
import java.sql.DriverManager;
import java.sql.Connection;
import java.sql.Statement;
import java.sql.ResultSet;
import java.sql.SQLException;

/**
 * This sample program is a minimal JDBC application showing
 * JDBC access to Cloudscape.
 *
 * Instructions for how to run this program are 
 * given in <A HREF=example.html>example.html</A>.
 * 
 * Cloudscape applications can run against Cloudscape running in an embedded 
 * or a client/server framework. When Cloudscape runs in an embedded framework,
 * the Cloudscape application and Cloudscape run in the same JVM. The application
 * starts up the Cloudscape engine. When Cloudscape runs in a client/server framework,
 * the application runs in a different JVM from Cloudscape. The application only needs
 * to start the client driver, and the connectivity framework provides network connections. 
 * (The server must already be running.)
 *
 * <p>When you run this application, give one of the following arguments:
 *    * embedded (default, if none specified)
 *    * rmijdbcclient (if Cloudscape is running embedded in the RmiJdbc Server framework)
 *    * sysconnectclient (if Cloudscape is running embedded in the Cloudconnector framework)
 *
 * @author janet
 */


public class CloudscapeSimpleApp {
	/* the default framework is embedded*/
	public String framework = "embedded";
	public String driver = "COM.cloudscape.core.JDBCDriver";
	public String protocol = "jdbc:cloudscape:";
	
	
    public static void main(String[] args) {
		
		new CloudscapeSimpleApp().go(args);
		
    }
	
    void go(String args[]) {
		/* parse the arguments to determine which framework is desired*/
		parseArguments(args);
		
        System.out.println("CloudscapeSimpleApp starting in " + framework + " mode.");	
        try {
            
            /*
			  The driver is installed by loading its class.
			  In an embedded environment, this will start up Cloudscape, since it is not already running.
			  */
            Class.forName(driver).newInstance();
            System.out.println("Loaded the appropriate driver.");
			
            /*
			  The connection specifies create=true to cause
			  the database to be created. To remove the database,
			  remove the directory CloudscapeDB and its contents.
			  The directory CloudscapeDB will be created under
			  the directory that the system property 
			  cloudscape.system.home points to, or the current
			  directory if cloudscape.system.home is not set.
			  
			  
			  */
            Connection conn = DriverManager.getConnection(
														  protocol + "CloudscapeDB;create=true");
            System.out.println("Connected to and created database CloudscapeDB");
			
            /*
			  We could also turn autocommit off by putting
			  ;autocommit=false on the URL.
			  */
            conn.setAutoCommit(false);
			
            /*
			  Creating a statement lets us issue commands against
			  the connection.
			  */
            Statement s = conn.createStatement();
			
            /*
			  We create a table, add a few rows, and update one.
			  */
            s.execute("create table CloudscapeDB(num int, addr varchar(40))");
            System.out.println("Created table CloudscapeDB");
            s.execute("insert into CloudscapeDB values (1956,'Webster St.')");
            System.out.println("Inserted 1956 Webster");
            s.execute("insert into CloudscapeDB values (1910,'Union St.')");
            System.out.println("Inserted 1910 Union");
            s.execute(
					  "update CloudscapeDB set num=180, addr='Grand Ave.' where num=1956");
            System.out.println("Updated 1956 Webster to 180 Grand");
			
            s.execute(
					  "update CloudscapeDB set num=300, addr='Lakeshore Ave.' where num=180");
            System.out.println("Updated 180 Grand to 300 Lakeshore");
			
            /*
			  We select the rows and verify the results.
			  */
            ResultSet rs = s.executeQuery("SELECT num, addr FROM CloudscapeDB ORDER BY num");
            if (!rs.next())
                throw new Exception("Wrong number of rows");
            if (rs.getInt(1)!=300)
                throw new Exception("Wrong row returned");
            if (!rs.next())
                throw new Exception("Wrong number of rows");
            if (rs.getInt(1)!=1910)
                throw new Exception("Wrong row returned");
            if (rs.next())
                throw new Exception("Wrong number of rows");
            System.out.println("Verified the rows");
			
            s.execute("drop table CloudscapeDB");
            System.out.println("Dropped table CloudscapeDB");
			
            /*
			  We release the result and statement resources.
			  */
            rs.close();
            s.close();
            System.out.println("Closed result set and statement");
			
            /*
			  We end the transaction and the connection.
			  */
            conn.commit();
            conn.close();
            System.out.println("Committed transaction and closed connection");
			
            /*
			  In embedded mode, an application should shut down Cloudscape.
			  If the application fails to shut down Cloudscape explicitly, 
			  the Cloudscape does not perform a checkpoint when the JVM shuts down, which means
			  that the next connection will be slower.
			  Explicitly shutting down Cloudscape with the URL is preferred.
			  
			  This style of shutdown will always throw an "exception".
			  */
            boolean gotSQLExc = false;
            if (framework.equals("embedded")) {
                try {
                    DriverManager.getConnection("jdbc:cloudscape:;shutdown=true");
                } catch (SQLException se) {
                    gotSQLExc = true;
                }
                if (!gotSQLExc)
                    System.out.println("Database did not shut down normally");
                else
                    System.out.println("Database shut down normally");
			}
		}
		catch (Throwable e) {
			System.out.println("exception thrown:");
			
            if (e instanceof SQLException) 
                printSQLError((SQLException)e);
            else
                e.printStackTrace();
        }
		
        System.out.println("CloudscapeSimpleApp finished");
    }
	
    static void printSQLError(SQLException e) {
        while (e != null) {
            System.out.println(e.toString());
            e = e.getNextException();
        }
    }
	
	private void parseArguments(String[] args) {
		
        int length = args.length;
		for (int index = 0; index < length; index++) {
			
			if (args[index].equalsIgnoreCase("rmijdbcclient")) {
				framework = "rmijdbc";
				driver = "COM.cloudscape.core.RmiJdbcDriver";
				protocol = "jdbc:cloudscape:rmi:";
			}
			if (args[index].equalsIgnoreCase("sysconnectclient")) {
				framework = "sysconnect";
				driver = "COM.cloudscape.core.WebLogicDriver";
				protocol = "jdbc:cloudscape:weblogic:";
			}
		}
		
		
	}
	
}




