package JBMSTours;
import java.sql.*;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import JBMSTours.Util;

/**
  * Look at the <a href="../JBMSTours/ApplicationMode.java">Source</a>.<P>
  * This class is used by the classes with main methods. It parses the
  * arguments on the command line. If a host name and port number 
  * are specified, then the application is a client application and
  * uses the Cloudconnector driver name and Cloudconnector URL. If there is additionally
  * a single character, "r," the application is an RmiJdbc client application,
  * and the correct driver and URL are used. Otherwise, the
  * application is in the embedded mode. The driver name is the standard
  * embedded Cloudscape driver and the URL is the standard Cloudscape URL.
  *
  * <p>It includes methods for connecting and disconnecting. You can use 
  * this to connect to a database.
  * It connects in the correct mode for the application (embedded or client).
  * <p>It assumes that the name of the database is <i>toursDB</i>. To connect
  * to a different database, supply a value for the property <i>JBMSTours.dbName</i>. 
  */
public class ApplicationMode {


/**
  * embedded mode
  *
  */
    public static final short EMBEDDED = 1;
/**
  * Cloudconnector client mode
  *
  */
    public static final short CLOUDCONNECTCLIENT = 2;

/**
  * RmiJdbc client mode
  *
  */
    public static final short RMIJDBCCLIENT = 4;


/**
  * embed driver name
  *
  */

    public static final String EMBEDDEDDRIVER = "COM.cloudscape.core.JDBCDriver";

/**
  * Cloudconnector client driver name
  *
  */

    public static final String CLIENTDRIVER = "COM.cloudscape.core.WebLogicDriver";

/**
  * RmiJdbc client driver
  *
  */


    public static final String RMICLIENTDRIVER = "COM.cloudscape.core.RmiJdbcDriver";




    private short whichmode= EMBEDDED;
/**
  * hostname for server
  *
  */
    private String hostname = "";
/**
  * portnumber server is listening on
  *
  */

    private String portnumber = "";
/**
  * driver class name
  *
  */
    private String driver = EMBEDDEDDRIVER;
/**
  * protocol (the protocol and subprotocol of the database connection URL)
  *  (e.g., jdbc:cloudscape: or jdbc:cloudscape:weblogic:)
  *
  */

    private String protocol = "jdbc:cloudscape:";

    /**
     * By default, the name of the database (dbName) is toursDB. 
     * You can override this by setting the
     * property <i>JBMSTours.dbName</i> to an alternate name.
     *
     */
    private String dbName="toursDB";

    /**
     * In case we need anything else on the URL.
     *
     */
	private String extra="";


	private boolean appropriateDriverBooted=false;
	private short numberOfConnections=0;

    public ApplicationMode () {
    setDatabaseName();
    }

    /**
     * constructor for Cloudconnector client application mode.
     */
    public ApplicationMode (String hst, String prt) {
      setCloudconnectVariables(hst, prt);
      setDatabaseName();
    }

    /**
     * Constructor for RmiJdbc client application mode
     *
     */
    public ApplicationMode (String hst, String prt, String r) {
      setRmiJdbcVariables(hst, prt);
      setDatabaseName();
    }



    /**
     * Constructor
     *
     */

  public ApplicationMode (String[] arg) {
    this();
    if (arg.length == 2) {
      setCloudconnectVariables(arg[0], arg[1]);
    }
    if (arg.length == 3) {
      setRmiJdbcVariables(arg[0], arg[1]);
    }
    if (arg.length != 0 && arg.length  != 2 && arg.length != 3) 
 
		System.out.println("Bad argument length. Assuming embedded mode.");
    setDatabaseName();
  }

    /**
     * Sets the database name--either the default, <i>toursDB</i>, 
     * or an alternate provided in a JVM property
     *
     */

  private void setDatabaseName() {
        dbName = getDatabaseName();
  }

    /**
     * Returns the database name--either the default, <i>toursDB</i>, 
     * or an alternate provided in a JVM property
	 * @return String database name
     *
     */
	public static String getDatabaseName() {
		String dbn = System.getProperty("JBMSTours.dbName", "toursDB");
		return dbn;
	}
	
	private void setCloudconnectVariables(String h, String p) {
		whichmode = CLOUDCONNECTCLIENT;
		hostname = h;
		portnumber = p;
		driver = CLIENTDRIVER;
		protocol = "jdbc:cloudscape:weblogic:";
		//we need to turn off prefetching for the connection so that we can use one updatable cursor.
		extra = "&weblogic.t3.cacheRows=0";
	}
	
	private void setRmiJdbcVariables(String h, String p) {
		whichmode = RMIJDBCCLIENT;
		hostname = h;
		portnumber = p;
		driver = RMICLIENTDRIVER;
		protocol = "jdbc:cloudscape:rmi:";
		
	}
	
	
    /**
     * Smart Method for connecting to Cloudscape.
     * Knows whether you are embedded or
     * a client. Takes the ;attribute=value portion of the URL as an argument<p>
     *
     *
     *
     * @param     String     ;attribute=value portion of the URL 
     *
     * @return    Connection  Connection to the database
     *
     * @exception Throwable  if an error occurs
     */

    public Connection connectToCloudscape(String andAttributes) throws Throwable {

        Connection conn;
        /* the default driver and URL are for the embedded mode.*/

        String URL = protocol + dbName + andAttributes;
        String appmode = "embedded";
        
	/* if the application is a client application to a Cloudconnect,
         * sets the full client URL
         */
	if (whichmode == CLOUDCONNECTCLIENT) {
	  URL = protocol +
	    "//" + hostname + ":" + portnumber + "/" + 
            dbName + andAttributes + extra;
          appmode = "Cloudconnector client";
	}

	if (whichmode == RMIJDBCCLIENT) {
		URL = protocol + "//" + hostname + ":" + portnumber + "/" +dbName + andAttributes;
		appmode = "RmiJdbc client";

	}

	if (whichmode == EMBEDDED) {
		
	  /* we can set Cloudscape properties programmatically, if we do it
		 before loading the cloudscape driver. Here, we're turning off
		 row-level locking in an embedded system, which is single-user in this
		 application.
		 Row-level locking uses more overhead and is not needed.*/
		
		java.util.Properties p = System.getProperties();
		p.put("cloudscape.storage.rowLocking", "false");
		
	}
	
	if (!appropriateDriverBooted) {
		Class.forName(driver).newInstance();
		appropriateDriverBooted=true;	
		println("Loaded the " + appmode + " JDBC driver");
	}
	try {
		conn = DriverManager.getConnection(URL);
		numberOfConnections++;
	}
	catch (SQLException sqle) {
		String sql_state = (sqle).getSQLState();
		Util.println(sql_state);
        if (sql_state.equals("XJ006")|| sql_state.equals("04501")) {
			
			conn=DriverManager.getConnection(URL, getUserName(), getPassword());
		}
		
		else throw(sqle);
		
	}
 	println("Connected to database " + dbName);
	conn.setAutoCommit(false);
	return conn;
    }
	
    /**
     * Smart Method for connecting to Cloudscape.
     * Knows whether you are embedded or
     * a client. No arguments (no attributes).<p>
     *
     *
     *
     * @return    Connection  Connection to the database
     *
     * @exception Throwable  if an error occurs
     */

    public Connection connectToCloudscape() throws Throwable {
        return connectToCloudscape("");
    }

    /**
     * Smart Method for quitting Cloudscape. Knows whether you are embedded or
     * a client. If you are embedded, it shuts down Cloudscape. 
     * If not, it doesn't.
     *
     * @return    boolean    if the application quit Cloudscape properly
     */

    public boolean quitCloudscape() {
 
        boolean correct = true;
        boolean gotSQLExc = false;

	if (whichmode == EMBEDDED) {
            try {

                DriverManager.getConnection("jdbc:cloudscape:;shutdown=true");
            } catch (SQLException se) {
				/* shut down always throw an exception if correct--catch it*/
                gotSQLExc = true;
                appropriateDriverBooted=false;

            }
            if (!gotSQLExc) {
                System.out.println("Database did not shut down normally");
                correct = false;
			}
            else
                System.out.println("Database shut down normally");
	}
	
	return true;
        
  }


	private static String getUserName() throws java.io.IOException {
		return queryUser("Provide a valid user name");
		
	}

    private static String getPassword() throws java.io.IOException {
		return queryUser("Provide the password for that user name");

	}


	private static String queryUser(String prompt) throws java.io.IOException {

        String reply = null;

        BufferedReader d = new BufferedReader( new InputStreamReader(System.in));
        System.out.print(prompt + ": ");
        try
			{
				reply = d.readLine();
			}
        catch (java.io.IOException e)
			{
				System.out.println("Could not read response from stdin");
				throw(e);
				
			}
		finally {
			d.close();
		}
        return reply;
		
	}
	
	private void println(Object o) {
		
		if (numberOfConnections<2)
			Util.println(o);
	}
	
}







