package JBMSTours;

import java.sql.*;
import java.util.Properties;
import JBMSTours.serializabletypes.City;
import JBMSTours.serializabletypes.Person;
import JBMSTours.serializabletypes.Adult;
import JBMSTours.serializabletypes.Child;
import JBMSTours.serializabletypes.Tour;
import JBMSTours.Util;
/**
 * Look at the <a href="../JBMSTours/BuildATour.java">source</a>.<p>
 * Run this program after creating the database with 
 * <a href="JBMSTours.CreateToursDB.html">CreateToursDB</a>.
 * 
 * (It outputs messages as it runs. It takes a few minutes.)<p>
 *
 * This program queries the core tables and builds one customized 
 * tour object, storing booking information in the database 
 * (<em>FlightBookings</em>, <em>FlightAvailability</em>,
 * <em>HotelBookings</em>, and <em>HotelAvailability</em>), 
 * and storing the tour objects in the table <em>CustomizedTours</em>. 
 * It also uses the table <em>FlightObjects</em> as a temporary 
 * repository for Flight objects.<p>
 *
 * This is a simple program that doesn't have a user interface. (Instead of
 * getting data from user input, it gets it from a hash table).
 * It does present a really simple scenario of how to use the classes
 * and the database.  It also provides an example of how to set a 
 * property within an application.
 *
 * <p>This class has a main() method and can be run from the command-line.
 * For complete instructions on how to run it, see the
 * <a href="../example.html#running">example instructions</a>.
 * 
 */

public class BuildATour {

	Adult   me;
	Adult   myhusband;
	Child   mychild;
	Group   myGroup; 
	City    city1;
	City    city2;
	City    city3;
	Date    firstTravelDate;
	Date    thirdTravelDay;
	Date    secondTravelDay;
	JCalendar myCalendar;
	CustomerData  cd;
	

    public ApplicationMode mymode;
	
	
    /* The start of each duration */
    private long beginSeconds;

	public BuildATour(ApplicationMode am) {
		mymode=am;
	}
	
    public static void main(String[] args) {
		
		/* testing for the client mode information (hostname and portnum)*/
		ApplicationMode runmode = new ApplicationMode(args);
        (new BuildATour(runmode)).go(false);
    }
	
    void go(boolean prebuild) {
        System.out.println("BuildATour starting");
		
        try {
            Connection      conn;
			
            /* Init the start time */

		    Date now = JCalendar.getToday();
			
            beginSeconds = now.getTime()/1000;

        /*   Adjust one property, cloudscape.language.statementCacheSize=100.
         *   The typical way of setting properties is in the 
         *   cloudscape.properties file in the system directory. You can 
         *   also set them as shown here.  Setting this property to a 
         *   high value allows us to re-use statements.  Since we are 
         *   building 4 different tours, we won't have to recompile
         *   statements or prepared statements for each tour. Must set 
         *   property before loading the driver.
         */
  
        Properties p = System.getProperties();
        p.put("cloudscape.language.statementCacheSize", "100");
        System.setProperties(p);


            /* Connect */
            conn = mymode.connectToCloudscape();

            buildOneTour(conn, prebuild);

            /* We're done */
            conn.close();
            mymode.quitCloudscape();

        } catch (Throwable e) {
            System.out.println("exception thrown:");

            CreateToursDB.errorPrint(e);
        }

        System.out.println("BuildATour finished");
    }


   /**
     * gets people and group data from a hash table (instead of from users!)
     * and sets the city and travel date variables according to 
     * preferences of the group
     *
     * @param conn    Current connection to the database.
     *
     *
     * @exception     SQLException   if database error
     * @exception     Throwable      if Connection error
     */

	public void getPeopleAndGroupData(Connection conn) throws SQLException, Throwable {

		conn.setAutoCommit(true);
		System.out.println("(. . . Getting information about group and travel preferences . . . )" + Util.lineSeparator());
		Statement s = conn.createStatement();
		
		int currentNumberOfGroups = 0;
		ResultSet curNum;
		curNum = s.executeQuery("SELECT COUNT(group_id) FROM GROUPS");
		while (curNum.next() ) {
			currentNumberOfGroups = curNum.getInt(1);

		}
		curNum.close();

        cd.setKeyStart(currentNumberOfGroups+1);    
		
		/* get the cities of the customer's preference out of database*/
		city1 = City.findCity(conn, cd.getCity1ID());
		city2 = City.findCity(conn, cd.getCity2ID());
		city3 = City.findCity(conn, cd.getCity3ID());
		
		/* construct the real group object */
		int gid = 0;
		myGroup = new Group(
							gid,
							1,
							2,
							1,
							city1.city_id,
							cd.getLevel(),
							cd.getBudget());
		
		/* set the group's address and phone number*/
		
		myGroup.setAddress(cd.getAddress());
		myGroup.setPhone(cd.getPhone());
		myGroup.insertGroup(conn);
		gid=myGroup.resetGroupIdToAutoincrementValue(conn);
		
	    me = cd.getPrimaryAdult();
		myGroup.addMainPersonToGroup(conn, me);
	    myhusband = cd.getSecondaryAdult();    
		myGroup.addToGroup(conn,myhusband);
		mychild = cd.getChild(me.getId());
		myGroup.addToGroup(conn, mychild);
		
		/* initialize the dates to the group's travel dates from hash table*/
		firstTravelDate = myCalendar.getDay(cd.getDay1());
		secondTravelDay = myCalendar.getDay(cd.getDay2());
		thirdTravelDay = myCalendar.getDay(cd.getDay3());
		
		conn.setAutoCommit(false);
	}
	
	
	/**
     * Builds one Tour object.
     *
     * @param conn    Current connection to the database.
     * @param boolean whether we're prebuilding the database or not
     *
     *
     * @exception     SQLException          if database error
     * @exception     CityChoiceException   if it can't build tour to
     *                                      requested cities
     * @exception     BudgetException       if it can't build tour to
     *                                      requested cities within budget
     * @exception     AvailabilityException if it can't build tour to
     *                                      requested cities because flights or hotels are not available
	 *
	 * @exception     Throwable             if database boot/connection problem
     */


	public void buildOneTour(Connection conn, boolean prebuild) throws java.sql.SQLException,
		CityChoiceException,
		BudgetException,
		AvailabilityException,
	    Throwable {

            System.out.println("Creating a tour");

            /*initializes an array of days beginning with today*/
            myCalendar = new JCalendar();
            Tour    myTour;

            /* initializes the hash table of data about people and groups*/
            cd = new CustomerData(prebuild);

            /* initializes the hash table of people and group data*/
            getPeopleAndGroupData(conn);

            /* construct the tour */
            myTour = new Tour(
                            myGroup,
                            city1,
                            city2,
                            city3,
                            firstTravelDate,
                            secondTravelDay,
                            thirdTravelDay);
            myTour.printWelcomeGroup(conn);
            myTour.buildTheTour(conn);
            conn.commit();
            myTour.printTourInfo(conn);
            myTour.storeInDatabase(conn);
            myGroup.updateRunningTotal(conn);    
            conn.commit();
            System.out.println("created one tour.");
  }

}






