package JBMSTours;

import java.sql.*;
import JBMSTours.inserters.*;
import JBMSTours.Util;

/**
 *   Look at the <a href="../JBMSTours/CreateToursDB.java">source</a>.<p>
 *   Main database creation program. Also populates several tables.
 *   To use, type
 *   <p>java <code>JBMSTours.CreateToursDB</code>.<p>
 *   Use
 *   the parameter "syncsource" if you want to create a synchronization source database.
 *
 *   Another legal parameter is <code>extra.</code>
 *   Use of this parameter is for
 *   building the pre-built version of the database included with
 *   the installation. It puts extra data into some tables, preloads the JDBC metadata
 *   stored prepared statements and the SysVisual stored prepared statements,
 *   and automatically runs JBMSTours.BuildATour four times. To use this
 *   parameter, you must have <i>tools.jar</i> in your class path in addition
 *   to <i>cloudscape.jar</i>.
 *   Most users will not use this parameter. Use one or the other 
 *   (<code>extra</code> or <code>syncsource</code>),  not both.<p>
 *
 *   <code>java JBMSTours.CreateToursDB syncsource</code> <p>
 *   <code>java JBMSTours.CreateToursDB extra</code> <p>
 *
 * By default, this application runs in the embedded mode. To run it as a
 * client application accessing a running Cloudconnect, 
 * give the hostname
 * and the port number of the Cloudconnector you want to connect to. For example:
 * <br><br><code>java JBMSTours.CreateToursDB jeeves 7001</code>
 * <p>To combine those arguments with other parameters, put them last.
 * <br><br><code>java JBMSTours.CreateToursDB syncsource jeeves 7001</code>
 * <br><code>java JBMSTours.CreateToursDB extra jeeves 7001</code>
 * <p>To run it as a client application accessing RmiJdbc Server, add the single character 'r' after
 * hostname and port number. For example:
 * <br><br><code>java JBMSTours.CreateToursDB localhost 1099 r
 * <br><br>java JBMSTours.CreateToursDB syncsource localhost 1099 r</code>
 * <p>What this application does:
 *   <ul>
 *   <li>Creates and then connects to <em>toursDB</em> database.
 *   <li>Creates tables and indexes.
 *   <li>Calls classes in the JBMSTours.inserters package to populate core tables.</li>
 *   <li>(If running in "extra" mode, calls
 *       <a href="JBMSTours.BuildATour.html">BuildATour</a>).<p>
 *   </ul>
 *   Run this program first. Then run
 *   <a href="JBMSTours.BuildATour.html">BuildATour</a>
 *   (constructs tour objects and populates other tables).
 *   Optional programs:
 *   <a href="JBMSTours.GenerateReport.html">GenerateReport</a>
 *   (uses aggregates to report about data),
 *   <a href="JBMSTours.ArchiveData.html">ArchiveData</a>
 *   (uses database-side JDBC methods, connecting to two different
 *   databases)
 *   <a href="JBMSTours.RunTime.html">RunTime</a>
 *   (illustrates use of runtime statistics to analyze query performance), and <a href="JBMSTours.AdminHelper.html">AdminHelper</a>.<p>
 *
 * @see CreateSchema
 * @see InsertCities
 * @see InsertCountries
 * @see InsertHotels
 * @see InsertFlights
 * @see InsertAvailability
 * @see InsertMaps
 * @see InsertJars
 * @see BuildATour
 */
public class CreateToursDB 
{
    /* without any parameters, this program builds the plain rather than
     * the syncsource database. It builds the plain database without the
     * "extra" data. If you specify syncsource, it builds the database as
     * a synchronization "source" database.
     * If you specify extra (used only for release), it loads extra random
     * data and also pre-builds some tours.
     */
    /**
     *
     * whether you are in embedded or client mode
     */
    ApplicationMode mymode;

    /**
     * 
     **/
    private short howToBuild;

    /**
     *
     **/
    private String buildmode = "plain";

    /**
     *
     **/
    private static final short NORMAL = 0;

    /**
     *
     **/
    private static final short EXTRA = 1;

    /**
     *
     **/
    private static final short SYNCSOURCE = 3;

    /**
     *
     **/
    private int exitstatus = 0;



    /**
     *
     **/
    public static void main(String[] args)
    {
        (new CreateToursDB()).go(args);
    }

    /**
     *
     **/
    void go(String[] args)
    {
        howToBuild = NORMAL;
	mymode = new ApplicationMode(parseArguments(args));
        System.out.println("CreateToursDB starting");
        System.out.println("Building the sample database, " + mymode.getDatabaseName()  + ".");
        System.out.println("Using the buildmode: " + buildmode);
        if(howToBuild == EXTRA) {
	  /* for the pre-built version of the database, we're loading it ahead of time
             with the stored prepared statements used by the JDBC metadata. */
            java.util.Properties p = System.getProperties();
            p.put("cloudscape.jdbc.metadataStoredPreparedStatements", "onDatabaseCreation");
	}

        try
        {

            Connection conn;
            String shortURL = "";
            if(howToBuild != SYNCSOURCE) {
                shortURL = ";create=true";
	    }
            else {
                /* to build a synchronization source database, we use the createSource attribute
                 to true on the URL"*/
                shortURL = ";createSource=true";
	    }
   
            conn = mymode.connectToCloudscape(shortURL);

            if(howToBuild == EXTRA) {
	  /* pause to allow thread creating metadata statements to finish*/

                try {Thread.sleep(120000L); } catch (InterruptedException e ) {}
                System.out.println("Added JDBC metadata stored prepared statements.");

	    }

            CreateSchema cs = new CreateSchema();
            if(!cs.createMainDatabaseObjects(conn))
                return;

            Inserters insertData;
            insertData = new InsertCountries();
            if(insertData.insert(conn))
                conn.commit();
            else
                conn.rollback();
            insertData.confirmInsert(conn);

            insertData = new InsertCities();
            if(insertData.insert(conn))
                conn.commit();
            else
                conn.rollback();
            insertData.confirmInsert(conn);

            insertData = new InsertFlights();
            if(insertData.insert(conn))
                conn.commit();
            else
                conn.rollback();
            insertData.confirmInsert(conn);

            insertData = new InsertHotels();
            if(insertData.insert(conn))
                conn.commit();
            else
                conn.rollback();
            insertData.confirmInsert(conn);

            insertData = new InsertMaps();
            if(insertData.insert(conn))
                conn.commit();
            else
                conn.rollback();
            insertData.confirmInsert(conn);


            /* the synchronization source version of the database uses
               fixed data starting with a fixed date. Otherwise,
               the database populates the FlightAvailability
               and HotelAvailability tables with random data starting
               with today.*/
            if(howToBuild == NORMAL)
                insertData = new InsertAvailability(400,true);
            if(howToBuild == SYNCSOURCE)
                insertData = new InsertAvailability(400,false);
            /*
             * if we're in the "extra" mode, inserts extra data
             * and automatically runs BuildATour 4 times.
             */
            if(howToBuild == EXTRA)
                insertData = new InsertAvailability(1000,true);
            if(insertData.insert(conn))
                conn.commit();

	    cs.createStoredPreparedStatements(conn);
	    cs.createTriggers(conn);
		conn.commit();

            if(howToBuild == EXTRA)
            {
                (new BuildATour(mymode)).buildOneTour(conn,true);
                (new BuildATour(mymode)).buildOneTour(conn,true);
                (new BuildATour(mymode)).buildOneTour(conn,true);
                (new BuildATour(mymode)).buildOneTour(conn,true);
				/* now add stored prepared statements used by SysVisual, so that initial boot-up time will be faster*/
                COM.cloudscape.tools.databasePreloader.addSysVisualStatements(conn); 
                System.out.println("Added SysVisual stored prepared statements.");
			}
            if (howToBuild == EXTRA || howToBuild == SYNCSOURCE) {

				//now store the logic in the database
				//for application-side class loading, be sure to remove JBMSTours from the user's class path
				
				//and to use the bootstrap program to start the app
				try {
					insertData = new InsertJars();
					insertData.insert(conn);
					insertData.confirmInsert(conn);
					cs.createDatabaseClasspath(conn);

				}
				catch (SQLException e) {
					exitstatus=1;
					if (e.getSQLState().equals("X0X63"))
						System.out.println("Couldn't find jar file, didn't store logic in database");
					else
						throw e;
				}
				
				
            }
            conn.commit();
            conn.close();
            System.out.println("Committed transactions and closed connection");
            mymode.quitCloudscape();
        }
        catch(Throwable e)
        {
            System.out.println("exception thrown:");
            errorPrint(e);
        }
        System.out.println("CreateToursDB finished");
		System.exit(exitstatus);
    }

    /**
     * Parses the arguments to main.
     **/
    private String[] parseArguments(String args[])
    {
        String newargs[] = args;
	
        int length = args.length;
	for (int index = 0; index < length; index++) {
	  if (args[index].equalsIgnoreCase("extra")) {
                howToBuild = EXTRA;
                buildmode = "extra";
		return stripAString(args, index);
	  }
	  if (args[index].equalsIgnoreCase("syncsource") | args[index].equalsIgnoreCase("repsource")) {
                howToBuild = SYNCSOURCE;
                buildmode = "synchronization source";
		return stripAString(args, index);
	  }
		
	}

	return newargs;
    }

    /* Removes a String from a String array at the specified index.*/
	
	private String[] stripAString(String[] arg, int index) {

    int length = arg.length;
    String[] newargs = new String[length-1];
    int index2 = 0;
    for (int j = 0; j < length-1; j++) {
        if (index2 ==index){
	    index2++;
        }
        newargs[j] = arg[index2];
	index2++;
    }
    return newargs;
  }
    /**
     *   Print out the SQL errors.
     **/
    static void errorPrint(Throwable e)
    {
        if(e instanceof SQLException)
            SQLExceptionPrint((SQLException)e);
        else
        {
            System.out.println("A non SQL error occured.");
            System.out.println(e);
            e.printStackTrace();
        }
    }

    /**
     * Get the details in the SQL errors.
     **/
    static void SQLExceptionPrint(SQLException sqle)
    {
        while(sqle != null)
        {
            System.out.println(Util.lineSeparator() + "--SQLException Caught--" + Util.lineSeparator());
            String sql_state = (sqle).getSQLState();
            System.out.println("SQLState:   " + sql_state);
            System.out.println("Error Code: " + (sqle).getErrorCode());
            System.out.println("Message:  " + (sqle).getMessage());
            sqle.printStackTrace();
            if(sql_state.equals("XJ004"))
            {
                System.out.println("Hint: Your system directory is set to " + System.getProperty("cloudscape.system.home",System.getProperty("user.dir")));
                System.out.println("That means that Cloudscape is looking for toursDB in that directory.");
                System.out.println("Pre-built version of database is in /demo/databases off of cloudscape_home");
                System.out.println("If you are doing the tutorial, your version of the database should be in your tutorial_system directory.");
            }
            sqle = sqle.getNextException();
        }
    }

}








