package JBMSTours;

import java.sql.Connection;
import java.sql.Statement;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Time;
import java.sql.Date;
import java.sql.SQLException;
import JBMSTours.serializabletypes.*;

/**
 * Look at the <a href="../JBMSTours/FlightBuilder.java">Source</a>.<p>
 * Build a table of flights that may be useful for a particular tour.
 * Given three cities, we collect useful flights between each pair of
 * cities.
 *
 * @see Flight
 * @see OneStopFlight
 * @see TransferFlight
 * @see CreateSchema
 */
public class FlightBuilder 
{
    /**
     * Amount of time we're willing to wait on a layover.  Used in
     * constructing a flight plan when there are no direct flights.
     * Expressed in hours.  Note that 23 hours is probably
     * unrealistically long.
     */
    public static final int LAYOVER_TOLERANCE = 23;

    /**
     *  Build flights between these cities.
     **/
    private City city1, city2, city3;

    
    /**
     *  id of the group that wants these flights.
     **/
    private int groupId;

    /**
     *  Connection to the database.
     **/
    private Connection conn;

    /**
     * These are all stored prepared statements.
     * See <a href="JBMSTours.CreateStoredPreparedStatements.html">CreateStoredPreparedStatements</a>.
     **/
    private static PreparedStatement getDirectFlights;

    /**
     *   
     **/
    private static PreparedStatement getFlightInfo;

    /**
     *
     **/
    private static PreparedStatement insertFlightObject;

    /**
     *
     **/
    private static PreparedStatement getFullFlightInfo;

    /**
     *
     **/
    private static PreparedStatement getFullTransferFlightInOneBlow;

    /**
     * Since some of these methods can be called as server-side methods, 
     * we set a flag to determine whether we can call the commit method.
     * By default, commits are allowed.
     *
     **/

    boolean commitsAllowed=true;

    /**
     * Constructs an object that sets up the
     * FlightObjects table of a particular database for a tour
     * involving certain cities.
     *
     * @param con A connection to the database
     * @param c1 The home city
     * @param c2 The first city to be visited
     * @param c3 The second city to be visited
     * @param gi The group_id
     */
    public FlightBuilder(Connection con, City c1, City c2, City c3, int gi)
    {
        conn = con;
        city1 = c1;
        city2 = c2;
        city3 = c3;
	getDirectFlights = null;
        getFlightInfo = null;
        insertFlightObject = null;
        getFullFlightInfo = null;
        getFullTransferFlightInOneBlow = null;
        groupId = gi;
        areCommitsAllowed();
    }


    /**
     * Constructs an object that sets up the
     * FlightObjects table of a particular database for flights
     * between only two particular cities.
     *
     * @param con A connection to the database
     * @param c1 city 1
     * @param c2 city 2
     * @param gi The group_id
     */
    public FlightBuilder(Connection con, City c1, City c2, int gi)
    {
        conn = con;
        city1 = c1;
        city2 = c2;
        city3 = null;
	getDirectFlights = null;
        getFlightInfo = null;
        insertFlightObject = null;
        getFullFlightInfo = null;
        getFullTransferFlightInOneBlow = null;
        groupId = gi;
        areCommitsAllowed();
    }

    /**
     * Find out whether the property "databaseSide" has been set. If so,
     * set the variable commitsAllowed to false.
     *     
     */

    private void areCommitsAllowed() {

    String p = System.getProperty("JBMSTours.databaseSide");
        if (p == null)
	  p = "false";
        if (p.equals("true")) 
	  commitsAllowed=false;

	}

    /**
     * Extract information on useful flights and store it in the
     * FlightObjects table.  The table stores instances of the Flight
     * class (including its two subclasses).  "Useful" flights are
     * those that fly between pairs of cities in our tour.
     *
     * @exception SQLException in case anything goes wrong during a SQL operation
     * @exception CityChoiceException if it can't build the tour between requested
     *                                cities
     *     
     */
    public void constructFlightObjectTable() throws SQLException, CityChoiceException
    {

        cleanOutFlightObjects(conn);
        commit();
        populateTheTable();
    }

    /**
     * Populate the FlightObjects table.
     *
     * @exception SQLException in case anything goes wrong during an
                  SQL operation
     * @exception CityChoiceException if it can't build the tour 
     *            between requested cities
     *     
     */
  public void populateTheTable() throws SQLException, CityChoiceException {



        Util.println("\nConstructing table of Flight objects for those cities.");
        prepareQueries(conn);
        /*
         * Populate the table with suitable flights between each pair
         * of cities in the tour.  If there are no direct flights
         * available, try flights that require a transfer.
         */
        if(!(constructDirectFlights(city1,city2)))
            constructTransferFlights(city1,city2);
        if (city3 != null) {
        if(!(constructDirectFlights(city2,city3)))
            constructTransferFlights(city2,city3);


        if(!(constructDirectFlights(city3,city1)))
            constructTransferFlights(city3,city1);
	}

        getDirectFlights.close();
        getFlightInfo.close();
        insertFlightObject.close();
        getFullFlightInfo.close();

  }


    /**
     * Deletes old information from the FlightAvailability table. The customer
     * doesn't need to know about the availability of flights that have already passed.
     * Used in SQL examples and called by archiveData.
     * There is an alias for this method called cleanOutFlightAvailability.
     *
     * @param  con    connection to the database
     *
     * @exception SQLException in case anything goes wrong during a SQL operation
     */
    public static void cleanOutFlightAvailability(Connection con) throws SQLException
    {

        Statement s = con.createStatement();
        s.executeUpdate("DELETE FROM FlightAvailability WHERE flight_date < CURRENT_JDATE()");
	s.close();
    }



    /**
     * "Prepares" the statements that execute the stored prepared statements.
     * Prepared statements now point to those originally prepared and stored in
     * CreateStoredPreparedStatements.
     * These stored prepared statements are already compiled. In this session,
     * the only thing we have to compile is the statement that executes them.
     *
     * @param  con    connection to the database
     *
     * @exception SQLException in case anything goes wrong during a SQL operation
     */
    private static void prepareQueries(Connection con) throws SQLException
    {

      if (getDirectFlights == null) {
        getDirectFlights = con.prepareStatement("EXECUTE STATEMENT getDirectFlights");
        getFlightInfo = con.prepareStatement("EXECUTE STATEMENT getFlightInfo");
        insertFlightObject = con.prepareStatement("EXECUTE STATEMENT insertFlightObject");
        getFullTransferFlightInOneBlow = con.prepareStatement("EXECUTE STATEMENT getFullTransferFlightInOneBlow");
        getFullFlightInfo = con.prepareStatement("EXECUTE STATEMENT getFullFlightInfo");
	}
    }

    /**
     * Builds direct flights--Flight and OneStopFlight object--for flights between two
     * cities and stores those objects in the database.
     *
     * @param c1 Originating city
     * @param c2 Destination city
     * @return the number of flights constructed
     *
     * @exception SQLException in case anything goes wrong during a SQL operation
     */
    private boolean constructDirectFlights(City c1, City c2) throws SQLException
    {
        int n = 0;
        Util.println("\nConstructing flight objects for flights from " + 
            c1.getName() + " to " + c2.getName());
        getDirectFlights.setString(1,c1.getAirport());
        getDirectFlights.setString(2,c2.getAirport());
        ResultSet rs = getDirectFlights.executeQuery();
        while(rs.next())
        {

            String flightNumber = rs.getString(1);
            n += putInTable(flightNumber,c1,c2);
        }
        commit();
        if(n > 0)
            return true;
        return false;
    }

    /**
     * After getting complete information about the direct flight,
     * Stores objects in the database.
     *
     * @ return   int   number of rows inserted
     *
     * @ exception SQLException in case anything goes wrong during a SQL operation
     */
    private int putInTable(String flightnumber, City origin, City destination) throws SQLException
    {
        Flight theFlight = makeADirectFlight(conn,flightnumber,origin.getAirport(),destination.getAirport());
        int rowsInserted = 0;
        if(theFlight.segmentNumber > 0)
        {
            insertFlightObject.setString(1,theFlight.getUniqueId());
            insertFlightObject.setObject(2,theFlight);
            insertFlightObject.setInt(3, groupId);
            Util.println(theFlight.toString()); 
            try {
                rowsInserted += insertFlightObject.executeUpdate();
            }

           catch(SQLException e)
           {
                alreadyExists(e);
            }

        }
        return rowsInserted;
    }

    /**
     * Construct TransferFlights for two cities.
     *
     * @param c1 Originating city
     * @param c2 Destination city
     *
     * @exception SQLException in case anything goes wrong during a SQL operation
     */
    private void constructTransferFlights(City c1, City c2) throws SQLException, CityChoiceException
    {
        boolean exists = false;
        Util.println("\nNo direct flights available. Looking for transfer flights.");
        String id1 = "";
        String id2 = "";
        int seg1 = 0;
        int seg2 = 0;
        getFullTransferFlightInOneBlow.setString(1,c1.getAirport());
        getFullTransferFlightInOneBlow.setString(2,c2.getAirport());
        ResultSet rs = getFullTransferFlightInOneBlow.executeQuery();
        while(rs.next())
        {
	    exists = true;
            id1 = rs.getString(1);
            seg1 = rs.getInt(2);
            id2 = rs.getString(3);
            seg2 = rs.getInt(4);
            makeATransferFlightAndPutInTable(id1,seg1,id2,seg2);
        }
        commit();
	if (!exists)
	  throw new CityChoiceException(c1, c2);
    }

    /**
     * Construct a TransferFlight for two cities and two flights.
     *
     * @param id1 First flight id
     * @param seg1 First segment id
     * @param id2 Second flight id
     * @param seg2 Second segment id
     * @exception SQLException in case anything goes wrong during a SQL operation
     */
    private void makeATransferFlightAndPutInTable(String id1, int seg1, String id2, int seg2) throws SQLException
    {
        TransferFlight comboFlight = makeATransferFlight(conn,id1,seg1,id2,seg2);
        if(comboFlight.segmentNumber > 0)
        insertFlightObject.setString(1,comboFlight.getUniqueId());
        insertFlightObject.setObject(2,comboFlight);
        insertFlightObject.setInt(3, groupId);
            try {
                insertFlightObject.executeUpdate();
            }

           catch(SQLException e)
           {
                alreadyExists(e);
            }
        Util.println(comboFlight.toString());
    }

    /**
     * A static method that finds a flight between two airports.
     * If there are no direct flights, looks for a transfer flights. Uses
     * stored prepared statements. If there are no transfer flights, it returns
     * an empty flights object. It doesn't necessarily return the best flight (although
     * it will always return a direct flight over a transfer flight).
     * Not used by this class, but available for use in SQL-J statements.<p>
     * For example, try "VALUES FlightBuilder.returnAnyFlight(getCurrentConnection(),
     * 'SFO', 'MAD').toString()"</p>
     * 
     * @param con  a connection to the databaseid1 First flight id
     * @param ap1  the first airport code
     * @param ap2  the second airport code
     *
     * @return    Flight  a flight
     * @exception SQLException in case anything goes wrong during a SQL operation
     *
     **/
    public static Flight returnAnyFlight(Connection con, String ap1, String ap2) throws SQLException
    {

	prepareQueries(con);
        getDirectFlights.setString(1,ap1);
        getDirectFlights.setString(2,ap2);
        ResultSet rs = getDirectFlights.executeQuery();
        while(rs.next())
        {
            String flightNumber = rs.getString(1);
            return makeADirectFlight(con,flightNumber,ap1,ap2);
        }
        /* if we got to this point, there were no direct flights*/

        getFullTransferFlightInOneBlow.setString(1,ap1);
        getFullTransferFlightInOneBlow.setString(2,ap2);
        rs = getFullTransferFlightInOneBlow.executeQuery();
        while(rs.next())
        {
            return makeATransferFlight(con,rs.getString(1),rs.getInt(2),rs.getString(3),rs.getInt(4));
        }
        Util.println("No flights.");
        return new Flight();
    }

    /**
     *  Constructs a Flight object or a OneStopFlight object given key information.
     **/
    private static Flight makeADirectFlight(Connection con, String flightnumber, 
        String ap1, String ap2) throws SQLException
    {
	prepareQueries(con);
        Flight partOne = null;
        Flight partTwo = null;

        /* flat SQL-92 values from a table, which we will use to create a new Flight object*/
        String flightid;
        int segnumber;
        String origairport;
        Time departtime;
        String destairport;
        Time arrivetime;
        double flyingtime;
        int miles;
        int whichrow = 0;
        int rowsInserted = 0;
        getFlightInfo.setString(1,flightnumber);
        rowsInserted = 0;
        /*
         * the query will have either one or two rows, depending
         * on the number of legs in the flight.
         */
        ResultSet rs = getFlightInfo.executeQuery();
        while(rs.next())
        {
            flightid = rs.getString(1);
            segnumber = rs.getInt(2);
            origairport = rs.getString(3);
            departtime = rs.getTime(4);
            destairport = rs.getString(5);
            arrivetime = rs.getTime(6);
            flyingtime = rs.getDouble(7);
            miles = rs.getInt(8);
            whichrow++;
            if(whichrow == 1)
            {
                /** constructing a flight object**/
                partOne = new Flight(flightid,segnumber,origairport,departtime,destairport,
                    arrivetime,flyingtime,miles);
                if(ap1.equals(partOne.orig_airport) && ap2.equals(partOne.dest_airport))
                {
                    return partOne;
                }
            }
            /* if the flight has two legs, then you need to make a one-stop flight*/
            if(whichrow == 2)
            {
                partTwo = new Flight(flightid,segnumber,origairport,departtime,destairport,
                    arrivetime,flyingtime,miles);
                OneStopFlight osf = new OneStopFlight(partOne,partTwo);
                if(ap1.equals(osf.orig_airport) && ap2.equals(osf.dest_airport))
                {
                    return osf;
                }
                else
                    if(ap1.equals(partTwo.orig_airport) && ap2.equals(partTwo.dest_airport))
                    {
                        return partTwo;
                    }
            }
        }
        rs.close();
        return new Flight();
    }

    /**
     *  This method constructs a transfer flight object given key information.
     **/
    private static TransferFlight makeATransferFlight(Connection con, String id1, 
        int seg1, String id2, int seg2) throws SQLException
    {
        Flight partOne;
        Flight partTwo;
        TransferFlight combo;

        /* create the first leg as a Flight object */
        getFullFlightInfo.setString(1,id1);
        getFullFlightInfo.setInt(2,seg1);
        ResultSet rs = getFullFlightInfo.executeQuery();
        if(!rs.next())
        {
            // there wasn't even one row in the result;
            // code could be added here to handle this unexpected
            // situation
            Util.println("No flight found.");
            return new TransferFlight();
        }
        partOne = new Flight(rs.getString(1),rs.getInt(2),rs.getString(3),rs.getTime(4),
            rs.getString(5),rs.getTime(6),rs.getDouble(7),rs.getInt(8));
        if(rs.next())
        {
        // there were multiple rows in the result; code could be
        // added here to handle this unexpected situation
        }
        /* now create the second leg as a Flight object */
        getFullFlightInfo.setString(1,id2);
        getFullFlightInfo.setInt(2,seg2);
        rs = getFullFlightInfo.executeQuery();
        if(!rs.next())
        {
            // there wasn't even one row in the result;
            // code could be added here to handle this unexpected
            // situation
            Util.println("No flight found.");
            return new TransferFlight();
        }
        partTwo = new Flight(rs.getString(1),rs.getInt(2),rs.getString(3),rs.getTime(4),
            rs.getString(5),rs.getTime(6),rs.getDouble(7),rs.getInt(8));
        if(rs.next())
        {
        // there were multiple rows in the result; code could be
        // added here to handle this unexpected situation
        }
        /* create the TransferFlight object */
        combo = new TransferFlight(partOne,partTwo);
        return combo;
    }

    /**
     * A static method that deletes stale entries from the flightObjects table.
     * Tours that have already started won't need the table anymore!
     * 
     * @param con  a connection to the database
     *
     * @exception SQLException in case anything goes wrong during a SQL operation
     *
     **/

     public static void cleanOutFlightObjects(Connection con) throws SQLException {

     Statement s = con.createStatement();
     s.executeUpdate("LOCK TABLE FlightObjects IN EXCLUSIVE MODE");
     s.executeUpdate("DELETE FROM flightobjects WHERE group_id IN (SELECT groups.group_id FROM groups, " +
                     "CustomizedTours WHERE groups.group_id = customizedtours.group_id AND " +
                     "customized_tour->begin < CURRENT_JDATE())");

  }

    /**
     * Catches expected exceptions that don't matter.
     *
     * @exception SQLException if exception is unexpected.
     *
     **/

   private static void alreadyExists(SQLException e) throws SQLException
   {
     //if an insert violates a primary key constraint, SQLException 23500 is thrown
     //this is expected if flightobject table has already been populated
     
      boolean expected = false;
      String sql_state = e.getSQLState();
      if(sql_state.equals("23500"))
         expected = true;
      if(!expected)
         throw e;
   }


  /* commits the connection, unless commits are not allowed */
  private void commit() throws SQLException {
        if (commitsAllowed)
            conn.commit();
  }



}







