package JBMSTours;

import java.lang.Throwable;
import java.sql.*;
import java.math.BigDecimal;
import JBMSTours.serializabletypes.Person;
import JBMSTours.serializabletypes.Adult;
import JBMSTours.serializabletypes.Child;
import JBMSTours.serializabletypes.Tour;
import JBMSTours.serializabletypes.City;
import JBMSTours.ApplicationMode;

/**
 * Look at the <a href="../JBMSTours/Group.java">Source</a>.<p>
 * Basic class for tracking the customer group. 
 * BuildATour constructs the group objects. Stored as relational
 * data in database.
 *
 * <p>Cloudscape automatically generates the integer value in the
 * group_id column (DEFAULT AUTOINCREMENT); this class uses the
 * <a href="#resetGroupIdToAutoincrementValue">resetGroupIdToAutoincrementValue</a>
 * method to find out what that
 * value was and to assign it to <code>this.groupId</code>.
 *
 */
public class Group {

    //////////////////
    // fields
    //////////////////

    /**
     * Identifier for the group. Stored in the group_id column.
     */
    private int groupId;

    /**
     * The number of children in the group. Stored in the number_kids column.
     */
    public int numberKids;

    /**
     * The number of adults in the group. Stored in the number_adults column.
     */
    public int numberAdults;

    /**
     * The total number of people in the group. Stored in the number_people column.
     */
    private int numberPeople;

    /**
     * The id of the main person in the group. Stored in the main_person column.
     * 
     */
    public int mainPerson;

    /**
     * The total number of hotel rooms needed by the group. Stored in the number_rooms col.
     */
    private int numberRooms;

    /**
     * The address to contact the group's main person. Stored in the address column.
     */
    public String address;

    /**
     * The phone number for contacting the group's main person. Stored in the phone column.
     */
    public String phone;


    /**
     * The city_id from which the group's trip will start. Stored in the city_id column.
     */
    public int city_id;

    /**
     * The level at which the group wishes to travel. Stored in the level column.  
     * See the static variables in JBMSTours.Tour
     */
    private short level;

    /**
     * The budget for the group's trip. Stored in the budget column.
     */
    private BigDecimal budget;

    /**
     * The total cost of the group's travel. Stored in the running_total column.
     */
    private BigDecimal runningTotal;


    private transient PreparedStatement get_number_of_seats;
    private transient PreparedStatement get_people_in_group;
    private transient PreparedStatement add_main_person_to_group;
    private transient PreparedStatement add_person;
    private transient PreparedStatement add_person_to_group;
    private static transient PreparedStatement get_the_group;

    //////////////////
    // constructors
    //////////////////

    /**
     * Create a new group, initializing it to have no members.
     */
    public Group() {

        groupId = 0;
        numberKids = 0;
        numberAdults = 0;
        numberPeople = 0;
        mainPerson = 0;
        numberRooms = 0;
        city_id = 0;
        address = "";
        phone = "";
        level = 0; 
        budget = new BigDecimal(0);
        runningTotal = new BigDecimal(0);
	setPreparedStatementsToNull();
    }


    /**
     * Create a new group with only a group_id.
     */
    public Group(int id) {
        this();
        groupId = id;
        level = Tour.ECONOMYTOURLEVEL;
    }



    /**
     * Constructs a group object. 
     *
     * @param id      The group_id.
     * @param k       Number of kids in the group.
     * @param a       Number of adults in the group.
     * @param id2     The person_id of the main person in the group.
     * @param rooms   The number of rooms needed by the group in a hotel.
     * @param acity   The id of the city where the group is from.
     * @param level   How the group wants to travel: first class, standard, or economy.
     * @param abudget How much money the group wants to spend.
     *
     */
    public Group (int id, int k, int a, int rooms, 
            int cid, String level, double abudget) {
        this();
        groupId = id;
        numberKids = k;
        numberAdults = a;
        numberPeople = k + a;
        numberRooms = rooms;
        city_id = cid;
        this.level = Tour.getShortForLevel(level); 
        budget = new BigDecimal(abudget);
        runningTotal = new BigDecimal(0);
    }



    /**
     * Constructs a group object from columns in the Groups table. 
     *  Uses all objects as parameters to make it easy
     *  for Cloudscape to find the signature in a SQL-J statement.
     *
     * @param id      The group_id.
     * @param k       Number of kids in the group.
     * @param a       Number of adults in the group.
     * @param np      Number of people
     * @param mp      The person_id of the main person in the group.
     * @param nr      The number of rooms needed by the group in a hotel.
     * @param acity   The id of the city where the group is from.
     * @param address The group's address
     * @param ph      the group's phone number
     * @param l       How the group wants to travel: first class, standard, or economy.
     * @param abudget How much money the group wants to spend.
     * @param rt      How much money the group has spent so far on the tour.
     *
     */


  public Group (Integer gid, Integer nk, Integer na, Integer np, Integer mp,
                   Integer nr, Integer cid, String ad, String ph, 
                   Short l, BigDecimal bud, BigDecimal rt) 
   {

        this(gid.intValue(), nk.intValue(), na.intValue(), nr.intValue(), 
        cid.intValue(), Tour.getStringForLevel(l.shortValue()), bud.doubleValue());
	mainPerson = mp.intValue();
	numberPeople = np.intValue();
        address = ad;
        phone = ph;
        runningTotal = rt;
	}

    /**
     * Constructs a group object from a row in a result set.
     *
     * @param     rs               a row in a a ResultSet (of select * from Groups)
     *
     * @exception SQLException     thrown on database access failure.
     */

    public Group (ResultSet rs) throws SQLException {
		//need to guard against nulls; fix after adding DEFAULTS
        this();
        groupId = rs.getInt(1);
        numberKids = rs.getInt(2);
		if (rs.wasNull())
			numberKids=0;
        numberAdults = rs.getInt(3);
		if (rs.wasNull())
			numberAdults=0;
        numberPeople = rs.getInt(4);
		if (rs.wasNull())
			numberPeople=numberKids+numberAdults;
		mainPerson = rs.getInt(5);
		if (rs.wasNull())
			mainPerson=0;
        numberRooms = rs.getInt(6);
		if (rs.wasNull())
			numberRooms=0;
        city_id = rs.getInt(7);
		if (rs.wasNull())
			city_id=1;
        address = rs.getString(8);
		if (rs.wasNull())
			address="";
        phone = rs.getString(9);
		if (rs.wasNull())
			phone="";
        level = rs.getShort(10);
		if (rs.wasNull())
			level=1;
        budget = rs.getBigDecimal(11, 2);
		if (rs.wasNull())
			budget=new BigDecimal(0);
        runningTotal = rs.getBigDecimal(12, 2);
		if (rs.wasNull())
			runningTotal=new BigDecimal(0);
    }




    /////////////////////
    // static methods
    /////////////////////

    /**
     * Get a group from the database given the identifier for it.
     *
     * @param     conn             the connection to the database.
     * @param     id               the id for the group to load.
     *
     * @return    Group            the group from the database.
     *
     * @exception SQLException     thrown on database access failure.
     */
    public static Group getGroup(Connection conn, int id)
            throws SQLException {

        Group theGroup = new Group();
	if (!prepareStaticStatements(conn))
	  Util.println("failed to prepare statements");
	get_the_group.setInt(1, id);
        ResultSet rs = get_the_group.executeQuery();
	rs.next();
        theGroup = new Group(rs);
        return theGroup;
    }

    //////////////////
    // output
    //////////////////

    /** 
     * Returns data about the group as a string..
     * @param     conn             the connection to the database.
     *
     * @return  String   object information about the Group
     * @exception SQLException     thrown on database access failure.
     */
    public String getInfo(Connection conn) throws SQLException {

        StringBuffer tourInfo = new StringBuffer();
        tourInfo.append("\nPrinting out information about " + this.toString(conn));
		tourInfo.append("\nTotal: " + numberPeople + 
						", " + 
						numberAdults + 
						" adults, " + 
						numberKids + " kids");

		tourInfo.append("\nMain person: " + 
						Person.findPerson(conn, mainPerson).toString());
	tourInfo.append("\nAddress: " + getAddress());
	tourInfo.append("\nPhone: " + getPhone());
	tourInfo.append("\nNumber of hotel rooms requested: " + numberRooms);
	tourInfo.append("\nNumber of airline seats needed: " + getSeats(conn));
	if (getSeats(conn) < numberPeople)
		tourInfo.append("\n(Children under age two typically do not need seats.)");
	tourInfo.append("\nTravelling " + Tour.getStringForLevel(level) + 
					" on a budget of $" + budget.toString());
	tourInfo.append("\nSo far, you have spent $" + runningTotal.toString());
	tourInfo.append("\nYou have this much left: $" + 
					(budget.subtract(runningTotal)).toString());
	return tourInfo.toString();
    }

    ///////////////////////
    // database accessors
    ///////////////////////

    /**

     * Write group info to the database.
     *
     * @param conn the connection to the database.
     *
     * @return the number of rows inserted (expected to be 1).
     *
     * @exception SQLException thrown if database access fails.
     */
    public int insertGroup(Connection conn) throws SQLException {
        PreparedStatement ps = conn.prepareStatement(
		   "INSERT INTO Groups (number_kids, number_adults, number_people, main_person, " +
		   "number_rooms, city_id, address, phone, tour_level, budget, running_total)" +
		   "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)");

		
        //ps.setInt(1, groupId);
        ps.setInt(1, numberKids);
		ps.setInt(2, numberAdults);
		ps.setInt(3, numberPeople);
		ps.setInt(4, mainPerson);
		ps.setInt(5, numberRooms);
		ps.setInt(6, city_id);
		ps.setString(7, address);
		ps.setString(8, phone);
		ps.setShort(9, level);
		ps.setBigDecimal(10, budget);
		ps.setBigDecimal(11, runningTotal);
		
        int j = ps.executeUpdate();
        return j;
		
    }
	
    /**

     * Update group info to the database (everything but group_id).
     *
     * @param conn the connection to the database.
     *
     * @return the number of rows updated (expected to be 1).
     *
     * @exception SQLException thrown if database access fails.
     */
  public int updateEntireGroup(Connection conn) throws SQLException {
        PreparedStatement ps = conn.prepareStatement(
            "UPDATE Groups SET number_kids = ?, number_adults = ?, number_people = ?, " +
            "main_person = ?, number_rooms = ?, city_id = ?, address = ?, phone = ?, " +
            "tour_level = ?, budget = ?, running_total = ? WHERE group_id = ?");
        ps.setInt(12, groupId);
        ps.setInt(1, numberKids);
	ps.setInt(2, numberAdults);
	ps.setInt(3, numberKids+numberAdults);
	ps.setInt(4, mainPerson);
	ps.setInt(5, numberRooms);
	ps.setInt(6, city_id);
	ps.setString(7, address);
	ps.setString(8, phone);
	ps.setShort(9, level);
	ps.setBigDecimal(10, budget);
	ps.setBigDecimal(11, runningTotal);
        int j = ps.executeUpdate();
        return j;
    

  }
    /**

     * Update the running total for the group in the database
     *
     * @param conn the connection to the database.
     *
     * @return the number of rows inserted (expected to be 1).
     *
     * @exception SQLException thrown if database access fails.
     */

  public int updateRunningTotal(Connection conn) throws SQLException {
        PreparedStatement ps = conn.prepareStatement(
            "UPDATE Groups SET running_total = ? WHERE group_id = ?");
        ps.setInt(2, groupId);
	ps.setBigDecimal(1, runningTotal);
        int j = ps.executeUpdate();
	Util.println("Group has been updated in the database.");
        return j;
    

  }

    /**
     * Get info about the last AUTOINCREMENT value inserted into the Groups table, and assign that
	 * number to this group as its groupId. 
     *
     * @param conn the connection to the database.
     *
     * @return the last AUTOINCREMENT value (the group_id for the group).
     *
     * @exception SQLException thrown if database access fails.
     */
	public int resetGroupIdToAutoincrementValue(Connection conn) throws SQLException {
		Statement s3 = conn.createStatement();
		// now get actual group number
		ResultSet rs3 = s3.executeQuery("VALUES ConnectionInfo.lastAutoincrementValue('APP', 'GROUPS', 'GROUP_ID')");
		while (rs3.next()) {
			
			groupId = rs3.getInt(1);
		}
		rs3.close();
		s3.close();
		return groupId;

	}



    /**
     * Add a the MainPerson to a group in the database. Updates the mainPerson
     * variable and inserts the
     * person into the PEOPLE table and also into Groups_People. Also updates
     * the main_person field in the Groups table.
     *
     * @param conn the connection to the database.
     * @param p the person to add.
     *
     * @return the number of rows inserted (expected to be 2).
     *
     * @exception SQLException thrown if database access fails.
     */
    public int addMainPersonToGroup(Connection conn, Person p) 
            throws SQLException {
		prepareStatements(conn);
              
        //add_person.setInt(1, p.id);
        add_person.setObject(1, p);
        int n=add_person.executeUpdate();
		p.updatePersonWithAutoincrementValue(conn);
		mainPerson = p.getId();
		add_main_person_to_group.setInt(1, mainPerson);
		add_main_person_to_group.setInt(2, groupId);
		n += add_main_person_to_group.executeUpdate();

		add_person_to_group.setInt(1, groupId);
        add_person_to_group.setInt(2, mainPerson);
        n+= add_person_to_group.executeUpdate();
        return n;
    }
    /**
     * Add a person to a group in the database. Inserts the
     * person into the PEOPLE table and also into Groups_People.
     *
     * @param conn the connection to the database.
     * @param p the person to add.
     *
     * @return the number of rows inserted (expected to be 2).
     *
     * @exception SQLException thrown if database access fails.
     */
    public int addToGroup(Connection conn, Person p) 
            throws SQLException {
	
        prepareStatements(conn);      
        //add_person.setInt(1, p.id);
        add_person.setObject(1, p);
        int n = add_person.executeUpdate();
		p.updatePersonWithAutoincrementValue(conn);

        add_person_to_group.setInt(1, groupId);
        add_person_to_group.setInt(2, p.id);
        n+= add_person_to_group.executeUpdate();
        return n;
    }

    /**
     * Returns a string listing all the people in the group. 
     * 
     * @param conn the connection to the database.
     *
     * @return a string concatenating the names of people in the group.
     *
     * @exception SQLException thrown if database access fails.
     */
    public String getPeopleInGroup(Connection conn) 
            throws SQLException {

        StringBuffer ourPeople = new StringBuffer();
        get_people_in_group.setInt(1, groupId);
        ResultSet rs = get_people_in_group.executeQuery();

        while (rs.next()) {
            if (ourPeople.length() > 0)
                ourPeople.append(", ");
            ourPeople.append(rs.getString(1)); 
        }

        return ourPeople.toString();
    }

    /**
     * Get the number of seats needed by the group. 
     * This is the number of people in the group over the age of 2.
     *
     * @param conn the connection to the database.
     *
     * @return number of seats needed
     *
     * @exception SQLException thrown on database access failure.
     */
    public int getSeats(Connection conn) throws SQLException {

        int seat = 0;
	prepareStatements(conn);
        get_number_of_seats.setInt(1, groupId);

        ResultSet rs = get_number_of_seats.executeQuery();

        while (rs.next()) {
            if (rs.getInt(1) > 2)
            seat++;
        }

        return seat;
    }
    /**
     * Get group's city of origin.
     *
     * @param conn the connection to the database.
     *
     * @return City  the group's city of origin.
     *
     * @exception SQLException thrown on database access failure.
     */
    public City getCityOrigin(Connection conn) throws SQLException {

        return City.findCity(conn, city_id);
    }

    /** Get the name of the group's city.
     *
     * @param conn the connection to the database.
     *
     * @return name of the group's city of origin.
     *
     * @exception SQLException thrown on database access failure.
     */
  public String getCityName(Connection conn) throws SQLException {

    return getCityOrigin(conn).toString();
  }


    /** 
     * Returns a welcome message for the Group
     *
     * @param conn the connection to the database.
     *
     *
     * @exception SQLException thrown on database access failure.
     */
    public String getWelcomeMessage(Connection conn) throws SQLException {
        
        StringBuffer welcomeMessage = new StringBuffer();
        welcomeMessage.append("\nWelcome to JBMSTours, ");
        welcomeMessage.append(getPeopleInGroup(conn));
        welcomeMessage.append("!");
        welcomeMessage.append("\nYour group has been stored in the database.\n");
        return welcomeMessage.toString();
    }




    //////////////////////////////////
    // state modifiers or accessors
    //////////////////////////////////

    /**
     * Add to the amount of money the group has spent.
     *
     * @param cost amount of money to add to the total spent.
     */
    public void addToRunning(BigDecimal cost) {
        runningTotal = runningTotal.add(cost).setScale(2, BigDecimal.ROUND_CEILING);
    }

    /**
     * Set the group's address.
     *
     * @param addres The address of the main person in the group.
     */
    public void setAddress(String add) {
        address = add;
    }


    /**
     * Set the group's phone.
     *
     * @param fone   The phone number of the main person in the group.
     */
    public void setPhone(String fone) {
        phone = fone;
    }


    /**
     * Return the group's available funds. This is the budget
     * less the current total spent.
     *
     * @return the available funds.
     */
    public BigDecimal getAvailableFunds() {
        return budget.subtract(runningTotal).setScale(2, BigDecimal.ROUND_CEILING);
    } 
    /**
     * Return the group's address.
     *
     * @return the address.
     */
    public String getAddress() {
        return address;
    } 
    /**
     * Return contact phone number for the group.
     *
     * @return phone.
     */
    public String getPhone() {
        return phone;
    } 

    /**
     * Return the id for the group
     *
     * @return the id for the group
     */
    public int getGroupId() {
        return groupId;
    }

    /**
     *
     * @return the level of travel.
     */
    public short getLevel() {
        return level;
    }

    /**
     * Return the number of rooms needed for the group.
     *
     * @return the number of rooms.
     */
    public int getNumberRooms() {
        return numberRooms;
    }

    /**
     * Return the number of people in the group.
     *
     * @return the number of people.
     */
    public int getNumberPeople() {
        return numberPeople; 
    } 

    /**
     * Convert the object to a string. Supplies the group Id
     * and originating city.
     *
     * @param Conn    Connection to the database
     *
     * @return string for the group object.
     *
     * @exception SQLException thrown on database access failure.
     */
    public String toString(Connection conn) throws SQLException {
        return "groupId " + groupId + " from " + getCityName(conn);
    }

    /**
     * Prepares some of the PreparedStatements for this class.
     *
     * @return boolean if statements are prepared without error
     *
     * @exception   SQLException   If there is a SQL error
     */

  public boolean prepareStatements(Connection conn) throws SQLException {

    if (add_main_person_to_group == null)

      add_main_person_to_group = conn.prepareStatement("UPDATE Groups set main_person = ? WHERE group_id = ?");

    if (add_person == null) 
      
        add_person = conn.prepareStatement("INSERT INTO people (Person) VALUES (?)");

    if (add_person_to_group == null) 

        add_person_to_group = conn.prepareStatement("INSERT INTO groups_people VALUES (?, ?)");
	

    if (get_people_in_group == null) 

        get_people_in_group = conn.prepareStatement(
            "SELECT person.getName() "+
            "FROM people, groups_people "+
            "WHERE people.person_id = groups_people.person_id "+
            "AND groups_people.group_id = ? ");

	
    if (get_number_of_seats == null) 

        get_number_of_seats = conn.prepareStatement(
          "SELECT (person.getAge() IS NULL) ? (30) : person.getAge() "+
          "FROM people, groups_people "+
          "WHERE people.person_id = groups_people.person_id "+
          "AND groups_people.group_id = ? ");
	

    if (add_person != null && add_main_person_to_group != null && add_person_to_group != null && get_number_of_seats != null)
      return true;
    else
      return false;

  }


 private void  setPreparedStatementsToNull() {
    get_number_of_seats=null;
    get_people_in_group=null;
    add_main_person_to_group=null;
    add_person=null;
    add_person_to_group=null;
    get_the_group=null;

  }
    /**
     * Prepares all the static PreparedStatements for this classs. 
     * This method is called by static methods.
     *
     * @return boolean if statements are prepared without error
     *
     * @exception   SQLException   If there is a SQL error
     */

  public static boolean prepareStaticStatements(Connection conn) throws SQLException {

	  if (get_the_group == null) {

		  get_the_group = conn.prepareStatement("SELECT * FROM Groups WHERE group_id = ?");
	  }
	  return true;
  }
	
}



