package JBMSTours;

import java.sql.*;
//import java.util.Date;
import java.math.BigDecimal;
import JBMSTours.serializabletypes.Tour;

/**
 * Look at the <a href="../JBMSTours/Hotel.java">Source</a>.<p>
 * A class to represent hotels.
 * Demonstrates: how to translate an object into flat relational data.
 
 *
 * @version   "%I%,%G%"
 *
 */
public class Hotel 
{
    /**
     *  Stored in the hotel_id column. 
     **/
    public int hotelId;

    /**
     * Name of the hotel. Stored in the hotel_name column.
     **/
    public String name;

    /**
     *  city_id of the hotel's city. Stored in the city_id column.
     **/
    private int cityId=0;

    /**
     *  A short that represents whether the hotel is economy, standard, or first class.
     *  Stored in the tour_level column.
     **/
    public short level;

 
    /**
     * The hotel's rate normally. Stored in the normal_rate column.
     **/
    public BigDecimal normalRate;

    /**
     * The hotel's rate during its busy season. Stored in the high_season_rate column
     **/
    public BigDecimal highSeasonRate;

    /**
     * Number of rooms this Hotel guarantees to our tour company. Stored in the
     * number_rooms_in_block column.
     */
    private int numberRoomsInTourBlock;


    /**
     * The date the hotel's high season begins. Stored in the high_season_begin column
     **/
    public short beginHighSeason;

    /**
     * The date the hotel's high season ends. Stored in the high_season_end column
     **/
    public short endHighSeason;

  
    /**
     * A short used if the hotel has no high season.
     **/

    public static final short NOHIGHSEASON = -1;


    private PreparedStatement find_number_of_rooms_taken;


     /**
     * public no-arg constructor for a hotel.
     *
     */
    public Hotel()
    {
        hotelId = 0;
        name = "unknown";
        cityId = 0;
        level = 0;
	normalRate = new BigDecimal(0);
	highSeasonRate = new BigDecimal(0);
        numberRoomsInTourBlock = 0;
	this.setHighSeason(NOHIGHSEASON, NOHIGHSEASON);
        find_number_of_rooms_taken = null;
    }

    /**
     * Constructs a hotel object for a hotel with no special high-season rate. 
     * Hotels are not stored in the database
     * as objects. They are stored as standard SQL-92 data.
     *
     * @param ID         hotel_id
     * @param hotelName  name of hotel
     * @param theCity    name of city
     * @param     The level of the tour. See the static variables in Tour.java.
     * @param normal     the rate for staying at the hotel in the normal season
     * @param high       the rate for staying in the hotel in the high season
     * @param rooms      the number of rooms in the hotel reserved for
     *                   our tour company
     */
    public Hotel(int ID, String hotelName, int theCity, short level, BigDecimal normal, BigDecimal high, int rooms)
    {
        this();
        hotelId = ID;
        name = hotelName;
        cityId = theCity;
        this.level = level;
        normalRate = normal;
        highSeasonRate = high;
        numberRoomsInTourBlock = rooms;
    }


    /**
     * Constructs a hotel object for a hotel with a special high-season rate. 
     * Hotels are not stored in the database
     * as objects. They are stored as standard SQL-92 data.
     *
     * @param ID         hotel_id
     * @param hotelName  name of hotel
     * @param theCity    name of city
     * @param     The level of the tour. See the static variables in Tour.java.
     * @param normal     the rate for staying at the hotel in the normal season
     * @param high       the rate for staying in the hotel in the high season
     * @param rooms      the number of rooms in the hotel reserved for
     *                   our tour company
     * @param begin_high_season  the month the high season begins
     * @param end_high_season    the month the high season ends
     */

   public Hotel(int ID, String hotelName, int theCity, short level, BigDecimal normal,
				BigDecimal high, int rooms, short begin_high_season, short end_high_season)
    {
        this(ID, hotelName, theCity, level, normal, high, rooms);
        this.setHighSeason(begin_high_season, end_high_season);
    }


    /**
     * Constructs a hotel object for a hotel that has a high-season rate using 
     * objects for parameters to make it easier for Cloudscape to recognize the
     * signature. <p>
     * Hotels are not stored in the database
     * as objects. They are stored as standard SQL-92 data.
     * <p>However, using this signature, you can construct Hotel objects within
     * SQL-J. For example:
     * <p><code>SELECT NEW Hotel(hotel_id, hotel_name, city_id, tour_level,
     * normal_rate, high_season_rate, number_rooms_in_block, high_season_begin,
     * high_season_end) FROM Hotels</code>
     *
     * @param ID         hotel_id
     * @param hotelName  name of hotel
     * @param theCity    name of city
     * @param level      levelof the tour
     * @param normal     the rate for staying at the hotel in the normal season
     * @param high       the rate for staying in the hotel in the high season
     * @param rooms      the number of rooms in the hotel reserved for
     *                   our tour company
     * @param begin_high_season  the month the high season begins
     * @param end_high_season    the month the high season ends
     */

   public Hotel(Integer ID, String hotelName, Integer theCity, Short level, 
                BigDecimal normal, BigDecimal high, Integer rooms, 
                Short begin_high_season, Short end_high_season)
    {
        this(ID.intValue(), hotelName, theCity.intValue(), 
             level.shortValue(), normal, high, rooms.intValue());
        this.setHighSeason(begin_high_season.shortValue(), end_high_season.shortValue());
    }


    /**
     * Constructs a Hotel given a row (ResultSet) in the Hotels table.
     *
     * @exception java.sql.SQLException if a database error occurs
     */

  public Hotel(ResultSet rs) throws SQLException {
                this(rs.getInt(1), rs.getString(2), rs.getInt(3), rs.getShort(4), 
                rs.getBigDecimal(5, 2), rs.getBigDecimal(6, 2), rs.getInt(7), 
                rs.getShort(8), rs.getShort(9)); 
  }


    /**
     * Create a string that describes this hotel.
     */
    public String toString()
    {
        return getName();
    }

    /**
     * Get the name of this hotel.
     */
    public String getName()
    {
        return name;
    }

    /**
     * Get the cityId for this hotel.
     */
    public int getCityId()
    {
            return cityId;
    }

    /**
     * Get the level of this hotel.
     */
    public short getLevel()
    {
        return level;
    }


    /**
     * Get the number of rooms in a tour block for
     * this hotel.
     */
    public int getNumberRoomsInTourBlock()
    {
        return numberRoomsInTourBlock;
    }


    /**
     * Define the months in the 'high season' for the hotel.
     *
     * <p> To use this provide the start (first month) and stop (last month)
     * for the high season. Months are zero-based so January is month 0.
     * You can use the static variables of java.lang.Calendar.
     * <p> To define a sequence of months that begins at the end of the year
     * and wraps around into the beginning of the next year the provide a first
     * month that is greater than the last month. For example to define a 'high
     * season' from December to January inclusive use 
     * 'setHighSeason(java.lang.Calendar.DECEMBER, java.lang.Calendar.JANUARY)'.
     *
     * @param start      the month the high season begins (0 based)
     * @param stop       the month the high season ends (0 based)
     *
     */
    public void setHighSeason(short start, short stop)
    {
      beginHighSeason = (short) start;
      endHighSeason = (short) stop; 

    }


    /**
     * Returns a "true" any of part of the stay occurs during high season.
     *
     * @param start     the start date of the stay
     * @param end       the end date of the stay
     */
    public boolean isHigh(Date start, Date end)
    {
        if (!isAHighSeason())
	    return false;
	    
	return overlapsOnClock(JCalendar.getMonth(start), 
                               JCalendar.getMonth(end), 
                               (int)beginHighSeason, 
                               (int) endHighSeason);

    }


  /* private method for seeing if two ranges of months overlap at all*/
  private static boolean overlapsOnClock(int a1, int a2, int b1, int b2) {

    boolean aflag = false;
    boolean bflag = false;
    for (int i = 0; i < 12; i++) {
      
      if (!aflag) {
 
      int t = i + a1;
      if (t > 11)
	t = t - 12;
      if (t == a2)
	aflag = true;
      bflag = false;
      for (int j = 0; j < 12; j++) {
	if (!bflag) {
	int s = j + b1;
	if (s > 12)
	  s = s - 12;
	if (s == b2)
	  bflag=true;
	if (t == s)
	  return true;
	}
      }
      }
    }

    return false;

  }


    /**
     * Returns a "true" if this hotel has a high season in which it raises its rates.
     *
     * @return  boolean    if there is a high season
     */
    public boolean isAHighSeason()
    {
      return beginHighSeason > NOHIGHSEASON;
    }

    /**
     * Sets rates for normal and high season, in dollars. Does not update rates in database,
     * if data has already been stored.
     *
     * @param normal     rate for normal season
     * @param high       rate for high season.
     */
      public void setRates(BigDecimal normal, BigDecimal high)
    {
        normalRate = normal;
        highSeasonRate = high;
    }

    /**
     * Updates the rates in the database.
     *
     * @param conn   a connection to the database
     *
     * @exception java.sql.SQLException if a database error occurs
     */
      public int updateRatesInDatabase(Connection conn) throws java.sql.SQLException
    {

      PreparedStatement ps = conn.prepareStatement("UPDATE Hotels SET normal_rate = ?, " +
          "high_season_rate = ? WHERE hotel_id = ?");

      ps.setBigDecimal(1, normalRate);
      ps.setBigDecimal(2, highSeasonRate);
      ps.setInt(3, hotelId);
      int j = ps.executeUpdate();
      ps.close();
      return j;
    }



    /**
     * Get the rate for the normal season
     */
    public BigDecimal getNormalRate()
    {
        return normalRate;
    }

    /**
     * Get the rate for the high season
     */
    public BigDecimal getHighRate()
    {
        return highSeasonRate;
    }

    /**
     * Get today's rate.
     */
    public BigDecimal getTodayRate()
    {

		Date now = JCalendar.getToday();
        if(this.isHigh(now,now))
            return getHighRate();
        else
            return getNormalRate();
    }

    /**
     * Return true if the hotel has a given number of rooms
     * available for a stay.
     *
     *
     * @param conn   a connection to the database
     * @param begin  beginning date for the proposed stay.
     * @param end    ending date for the proposed stay.
     * @param rooms  number of rooms needed
     *
     * @return  true if hotel has the requested number of
     *          rooms available for the proposed stay.
     * @exception java.sql.SQLException a error occurred.
     */
    public boolean isAvailable(Connection conn, java.sql.Date begin, java.sql.Date end, int rooms) throws SQLException
    {
        boolean avail = true;
        prepareStatements(conn);
        find_number_of_rooms_taken.setDate(1,begin);
        find_number_of_rooms_taken.setDate(2,end);
        find_number_of_rooms_taken.setInt(3,hotelId);
        ResultSet rs = find_number_of_rooms_taken.executeQuery();
        while(rs.next())
        {
            int p = 0;
            p = rs.getInt(1);
            if((numberRoomsInTourBlock - p) < rooms)
                avail = false;
        }
        //close the result set
        rs.close();
        return avail;
    }

    /**
     * Return a message describing this hotel.
     * @return  String   object information about the Hotel
     */
    public String getInfo()
    {
        StringBuffer hotelInfo = new StringBuffer();
        hotelInfo.append("\nYou're staying at the: " + name);
        hotelInfo.append("\nIt is a " + Tour.getStringForLevel(level) + " hotel");
        return hotelInfo.toString();
    }

    /**
     * Prepares some or all of the PreparedStatements for this class.
     *
     * @param conn    a connection to the database
     *
     * @exception  SQLException if there is an SQL error
     */
    private boolean prepareStatements(Connection conn) throws SQLException
    {
      /* this statement is a prepared statement, because conceivably the booking date
         values could change. The hotel_id will never change.*/

        if(find_number_of_rooms_taken == null)
            find_number_of_rooms_taken = conn.prepareStatement(
                "SELECT rooms_taken FROM HotelAvailability  " + 
                "WHERE booking_date BETWEEN ? AND ? AND hotel_id = ?");
        if(find_number_of_rooms_taken != null)
            return true;
        else
            return false;
    }




    /**
     * Inserts the hotel into the database, "translating" it into flat relational
     * data.
     *
     * @param conn    a connection to the database
     *
     * @exception  SQLException if there is an SQL error
     */
  public int insertInDatabase(Connection conn) throws SQLException {
    
  String insertString="INSERT INTO hotels (hotel_id, hotel_name, city_id, tour_level, normal_rate, "+
      "high_season_rate, number_rooms_in_block, high_season_begin, high_season_end) "+
      "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)";

    PreparedStatement insert1 = conn.prepareStatement(insertString);
            insert1.setInt(1, hotelId);
        insert1.setString(2, name);
        insert1.setInt(3, cityId);
        insert1.setShort(4, level);
        insert1.setObject(5, normalRate);
        insert1.setObject(6, highSeasonRate);
        insert1.setInt(7, numberRoomsInTourBlock);
        insert1.setShort(8, beginHighSeason);
        insert1.setShort(9, endHighSeason);

        int j = insert1.executeUpdate(); 

       insert1.close();
       return j;

  }

    /**
     * Updates a this hotel's record in the database with the data contained a new Hotel object;
	 * updates only changed fields.
     *
     * @param conn    a connection to the database
	 * @param Hotel   the new Hotel object
     *
     * @exception  SQLException if there is an SQL error
     */
 public int updateInDatabase(Connection conn, Hotel newhotel) throws SQLException {
	
	 StringBuffer updateString = new StringBuffer("UPDATE Hotels SET  ");
	 
	 boolean hotelChanged = false;
	 boolean nameChanged = false;
	 boolean levelChanged = false;
	 boolean numberRoomsChanged = false;
	 boolean normalRateChanged = false;
	 boolean highSeasonRateChanged = false;
	 boolean beginHighSeasonChanged = false;
	 boolean endHighSeasonChanged = false;
	 
	 if (!this.name.equals(newhotel.name)) {
		 hotelChanged = true; 
		 nameChanged = true;
		 updateString.append("hotel_name = " + newhotel.name);    
	 }
	 
	 if (this.level != newhotel.level) {
		 
		 levelChanged = true;
		 updateString.append(buildStmt(hotelChanged) + "tour_level = " + newhotel.level);
		 hotelChanged = true;
	 }
	 
	 if (!this.normalRate.equals(newhotel.normalRate)) {
		 normalRateChanged = true;
		 updateString.append(buildStmt(hotelChanged) + "normal_rate = " +
							 newhotel.normalRate.toString()); 
		 hotelChanged = true;
	 }
	 
	 if (!this.highSeasonRate.equals(newhotel.highSeasonRate)) {
		 highSeasonRateChanged = true;
		 updateString.append(buildStmt(hotelChanged) + "high_season_rate = " +
							 newhotel.highSeasonRate.toString());
		 hotelChanged = true;	
	 }
	 
	 if (this.getNumberRoomsInTourBlock() != newhotel.getNumberRoomsInTourBlock() ) {
		 numberRoomsChanged = true;
		 updateString.append(buildStmt(hotelChanged) + "number_rooms_in_block = " +
							 newhotel.getNumberRoomsInTourBlock());
		 hotelChanged = true;
	 }
	 
	 
	 
	 if (this.beginHighSeason != newhotel.beginHighSeason) {
		 beginHighSeasonChanged = true;
		 updateString.append(buildStmt(hotelChanged) + "high_season_begin=" +
							 newhotel.beginHighSeason);
		 hotelChanged = true;      
	 }
	 
	 
	 if (this.endHighSeason != newhotel.beginHighSeason) {
		 endHighSeasonChanged = true;
		 updateString.append(buildStmt(hotelChanged) + "high_season_end = " +
							 newhotel.endHighSeason);
		 hotelChanged = true;      
		 
	 }
	 
	 if (!hotelChanged)
		 return 0;
	 
	 updateString.append(" WHERE hotel_id = " + hotelId);
	 System.out.println(updateString.toString());
	 Statement st = conn.createStatement();
	 int j = st.executeUpdate(updateString.toString());
	 st.close();
	 return j;
 }
	
	private static String buildStmt(boolean notFirst) {
		if (notFirst)
			return ", ";
		else
			return "";
	}
	
}










