package JBMSTours;

import java.sql.*;

/**
 * Look at the <a href="../JBMSTours/RunTime.java">Source</a>.<p>
 * This class executes some statements to show runtime statistics.
 * A variety of selects are used, including basic selects, joins, 
 * and order bys.
 * <p>The class inserts results  
 * into a table called <i>runtime</i> that you can examine later at your convenience.
 * It also outputs information. For some queries, only getScanStatistics() text is called. For
 * Other queries, toString() is called, and the entire statement execution plan is displayed</p>
 *
 * <p>This class has a main() method and can be run from the command-line.
 * Set the property JBMSTours.runtimeVerbose to false when running this program
 * in the tests.
 * For complete instructions on how to run it, see, <a href="../example.html#running">example instructions</a>.
 *
 */
public class RunTime {

    /**
    For statements executed against the database.
     */
    private transient Statement s;
    /**
    Holds the repeated insert into the runtime table
     */
    private transient PreparedStatement psInsert;
    /**
    Holds the select of full runtime text the runtime table
     */
    private transient PreparedStatement psSelect;
    /**
    Holds the select of scan text only from the runtime table
     */
    private transient PreparedStatement psSelectPartial;

    /**
      A ResultSet for various uses.
     */
    private transient ResultSet rs;

    /**
    To manage server and embedded connection use.
     */
    public ApplicationMode mymode;
    /**
    To manage inserts into the runtime table.
     */
    private int runtimeId=0;
    private long instanceId = new java.util.Date().getTime();


    public static void main(String[] args) {

    /* testing for the client mode information (hostname and portnum)*/
      ApplicationMode runmode = new ApplicationMode(args);
        (new RunTime()).go(runmode);
    }

    void go(ApplicationMode amode) {
        System.out.println("RunTime starting");
        mymode = amode;

        try {

            Connection conn = mymode.connectToCloudscape();


            /*
                Creating a statement lets us issue commands against
                the connection.
             */

            s = conn.createStatement();
            System.out.println("created statement");
            conn.setAutoCommit(true);

            System.out.println("setting up statistics table");
            try {
                s.execute("CREATE TABLE runtime("+
                  "instant LONGINT, " +
                  "id INT , " +
                  "desired SERIALIZE(java.lang.String) NOT NULL, "+
                  "result SERIALIZE(COM.cloudscape.types.RunTimeStatistics), " +
                  "CONSTRAINT runtime_pk PRIMARY KEY(instant, id))");

            } catch (SQLException se) {
                // table must have existed
                // ignores catastrophic problems like log full...

            }
            psInsert = conn.prepareStatement("INSERT INTO runtime VALUES (?, ?, ?, RunTimeStatistics())");
            psSelect = conn.prepareStatement("SELECT result.toString() FROM runtime WHERE instant = ? AND id = ?");
            psSelectPartial = conn.prepareStatement("SELECT result.getScanStatisticsText() FROM runtime WHERE instant = ? AND id = ?");
            //conn.commit(); // done setting up runtime and the prepares 

            /*
               Turn on statistics gathering mode
             */
            s.execute("SET RUNTIMESTATISTICS ON");
            s.execute("SET STATISTICS TIMING ON");


            String[] select = new String[2];
            select[0] ="SELECT miles FROM Flights";
            select[1] = "Should show a table scan of Flights.";
            processQuery(select, false);

            select[0] = "SELECT Country FROM Countries " +
                                        "where region = 'Europe'";

            select[1] = "Should show a table scan of Hotels, because the Region column is not indexed.";
            processQuery(select, false);

            select[0] = "SELECT * from Hotels where hotel_id < 50";
            select[1] = "Should show an index scan of Hotels using the index on hotel_id; then another node where the corresponding rows are retrieved.";
            processQuery(select, false);

            select[0] = "SELECT country, city.getName() FROM countries, cities WHERE countries.country_ISO_code = cities.country_ISO_code AND cities.country_ISO_code BETWEEN 'B' AND 'E'";
            select[1] = "This query and the one right after this one show that Cloudscape chooses an optimal order for the join, no matter how the query is written.";
            processQuery(select, true);

            select[0] = "SELECT country, city.getName() FROM cities, countries WHERE countries.country_ISO_code = cities.country_ISO_code AND countries.country_ISO_code BETWEEN 'B' AND 'E'";            
            select[1] = "See the notes for the query before this one.";
            processQuery(select, true);


            select[0] = "select orig_airport, dest_airport, flight_date from Flights, FlightAvailability WHERE flights.flight_id = FlightAvailability.flight_id AND Flights.segment_number = FlightAvailability.segment_number";
            select[1] = "This query should do a hash join.";
            processQuery(select, false);


            select[0] = "SELECT orig_airport FROM Flights";
            select[1] = "Should show how Cloudscape scans index only, never table. Index covers query.";
            processQuery(select, false);

            select[0] = "SELECT city.getName(), city.showTemperature() as TEMPO from cities ORDER BY TEMPO";
            select[1] = "Should show a node where sorting is done.";
            /* use the true argument, because we want to see the full runtime statistics text in order to see the sort node.*/
            processQuery(select, true);

            String createtemptable = "CREATE TABLE mytemp (flight_id CHAR(6), airport CHAR(3))";

            conn.setAutoCommit(false);
            s.execute(createtemptable);
            s.execute("LOCK TABLE mytemp in EXCLUSIVE MODE");

           select[0] = "INSERT INTO mytemp SELECT flight_id, dest_airport from flights where orig_airport = 'SFO'";
           select[1] = "should show nodes for both selecting and inserting.";
           processStatement(select, true);
            s.execute("DROP TABLE mytemp");
	    conn.commit();
	   conn.setAutoCommit(true);

           select[0] = "SELECT count(*) from hotels";
           select[1] = "should use an index and avoid a table scan. Should show a node for sorting.";
           processQuery(select, true);


           select[0] = "EXECUTE STATEMENT GETFULLTRANSFERFLIGHTINONEBLOW USING VALUES ('SFO', 'BUE')";
           select[1] = "Should show information about the stored prepared statement.";
           processQuery(select, true);

            /*
               Turn off statistics gathering mode
             */
            s.execute("SET RUNTIMESTATISTICS OFF");
            s.execute("SET STATISTICS TIMING OFF");

            System.out.println("turned off runtimestatistics");

            /*
               free up our JDBC resources.
             */
            s.close();
            psInsert.close();
            psSelect.close();
			conn.close();
            mymode.quitCloudscape();

        } catch (Throwable e) {
            System.out.println("exception thrown:");
            JBMSTours.CreateToursDB.errorPrint(e);
        }

         System.out.println("RunTime finished");
    }


    /**
	 * Given a query and description of its desired result,
	 * execute it and store away the desired result and
	 * actual runtime statistics. Display the results as well.
	 * 
	 * @exception SQLException thrown on dbaccess error.
     */
    private void processQuery(String[] query, boolean fullText) throws SQLException {

            System.out.println(query[1]);
            rs = s.executeQuery(query[0]);
            System.out.println("executed query.");
            while (rs.next()) {
                // let it gather data on the fetches.
            }
            rs.close();
			processStatistics(fullText,query[1]);
			
	}

    /**
	 * Given a statement and description of its desired result,
	 * execute it and store away the desired result and
	 * actual runtime statistics.
	 * @exception SQLException thrown on dbaccess error.
     */
    private void processStatement(String[] query, boolean fullText) throws SQLException {

            System.out.println(query[1]);
            s.executeUpdate(query[0]);
            System.out.println("executed statement.");
			processStatistics(fullText,query[1]);
	}

	private void processStatistics(boolean fullText, String querytext) throws SQLException {
           
		psInsert.setLong(1, instanceId);
		psInsert.setInt(2, runtimeId);
		psInsert.setString(3,querytext);
	    psInsert.execute();
             
		PreparedStatement ps = psSelect; 
           if (fullText == false)
	     ps = psSelectPartial;

		   ps.setLong(1, instanceId);
            ps.setInt(2, runtimeId++);
            rs = ps.executeQuery();
            while (rs.next()) {
				if (getVerboseMode())
                System.out.println(rs.getString(1));
            }
            rs.close();
			
    }

    /**
     * Returns the verbose boolean
	 *
	 * @return boolean 
     *
     */
	private static boolean getVerboseMode() {
		String verbose = System.getProperty("JBMSTours.runtimeVerbose", "true");
		return new Boolean(verbose).booleanValue();
	}
	
	

}







