package JBMSTours.aggregates;
import COM.cloudscape.aggregates.Aggregator;
import COM.cloudscape.aggregates.AggregateDefinition;
import COM.cloudscape.types.TypeDescriptor;
import java.io.Externalizable;
import java.io.ObjectOutput;
import java.io.ObjectInput;
import java.io.IOException;
import java.sql.ResultSet;
import java.sql.Types;


/**
 * This class implements the aggregation for MaxButONE when the input type is
 * a double.
 * <p>
 * We implement Externalizable to save space in case we
 * are written out (Serialization tends to take up a lot
 * of space).
 */
public class MaxButOneDoubleAggregator implements Externalizable, Aggregator
{

	private Double topValue;
	private Double secondValue;
	
	/**
	 * Niladic constructor
	 */	
	public MaxButOneDoubleAggregator()
	{
	}

	/////////////////////////////////////////////////////////////////
	//
	// AGGREGATOR
	//
	/////////////////////////////////////////////////////////////////

	/**
	 * Initializes the aggregate for processing.  
	 */
	public	void initialize()
	{
	}

	/**
	 * Iteratively accumulates the addend into the aggregator.
	 * Called on each member of the set of values that is being
	 * aggregated.
	 *
	 * @param addend	the ResultSet addend (current input to 
	 * 					the aggregation)
	 *
	 * @exception Throwable on error
	 */
	public	void accumulate(ResultSet addend) throws Throwable
	{
		double currentValue;
		currentValue = addend.getDouble(1);
		if (addend.wasNull())
			return;
		accumulate(new Double(currentValue));
	}

	private void accumulate(Double currentValue) throws Throwable {
		//the first time this is called, topValue will be null
		if (topValue==null || 
			(currentValue.doubleValue() > topValue.doubleValue()))
		{
			secondValue=topValue;
			topValue=currentValue;
		}
		else if (secondValue==null || 
			(currentValue.doubleValue() > secondValue.doubleValue()))
		{
			secondValue=currentValue;
		}
	}

	/**
	 * Merges one aggregator into a another aggregator. Used in Group by.
	 *
	 * @param inputAggregator	the other Aggregator 
	 *							(input partial aggregate)
	 *
	 * @exception Throwable on error
	 */
	public	void merge(Aggregator inputAggregator) throws Throwable
	{
		MaxButOneDoubleAggregator otherGuy = (MaxButOneDoubleAggregator) inputAggregator;
		accumulate(otherGuy.getTopValue());
		accumulate(otherGuy.getSecondValue());
	}

	/**
	 * Produces the result to be returned by the query.
	 * The last processing of the aggregate.
	 * <P>In this instance, returns the second highest value in the column. <p>
	 * See <a href="../../../../javadoc/COM.cloudscape.aggregates.AggregateDefinition.html#getResult()">
	 * COM.cloudscape.aggregates.AggregateDefinition.getResult()</a>.
	 *
	 *
	 * @return the result column.  
	 *
	 */
	public Object getResult()
	{
		return secondValue;
	}

	/**
	 * Return a new instance of this class.  Identical
	 * to Class.newInstance() but much cheaper because
	 * reflection is costly.  
	 * <p>
	 * Just implement it to do <i>new myClass()</i>.
	 *
	 * @return Aggregator
	 */
	public Aggregator newInstance()
	{
		return new MaxButOneDoubleAggregator();
	}

	/////////////////////////////////////////////////////////////////
	//
	// CLASS INTERFACE
	//
	/////////////////////////////////////////////////////////////////

	/**
	 * Returns the highest value in the column found so far. This method
	 * needs to be public because instances call each other during a
	 * merge.
	 *
	 * @return topValue
	 */
	public Double getTopValue()
	{
		return topValue;
	}
	/**
	 * Returns the second highest value in the column found so far. This method
	 * needs to be public because instances call each other during a
	 * merge.
	 *
	 * @return secondValue
	 */
	public Double getSecondValue()
	{
		return secondValue;
	}

	
	/////////////////////////////////////////////////////////////////
	//
	// EXTERNALIZABLE 
	//
	/////////////////////////////////////////////////////////////////

	/**
	 * Write out our two fields.
	 *
	 * @param out the output stream
	 *
	 * @exception IOException on error
	 */
	public void writeExternal(ObjectOutput out) throws IOException
	{
		out.writeDouble(topValue.doubleValue());
		out.writeDouble(secondValue.doubleValue());
	}

	/**
	 * Read in our two fields.
	 *
	 * @param int the input stream
	 *
	 * @exception IOException on error
	 * @exception ClassNotFoundException on error
	 */
	public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException 
	{
		topValue = new Double(in.readDouble());
		secondValue = new Double(in.readDouble());
	}
}
