package JBMSTours.aggregates;
import COM.cloudscape.aggregates.Aggregator;
import COM.cloudscape.aggregates.AggregateDefinition;
import COM.cloudscape.types.TypeDescriptor;
import java.io.Externalizable;
import java.io.ObjectOutput;
import java.io.ObjectInput;
import java.io.IOException;
import java.sql.ResultSet;
import java.sql.Types;
import java.lang.Comparable;


/**
 * Able to Define and implement the aggregation for MaxButONE in a JDK12
 * environment. This aggregate returns the runnner up value for the column
 * (not the max, but the value right
 * after that). 
 * <p>
 * We implement Externalizable to save space in case we
 * are written out (Serialization tends to take up a lot
 * of space).
 */
public class MaxButOneJDK12 implements Externalizable, Aggregator, AggregateDefinition
{

	private Object topValue;
	private Object secondValue;
	
	/**
	 * Niladic constructor
	 */	
	public MaxButOneJDK12()
	{
	}

	/////////////////////////////////////////////////////////////////
	//
	// AGGREGATE DEFINITION
	//
	/////////////////////////////////////////////////////////////////
	
	/**
	 * @param inputType	the input type descriptor
	 * @param aggregatorClassName	output parameter, filled in
	 *		with the class name that implements COM.cloudscape.aggregates.Aggregator
	 *
	 * @return the output type descriptor (which may or may not
	 *		be the same as the input type -- it is ok to simply
	 *		return the input type).  Null is returned
	 *		if the aggregate cannot process the input type.
	 *
	 * @see COM.cloudscape.types.TypeDescriptor
	 */
	public	TypeDescriptor getAggregator(TypeDescriptor inputType,
							StringBuffer aggregatorClassName)
	{
		aggregatorClassName.append("JBMSTours.aggregates.MaxButOneJDK12");	
		return inputType;
	}	
	
	/////////////////////////////////////////////////////////////////
	//
	// AGGREGATOR
	//
	/////////////////////////////////////////////////////////////////

	/**
	 * Initializes the aggregate for processing.  
	 */
	public	void initialize()
	{
	}

	/**
	 * Iteratively accumulates the addend into the aggregator.
	 * Called on each member of the set of values that is being
	 * aggregated.
	 *
	 * @param addend	the ResultSet addend (current input to 
	 * 					the aggregation)
	 *
	 * @exception Throwable on error
	 */
	public	void accumulate(ResultSet addend) throws Throwable
	{
        Comparable currentValue = (Comparable) addend.getObject(1);
		if (addend.wasNull())
			return;

		accumulate(currentValue);
	}

	private void accumulate(Object cv) throws Throwable {
		//the first time this is called, topValue will be null
		Comparable currentValue = (Comparable) cv;
		if (topValue==null || currentValue.compareTo(topValue)==1)
		{
			secondValue=topValue;
			topValue=currentValue;
		}
		else if (secondValue==null || currentValue.compareTo(secondValue)==1)
		{
			secondValue=currentValue;
		}
	}

	/**
	 * Merges one aggregator into a another aggregator. Used in Group by.
	 *
	 * @param inputAggregator	the other Aggregator 
	 *							(input partial aggregate)
	 *
	 * @exception Throwable on error
	 */
	public	void merge(Aggregator inputAggregator) throws Throwable
	{
		MaxButOneJDK12 otherGuy = (MaxButOneJDK12) inputAggregator;
		accumulate(otherGuy.getTopValue());
		accumulate(otherGuy.getSecondValue());
	}

	/**
	 * Produces the result to be returned by the query.
	 * The last processing of the aggregate.
	 * <P>In this instance, returns the second highest value in the column.<p>
	 * See <a href="../../../../javadoc/COM.cloudscape.aggregates.AggregateDefinition.html#getResult()">
	 * COM.cloudscape.aggregates.AggregateDefinition.getResult()</a>.
	 *
	 *
	 * @return the result column.  
	 *
	 */
	public Object getResult()
	{
		return secondValue;
	}

	/**
	 * Return a new instance of this class.  Identical
	 * to Class.newInstance() but much cheaper because
	 * reflection is costly.  
	 * <p>
	 * Just implement it to do <i>new myClass()</i>.
	 *
	 * @return Aggregator
	 */
	public Aggregator newInstance()
	{
		return new MaxButOneJDK12();
	}

	/////////////////////////////////////////////////////////////////
	//
	// CLASS INTERFACE
	//
	/////////////////////////////////////////////////////////////////

	/**
	 * Returns the highest value in the column found so far. This method
	 * needs to be public because instances call each other during a
	 * merge.
	 *
	 * @return topValue
	 */
	public Object getTopValue()
	{
		return topValue;
	}
	/**
	 * Returns the second highest value in the column found so far. This method
	 * needs to be public because instances call each other during a
	 * merge.
	 *
	 * @return secondValue
	 */
	public Object getSecondValue()
	{
		return secondValue;
	}

	
	/////////////////////////////////////////////////////////////////
	//
	// EXTERNALIZABLE 
	//
	/////////////////////////////////////////////////////////////////

	/**
	 * Write out our two fields.
	 *
	 * @param out the output stream
	 *
	 * @exception IOException on error
	 */
	public void writeExternal(ObjectOutput out) throws IOException
	{
		out.writeObject(topValue);
		out.writeObject(secondValue);
	}

	/**
	 * Read in our two fields.
	 *
	 * @param int the input stream
	 *
	 * @exception IOException on error
	 * @exception ClassNotFoundException on error
	 */
	public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException 
	{
		topValue = in.readObject();
		secondValue = in.readObject();
	}
}
