package JBMSTours.aggregates;

import VisualNumerics.math.Statistics;
import COM.cloudscape.aggregates.Aggregator;
import COM.cloudscape.aggregates.AggregateDefinition;
import COM.cloudscape.database.Factory;
import COM.cloudscape.aggregates.AggregateDefinition;
import COM.cloudscape.types.TypeDescriptor;
import java.util.Vector;
import java.util.Enumeration;
import java.io.Externalizable;
import java.io.ObjectOutput;
import java.io.ObjectInput;
import java.io.IOException;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.Types;
import java.sql.SQLException;

/**
 * This is a poor man's standard deviation.  It only works on 
 * input values of type DOUBLE PRECISION.  It is implemented
 * using VisualNumerics statistics package.  It is pretty
 * primitive in how it works -- i just builds up a big
 * array of doubles and feeds that to visual numerics.
 * <p>
 * We implement Externalizable to save space in case we
 * are written out (Serialization tends to take up a lot
 * of space).
 */
public class StandardDeviation implements Externalizable, Aggregator, AggregateDefinition
{
	private Vector doubleVector;


	/**
	 * Niladic constructor
	 */	
	public StandardDeviation()
	{
	}

	/////////////////////////////////////////////////////////////////
	//
	// AGGREGATE DEFINITION
	//
	/////////////////////////////////////////////////////////////////
	
	/**
	 * @param inputType	the input type descriptor
	 * @param aggregatorClassName	output parameter, filled in
	 *		with the class name that implements COM.cloudscape.aggregates.Aggregator
	 *
	 * @return the output type descriptor (which may or may not
	 *		be the same as the input type -- it is ok to simply
	 *		return the input type).  Null is returned
	 *		if the aggregate cannot process the input type.
	 *
	 * @exception SQLException if a problem finding type descriptor
	 *
	 * @see COM.cloudscape.types.TypeDescriptor
	 */
	public	TypeDescriptor getAggregator(TypeDescriptor inputType,
							StringBuffer aggregatorClassName)
							throws SQLException
	{
		aggregatorClassName.append("JBMSTours.aggregates.StandardDeviation");	
		
		// only take DOUBLE PRECISION
		if (inputType.getJDBCTypeId() != Types.DOUBLE)
		{
			return null;
		}

		return Factory.getTypeFactory().getTypeDescriptor("java.lang.Double", true);
	}	

	/////////////////////////////////////////////////////////////////
	//
	// AGGREGATOR
	//
	/////////////////////////////////////////////////////////////////

	/**
	 * Initializes the aggregate for processing.  
	 */
	public	void initialize()
	{
		doubleVector = new Vector();
	}

	/**
	 * Iteratively accumulates the addend into the aggregator.
	 * Called on each member of the set of values that is being
	 * aggregated.
	 *
	 * @param addend	the ResultSet addend (current input to 
	 * 					the aggregation)
	 *
	 * @exception Throwable on error
	 */
	public	void accumulate(ResultSet addend) throws Throwable
	{
		if (addend != null)
		{
			double d = addend.getDouble(1);
			if (!addend.wasNull())
			{
				doubleVector.addElement(new Double(d));
			}
		}
	}

	/**
	 * Merges one aggregator into a another aggregator.
	 *
	 * @param inputAggregator	the other Aggregator 
	 *							(input partial aggregate)
	 *
	 * @exception Throwable on error
	 */
	public	void merge(Aggregator inputAggregator) throws Throwable
	{
		Vector otherVector = ((StandardDeviation)inputAggregator).getVector();

		for (Enumeration e = otherVector.elements(); e.hasMoreElements();)
		{
			doubleVector.addElement(e.nextElement());
		}
	}

	/**
	 * Produces the result to be returned by the query.
	 * The last processing of the aggregate.  Feed
	 * the array of doubles into VisualNumerics. <p>
	 * See <a href="../../../../javadoc/COM.cloudscape.aggregates.AggregateDefinition.html#getResult()">
	 * COM.cloudscape.aggregates.AggregateDefinition.getResult()</a>.
	 *
	 * @return the result column.  
	 *
	 */
	public Object getResult()
	{
		double[] doubleArray = new double[doubleVector.size()];
		int i = 0;
		for (Enumeration e = doubleVector.elements(); e.hasMoreElements(); i++)
		{
			doubleArray[i] = ((Double)e.nextElement()).doubleValue();
		}
		return new Double(Statistics.standardDeviation(doubleArray));
	}

	/**
	 * Return a new instance of this class.  Identical
	 * to Class.newInstance() but much cheaper because
	 * reflection is costly.  
	 * <p>
	 * Just implement it to do <i>new myClass()</i>.
	 *
	 * @return Aggregator
	 */
	public Aggregator newInstance()
	{
		return new StandardDeviation();
	}

	/////////////////////////////////////////////////////////////////
	//
	// CLASS INTERFACE
	//
	/////////////////////////////////////////////////////////////////
	public Vector getVector()
	{
		return doubleVector;
	}

	
	/////////////////////////////////////////////////////////////////
	//
	// EXTERNALIZABLE 
	//
	/////////////////////////////////////////////////////////////////

	/**
	 * Write out our double vector as an array of doubles,
	 * with an integer size.
	 *
	 * @param out the output stream
	 *
	 * @exception IOException on error
	 */
	public void writeExternal(ObjectOutput out) throws IOException
	{
		out.writeInt(doubleVector.size());
		for (Enumeration e = doubleVector.elements(); e.hasMoreElements();) 
		{
			out.writeDouble(((Double)e.nextElement()).doubleValue());
		}
	}

	/**
	 * Read in out doubles.  First read in the size and then
	 * reinstantiate our vector and fill it up.
	 *
	 * @param int the input stream
	 *
	 * @exception IOException on error
	 */
	public void readExternal(ObjectInput in) throws IOException
	{
		int size = in.readInt();
		doubleVector = new Vector(size);
		while (size-- > 0)
		{
			doubleVector.addElement(new Double(in.readDouble()));
		}
	}
}
