package JBMSTours.serializabletypes;

import java.util.Locale;
import java.sql.*;
//import java.lang.Comparable;

import java.io.Externalizable;
import java.io.ObjectOutput;
import java.io.ObjectInput;
import java.io.IOException;
import JBMSTours.Util;
import JBMSTours.JCalendar;
import JBMSTours.WeatherSpot;



/**
 * Look at the <a href="../JBMSTours/serializabletypes/City.java">Source</a>.<p>
 * The City class includes information about the city,
 * the country, the language, average monthly temperature, airport 
 * and alternate airport, Location, and Time Zone. It includes 
 * methods for calculating the distance between
 * cities and time differences between cities.<p>
 *
 * This class is the richest data type in JBMSTours. <p>
 *
 * This class is a good example of
 * how to use Externalizable instead of Serializable
 * for storage and retrieval efficiency.<p>
 *
 * Beginning with the 3.0 release, this class is ORDERABLE (because it
 * supplies the compareTo method). It also supplies the equals and hashCode
 * methods for correct = and <> comparisons.
 *
 */
public class City implements Externalizable, WeatherSpot {

    //////////////////
    // fields
    //////////////////

    /**
     * identifier for the city
     */
    public int city_id;

    /**
     * name of city
     */
    public String name;

    /**
     * the predominant language spoken in the city
     */
    public String language;

    /**
     * the country the city is located in
     */
    public String country;

    /**
     * the long name of the country, initialized to an empty string
     */
    public String countryLong = "";

    /**
     * preferred airport in the city
     */
    private String airport;


    /**
     * array to store the average monthly temperature
     */
    public double averageTemperature[] = 
                            {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    
    private static final String FARENHEIT = "Farenheit";
    private static final String CELSIUS = "Celsius";
    private static final String MILES = "miles";
    private static final String KILOMETERS = "kilometers";

    /**
     * city's location
     */
    private Location cityLocation;

    /**
     * city's timezone
     */
    private TimeZone cityTimeZone;
  
    /**
     * This is the version number for the first version the City class.
     * When we save an object of this class we begin by writing this version
     * number (see writeExternal). When we read the saved object we begin by
     * reading the version number (see readExternal).  In the future, if a
     * later version of this class saves different information (such as an
     * extra field), the reading logic can use the version number to determine
     * what has been saved.
     *
     */

    private static final int currentClassVersion = 1;
  
    private static transient PreparedStatement find_city_by_id;
    private static transient PreparedStatement find_city_by_name_and_country;
    private static transient PreparedStatement find_city_by_airport;


    /**
     * Cloudscape uses java serialization to write user objects
     * to the database and read them from the database.
     * To support future versions of our class we assign
     * an explicit value to 'serialVersionUID.
     *
     * <p>For more information java serialization please
     * refer to the <a href="../../../doc/main/dgolt.htm">Developer's Guide</a>
     * and to the JDK documentation.
     **/

    static final long serialVersionUID = -2869026592638800364L;

    /**
     * default system: metric or non-metric
     */
    public String measurementSystem = "nonmetric";

    /**
     * No-arg constructor.  Default city is San Francisco.
     */
    public City () {
      this(84, "San Francisco", "English", "US", "SFO");
      find_city_by_id = null;
      find_city_by_name_and_country=null;
      find_city_by_airport = null;
     
    }

    /**
     * Constructs a City object
     *
     * @param id          the city id
     * @param theName     the Name of the City
     * @param theLanguage lower-case two character code for language
     * @param countryCode upper-case two-character ISO code
     * @param ap          three-letter code for the airport   
     */ 
    public City (int id, String theName, String theLanguage, 
                 String countryCode, String ap) {
        city_id = id;
        name = theName;
        language = theLanguage;
        country = countryCode;
        airport = ap;

        // US uses metric system, all others metric
        if ( ! country.equals("US"))
            measurementSystem = "metric";

    }


    /**
     * Get a String representation of this object
     *
     * @return     String name of city
     */
    public String toString() {
        return name + ", " + countryLong;
    }


    /**
     *  Sets the array of average monthly temperatures (Farenheit) for the city. 
     *
     * @param Jan        average monthly temperature for January.
     * @param Feb        average monthly temperature for February.
     * @param Nar        average monthly temperature for March.
     * @param Apr        average monthly temperature for April.
     * @param May        average monthly temperature for May.
     * @param Jun        average monthly temperature for June.
     * @param Jul        average monthly temperature for July.
     * @param Aug        average monthly temperature for August.
     * @param Sep        average monthly temperature for Sept.
     * @param Oct        average monthly temperature for October.
     * @param Nov        average monthly temperature for November.
     * @param Dec        average monthly temperature for December.
     */
    public void setTemperatures(double Jan,
                                double Feb,
                                double Mar,
                                double Apr,
                                double May,
                                double Jun,
                                double Jul,
                                double Aug,
                                double Sep,
                                double Oct,
                                double Nov,
                                double Dec) {

        this.setTemperatures(Jan, Feb, Mar, Apr, May, Jun, Jul, Aug, Sep, Oct, Nov, Dec, "F");  

    }
    
    /**
     * For those cities for which only January, April, July, and October average temperatures
     * are available, interpolates the temperatures for the missing months to set the
     * values in the temperature array.
     *
     * @param Jan        average monthly temperature for January.
     * @param Apr        average monthly temperature for April.
     * @param Jul        average monthly temperature for July.
     * @param Oct        average monthly temperature for October.
     */
    public void setTemperatures(double Jan,
                                double Apr,
                                double Jul,
                                double Oct) {
        double Feb, Mar, May, Jun, Aug, Sep, Nov, Dec;

        Feb = (Jan + Apr) / 2;
        Mar = Feb;
        May = (Apr + Jul) / 2;
        Jun = May;
        Aug = (Jul + Oct) / 2;
        Sep = Aug;
        Nov = (Oct + Jan) / 2;
        Dec = Nov;

        this.setTemperatures(Jan, Feb, Mar, Apr, May, Jun, Jul, Aug, Sep, Oct, Nov, Dec, "F");  
    }

    /**
     * if you want to specify the scale (farenheit or celsius) you are using
     * when setting the temperature array,
     * use this method.
     *
     * @param Jan        average monthly temperature for January.
     * @param Feb        average monthly temperature for February.
     * @param Nar        average monthly temperature for March.
     * @param Apr        average monthly temperature for April.
     * @param May        average monthly temperature for May.
     * @param Jun        average monthly temperature for June.
     * @param Jul        average monthly temperature for July.
     * @param Aug        average monthly temperature for August.
     * @param Sep        average monthly temperature for Sept.
     * @param Oct        average monthly temperature for October.
     * @param Nov        average monthly temperature for November.
     * @param Dec        average monthly temperature for December.
     * @param scale      "F" or "C" (for Farenheit or Celsius).
     * 
     */        
    public void setTemperatures(double Jan,
                                double Feb,
                                double Mar,
                                double Apr,
                                double May,
                                double Jun,
                                double Jul,
                                double Aug,
                                double Sep,
                                double Oct,
                                double Nov,
                                double Dec,
                                String scale) {

        if (scale.regionMatches(true, 0, "C", 0, 1)) {
            averageTemperature[0] =  Jan;
            averageTemperature[1] =  Feb;
            averageTemperature[2] =  Mar;
            averageTemperature[3] =  Apr;
            averageTemperature[4] =  May;
            averageTemperature[5] =  Jun;
            averageTemperature[6] =  Jul;
            averageTemperature[7] =  Aug;
            averageTemperature[8] =  Sep;
            averageTemperature[9] =  Oct;
            averageTemperature[10] =  Nov;
            averageTemperature[11] =  Dec;
        } else if (scale.regionMatches(true, 0, "F", 0, 1)) {
            averageTemperature[0] =  toCelsius(Jan);
            averageTemperature[1] =  toCelsius(Feb);
            averageTemperature[2] =  toCelsius(Mar);
            averageTemperature[3] =  toCelsius(Apr);
            averageTemperature[4] =  toCelsius(May);
            averageTemperature[5] =  toCelsius(Jun);
            averageTemperature[6] =  toCelsius(Jul);
            averageTemperature[7] =  toCelsius(Aug);
            averageTemperature[8] =  toCelsius(Sep);
            averageTemperature[9] =  toCelsius(Oct);
            averageTemperature[10] =  toCelsius(Nov);
            averageTemperature[11] =  toCelsius(Dec);
        } else {
            Util.println("Please enter C for Celsius or F for Farenheit.");
        }
   }

    /** 
     * Returns the average temperature for the city for the month of the date specified.
     * You also specify whether you want to see the temperature as farenheit ("F")
     * or as Celsius ("C").
     *
     * @param    adate        Date you are inquiring about.
     * @param     scale      "F" or "C"
     *
     */
    public double showTheTemperature(Date adate, String scale) {
        String      s = scale.substring(0, 1);
        int         month = JCalendar.getMonth(adate);
        double      temp;

        temp = averageTemperature[month];
                
        if (s.equalsIgnoreCase("F"))
            temp = toFarenheit(temp);
        return (new java.math.BigDecimal(temp).setScale(1, java.math.BigDecimal.ROUND_HALF_DOWN)).doubleValue();
    }

    /** 
     * Displays the average temperature for the city for the month of the date specified.
     * The group's home city determines whether they want to see the temperature as farenheit ("F")
     * or as Celsius ("C").
     *
     * @param Dec        Date you are inquiring about.
     * @param homeCity   the group's home city/
     *
     */
    public double showTemperature(Date stayDate, City homeCity) {
        String scale = "";
        if (homeCity.measurementSystem.equals("nonmetric"))
            scale = "F";
        else
            scale = "C";

        return this.showTheTemperature(stayDate, scale);
    }

    /** 
     * Returns the average temperature for the city for the month of the date specified as a String
     * The group's home city determines whether they want to see the temperature as farenheit ("F")
     * or as Celsius ("C").
     *
     * @param Dec        Date you are inquiring about.
     * @param homeCity   the group's home city/
     *
     */
    public String showTemperatureAsString(Date stayDate, City homeCity) {
        String scale = "";
        String fullScale = FARENHEIT;
        if (homeCity.measurementSystem.equals("nonmetric"))
            scale = "F";
        else {
            scale = "C";
			fullScale = CELSIUS;
		}
		
        return this.showTheTemperature(stayDate, scale) + " degrees " + fullScale;
    }
	
	
	
    /**
     * shows the city's average temperature for the current month, using Farenheit.
     *
     */   
    public double showTemperature() {
        /*
        ** We always want to get the current date and time using the TourDate
        ** class, so that we can get consistent output for testing.
        */
		Date now = JCalendar.getToday();
        return this.showTheTemperature(now, "F");
    }
	
    /**
     * Returns city name.
     *
     */
    public String getName() {
        return name;
    }
	
    /**
     * Returns the full country name.
     *
     * @return      String      full country name.
     */
    public String getDisplayCountry() {
        return countryLong;
    }
	
    /**
     * sets the full country name (looks it up in a table based on the code).
     * called by InsertCities.
     *
     * @param conn        Connection.
     * @param pstmt       a prepared statement (prepared in InsertCities).
     *
     * @exception java.sql.SQLException a error occurred.
     */
    public void setDisplayCountry(Connection conn, PreparedStatement pstmt)
		 throws SQLException {
			 
			 String acountry = "";
			 
			 pstmt.setString(1, this.country);                 
			 ResultSet rs = pstmt.executeQuery();
			 boolean flag = false;   // is country found
			 
			 while (rs.next()) {
				 acountry = rs.getString(1);
				 flag = true;
			 }
			 
			 if (!flag)
				 Util.println("no country found for " + this.country);  
			 
			 countryLong = acountry;
    }
	
    /**
     * Returns language.
     *
     */
    public String getLanguage() {
        return language;
		
    }
	
    /**
     * Method that queries the City table to return a City object. <p>
     *
     * <em>SELECT city FROM Cities WHERE city.getName() = ? and city.getDisplayCountry() = ?</em><p>
     *
     * server-side method <code>getDisplayCountry</code> is used instead of
     * a join.
     *
     * @see #getDisplayCountry().
     *
     * @param     conn       Connection 
     * @param     acity      Cityname (in english)
     * @param     acountry   Countryname (in english)
     *
     * @exception java.sql.SQLException a error occurred.
     */
	public static City findCity(Connection conn,
								String acity,
								String acountry)
		 throws SQLException {
			 
			 City mycity = null;
			 prepareStaticStatements(conn, true);
			 
			 Util.println(
						  "Retrieving city object that matches " +
						  acity +
						  " " +
						  acountry);
			 
			 find_city_by_name_and_country.setString(1, acity);
			 find_city_by_name_and_country.setString(2, acountry);          
			 
			 ResultSet rs = find_city_by_name_and_country.executeQuery();
			 while (rs.next()) {
				 mycity = (City) rs.getObject(1);
				 
				 Util.println("\n" +
							  mycity.getName() + ", in " + acountry + " retrieved\n");
			 }
			 
			 //close the result set 
			 rs.close();
			 
			 Util.println("\n\n");
			 
			 return mycity;
    }
	
    /**
     * Queries the City table to return the appropriate city 
     * object based on the city_id. <p>
     *
     * <em>SELECT city FROM Cities WHERE city_id = ?</em><p>
     *
     * @param     conn       Connection 
     * @param     cityId     City ID
     * @return    City object
     *
     * @exception SQLException  Thrown if there is a SQL error.
     */
    public static City findCity(Connection conn,
                                int cityId)
		 throws SQLException {

			 City mycity = null;
			 
			 prepareStaticStatements(conn, false);
			 find_city_by_id.setInt(1, cityId);
			 
			 ResultSet rs = find_city_by_id.executeQuery();
			 
			 /* Should get at most one row, since city_id is the primary key */
			 while (rs.next()) {
				 mycity = (City) rs.getObject(1);
			 }
             
			 //close the result set 
			 rs.close();
			 return mycity;
    }
	
	
    /**
     * Method that queries the City table. <p>
     *
     * <em>SELECT city FROM Cities WHERE 
     * city.getAirport() = ?</em><p>
     *
     * @param     conn       Connection 
     * @param     ap         airport code
     *
     * @return    City object
     *
     * @exception java.sql.SQLException a error occurred.
     */
    public static City findCity(Connection conn,
                                String ap)
		 throws SQLException {
			 
			 if (ap.length() != 3) {

				 //if executed as a database-side method,
				 // and ? parameter was not cast,
				 // perhaps the second parameter is actually
				 // an int.
				 try {
					 
					 int i = new Integer(ap).intValue();
					 return findCity(conn, i);
					 
				 }
				 catch (java.lang.NumberFormatException e) {
					 return new City();
				 }
			 }
			 City mycity = null;
			 prepareStaticStatements(conn, true);
			 find_city_by_airport.setString(1, ap); 
			 
			 ResultSet rs = find_city_by_airport.executeQuery();
			 
			 while (rs.next()) {
				 mycity = (City) rs.getObject(1);
			 }
             
			 //close the result set 
			 rs.close();
			 
			 return mycity;
    }
	
	
    /**
     * Returns country code.
     *
     * @return String country
     */
    public String getCountryCode() {
        return country;
    }
	
    /**
     * Returns code for city's main airport.
     *
     */
    public String getAirport() {
        return airport; 
    }
	
	
	
    /**
     * Returns a boolean indicating whether the airport belongs to the city object.
     *
     * @return boolean airport
     */
	
    public boolean airportInCity(String ap) {
        return (ap.equals(airport));
    } 
	
    /**
     * Utility to convert a temperature degree in Celsius to Farenheit
     */
    private static double toFarenheit(double c) {
        double f;
		
        f = ((c*9)/5)+32;
        return f;
    }
	
    /**
     * Utility to convert a temperature degree in Farenheit to Celsius
     */
    private static double toCelsius (double f) {
        double c;
		
        c = ((f - 32) * 5) / 9;
		
        return c;
    }
	
    /**
     * Print city information to the output stream.
     *
     * @param     adate       date of visit to the city
     * @param     homeCity    the visitor's home city 
     *
     * @return    String      object information about the city
     */
    public String getInfo(Date adate, City homeCity) {

        StringBuffer cityInfo = new StringBuffer();

        cityInfo.append("\nA city in your two-city tour is: " + getName());
        cityInfo.append("\nUsing stored object data and class methods to profile the city");
        if (isTropical())
            cityInfo.append("\nThis city is in the tropics.");

        cityInfo.append("\nThe average day-time temperature in that city during your visit is: \n" +
            showTemperatureAsString(adate, homeCity));
        
        cityInfo.append("\nSpoken language: " + getLanguage());
        cityInfo.append("\n" + getName() + " is " +
                           getDistanceFromHomeCityAsString(homeCity) +
                           " from " + homeCity.getName() + "."); 
        cityInfo.append("\nThe time zone of that city is " +
                           getTimeDifferenceAsString(adate, homeCity) +
                           " than " + homeCity.getName() + " during your visit.\n");
	return cityInfo.toString();
    }


    /**
     * Sets the Location object for the city.
     *
     * @param     latDeg      degrees latitude

     * @param     latMin      minutes latitude
     * @param     latDir      "N" or "S" for  latitude
     * @param     longDeg     degrees longitude
     * @param     longMin     minutes longitude
     * @param     longDir     "E" or "W" for  latitude
     *
     */
    public void setLocation (int latDeg,
                             int latMin,
                             String latDir,
                             int longDeg,
                             int longMin,
                             String longDir) {
        cityLocation = new Location(latDeg, latMin, latDir, longDeg, longMin, longDir);
    }
	
    /**
     * Returns the city's Location. 
     *
     * @return Location  cityLocation
     */
	
    public Location getLocation() {
        return cityLocation;
    }
	
    /**
     * Returns the city's Location's toString() value. For use in Cloudview. 
     *
     * @return Location  cityLocation
     */
    public String getLocationAsString() {
        return getLocation().toString();
    }
	
    /**
     * Calculates the distance between current city and another city.
     * You specify the second city by stating its airport code.
     *
     * @param     conn       Connection 
     * @param     ap         the other city's airport code
     *
     * @return    int        number of miles between them.
     *
     * @exception SQLException  Thrown if there is a SQL error.
     */
    public int getDistanceFrom(Connection conn, String ap) throws SQLException {
        City cityb = findCity(conn, ap);
        return this.getDistanceFrom(cityb);
    }
	
    /**
     * Calculates the distance between current city and another city.
     * You specify the second city by stating its city id.
     *
     * @param     conn       Connection 
     * @param     i          the other city's id
     *
     * @return    int        number of miles between them.
     *
     * @exception SQLException  Thrown if there is a SQL error.
     */
    public int getDistanceFrom(Connection conn, int i) throws SQLException {
        City cityb = findCity(conn, i);
        return this.getDistanceFrom(cityb);
    }
	
    /**
     * Calculates the distance between current city and another city.
     *
     * @param     conn       Connection 
     * @param     cityb      the other city
     *
     * @return    int        number of miles between them.
     *
     * @see Location#getDistance
     */
    public int getDistanceFrom(City cityb) {
        int distance;
        Location locationB = cityb.getLocation();
        distance = (getLocation()).getDistance(locationB);
       return distance;
    }

    /**
     * Calculates the distance between current city and another city
     *  and converts units to the appropriate measurement system.
     *
     * @param     cityb      the other city
     *
     * @return    int        number of miles between them OR number of kilometers.
     *
     * @see Location
     */
    public int getDistanceFrom(City cityb, int i) {
        double distance;
        Location locationB = cityb.getLocation();
        distance = (double) (getLocation()).getDistance(locationB);
        if (cityb.measurementSystem == "metric")
           distance = distance * 1.6093; 
        return (new java.math.BigDecimal(distance).setScale(0, java.math.BigDecimal.ROUND_CEILING)).intValue();
    }


    /**
     * Calculates the distance between current city and home city using
     * measurement system of home city, and converts units to kilometers if
     * necessary
     *
     * @param     cityb      the other city
     *
     * @return    String     String showing distance between cities
     *
     * @see Location
     */

  public String getDistanceFromHomeCityAsString(City homeCity) {
  String units = " " + MILES;
  if (homeCity.measurementSystem == "metric")
     units = " " + KILOMETERS;
  return getDistanceFrom(homeCity, 1) + units;
  }

    /**
     * Sets the TimeZone object for the city.
     *
     * @param GMToff    Offset in hours from Greenwich Mean Time
     * @param DSTGMToff Offset in hours from Greenwich Mean Time when
     *                  Daylight Savings Time is in effect.
     * @param beginDST  A String code that represents when Daylight Savings
     *                  Time begins.  The possible values are:
     *  <PRE>
     *      "FSA" - First Sunday in April
     *      "FSO" - First Sunday in October
     *      "SSO" - Second Sunday in October
     *      "SSM" - Second Sunday in March
     *      "SSF" - Second Sunday in February
     *      "FSS" - First Sunday in September
     *      "FSM15" - First Sunday on or after March 15
     *      "LSO" - Last Sunday in October
     *      "LSM" - Last Sunday in March
     *      "FSM" - First Sunday in May
     *      "TSO" - Third Sunday in October
     *      "TSS" - Third Sunday in September
     *      "FFM" - First Friday in May
     *      "LWS" - Last Wednesday in September
     *      "LSS" - Last Sunday in September
     *      "SSA" - Second Sunday in April
     *      "TSM" - Third Sunday in March
     *  </PRE>
     * @param endDST    A String code that represents when Daylight Savings
     *                  Time ends.  See above for possible values.
     *
     * @see TimeZone
     *
     */
    public void setTimeZone(double GMToff,
                            double DSTGMToff,
                            String beginDST,
                            String endDST) {

       cityTimeZone = new TimeZone(GMToff,
                                    DSTGMToff,
                                    beginDST,
                                    endDST);
    }

    /**
     * Sets the TimeZone object for the city. Use for cities that don't have
     * Daylight Savings Times.
     *
     * @param     GMToff      the GMT time offset (e.g., -3)
     *
     */
    public void setTimeZone(double GMToff) {

        cityTimeZone = new TimeZone(GMToff);
    }

    /**
     * Returns the city's TimeZone object. 
     *
     * @return TimeZone  cityTimeZone
     */
    public TimeZone getTimeZone() {
         return cityTimeZone;
    }
  
    /**
     * Returns the city's TimeZone's toString() value. For use in Cloudview.
     *
     * @return TimeZone  cityTimeZone
     */
    public String getTimeZoneAsString() {
        return getTimeZone().toString();
    }

    /**
     * Calculates the time zone difference between current city and another city for date specified.
     *
     * @param     theDate      desired date
     *
     * @param    cityb        the other city
     *
     * @return   double        the time difference
     *
     */
    public double getTimeDifference(java.sql.Date theDate, City cityb) {
        return this.getTimeZone().getTimeDifference(cityb.getTimeZone(), theDate);
    }


    /**
     * Calculates the time zone difference between current city and another city for date specified
     * and returns a String.
     *
     * @param     theDate      desired date
     * @param     cityb        secondCity
     *
     * @return    String       string showing time difference
     *
     */
    public String getTimeDifferenceAsString(java.sql.Date theDate, City cityb) {
        String laterOrEarlier = "later"; 
        double timedifference = this.getTimeDifference(theDate, cityb);
        if (timedifference > 0)
            laterOrEarlier = "earlier";
        return java.lang.Math.abs(timedifference) + " hours " + laterOrEarlier;

    }



    /**
     * Calculates the time zone difference between current city and 
     * another city for date specified.
     * You specify the second city by its airport code.
     *
     * @param     theDate      desired date
     *
     * @return    ap           the airport code of the other city.
     *
     * @exception SQLException  Thrown if there is a SQL error.
     */
    public double getTimeDifference(Connection conn, java.sql.Date theDate, String ap) throws SQLException {
 
        City cityb = findCity(conn, ap);
  
        return this.getTimeDifference(theDate, cityb);
    }

    /**
     * Returns true if the city lies within the tropics.
     *
     */
    public boolean isTropical() {
        return getLocation().isTropical();
    }

    /**
     * Returns a double latitude value. Allows you to order cities 
     * from north to south.
     *
     */
    public double getNorthSouthOrderableLocation() {
 
        return getLocation().getNorthSouthOrderableLocation();
    }

    /**
     * Returns a double latitude value. Allows you to order cities from east
     * to west.
     *
     */
    public double getEastWestOrderableLocation() {
        return getLocation().getEastWestOrderableLocation();
    }

    /**
     * Good Cloudscape practice requires that an object override
	 * equals and hashCode for correct implementation of = and <>
	 * in queries.
     *
     * @return    boolean        if the two objects are equals.
     *
     */
	public boolean equals(Object o)  {
		City oc = null;
		try {
			oc = (City) o;
		}
		catch (ClassCastException ce) {
			return false;
		}
		if (this.compareTo(oc)==0)
			return true;
		else
			return false;
	}

    /**
     * This method allows a City object to be <i>orderable</i> within SQL-J
	 * statements.<p>
	 * 
	 * We have arbitrarily decided to order cities primarily by their Locations, and
	 * secondarily by their time zones. See the javadoc for Location and TimeZone to see
	 * how those objects are ordered.
     *
     * @return    int        -1 if this object is "less than" the parameter, 0 if they or equal, or 1 if it is "greater than" the parameter..
	 *
     * @exception ClassCastException  Thrown if the objects cannot be compared.
     *
	 * @see TimeZone
	 * @see Location
     *
     */
	public int compareTo(Object o) throws ClassCastException {
		//handle nulls;
		// nulls sort lower than everything
		if (o == null)
			return 1;
		// may throw an exception;
        City oc = (City) o;
		int locations = this.getLocation().compareTo(oc.getLocation());
		int timezones = this.getTimeZone().compareTo(oc.getTimeZone());
		

		if (locations == 0 && timezones ==0)
			return 0;
		
		// order the Cities primarily by their locations
		if (locations > 0)
			return 1;
		if (locations < 1)
			return -1;
		// if their locations are equal, then look at time zones.
		if (timezones > 0)
			return 1;
		//the only other possibility is that timezones < 1
		
		else
			return -1;
	}

	/* for correct equality comparisons
	 *
     *
	 * @return int   a hash code
	 *
	 */

	public int hashCode() {
		return this.getLocation().hashCode();
	}

	
    /**
     * Prepares some or all of the PreparedStatements. This method is called
     * by static methods.
     *
     * @param connn    a connection to the database
     *
     * @return boolean if statements are prepared without error
     *
     * @exception  SQLException if there is an SQL error
     */

  public static boolean prepareStaticStatements(Connection conn, boolean all) throws SQLException {
    if (find_city_by_id == null)
      find_city_by_id = conn.prepareStatement("EXECUTE STATEMENT getCityById");

    /* the boolean in this case means that you want the extra statements prepared, which aren't currently being used by the app. But a user may want to use them in ij or some other place.  If so, prepare them, too.*/

    if (all) {
      if (find_city_by_name_and_country == null)
	find_city_by_name_and_country = conn.prepareStatement("SELECT city " +
            "FROM Cities " +
            "WHERE city.getName() = ? AND city.getDisplayCountry() = ?"
            );
      if (find_city_by_airport == null)
	find_city_by_airport = conn.prepareStatement("SELECT city " +
                "FROM Cities " +
                "WHERE city.getAirport() = ?"); 
    }

    if (all) {
    if (find_city_by_id != null && find_city_by_name_and_country != null && find_city_by_airport != null)
      return true;
    else
      return false; }

    else {
      if (find_city_by_id != null)
	return true;
      else
	return false;

    }

  }

    /**
     * Externalizable interface
     * Write this object to ObjectOutput stream
     *
     * @exception IOException Thrown if there is an IO Error writing the object
     */
    public void writeExternal(ObjectOutput out) throws IOException {
        out.writeInt(currentClassVersion);
        out.writeInt(city_id);
        out.writeUTF(name);
        out.writeUTF(language);
        out.writeUTF(country);
        out.writeUTF(countryLong);
        out.writeUTF(airport);
        out.writeObject(cityLocation);
        out.writeObject(cityTimeZone);
        for (int i = 0; i < averageTemperature.length; i++) {
            out.writeDouble(averageTemperature[i]);
        }
    }

    /**
     * Externalizable interface
     * Read this object from ObjectOutput stream
     *
     * @exception IOException Thrown if there is an IO Error reading the object
     *      or the input stream does not contain the expected objects in the
     *      expected order
     */
    public void readExternal(ObjectInput in) throws IOException {
        int version = in.readInt();
        if (version != currentClassVersion) {
            throw new IOException(
                        "Expected class version " +
                        currentClassVersion +
                        " when reading City, but got class version "
                        + version);
        }

        try {
            city_id     = in.readInt();
            name        = in.readUTF();
            language    = in.readUTF();
            country     = in.readUTF();
            countryLong = in.readUTF();
            airport     = in.readUTF();
            cityLocation    = (Location) in.readObject();
            cityTimeZone    = (TimeZone) in.readObject();

            for (int i = 0; i < averageTemperature.length; i++) {
                averageTemperature[i] = in.readDouble();
            }

            if (country.equals("US"))
                measurementSystem = "nonmetric";
            else
                measurementSystem = "metric";
        } catch (ClassNotFoundException e) {
            throw new IOException("Unexpected exception: " + e);
        } 
    }

}
