package JBMSTours.serializabletypes;

import java.sql.*;
import java.math.BigDecimal;
import java.io.Serializable;
import JBMSTours.Util;
import JBMSTours.JCalendar;
import JBMSTours.AvailabilityException;
import JBMSTours.BudgetException;


/**
 * Look at the <a href="../JBMSTours/serializabletypes/CustomerFlight.java">Source</a>.<p>
 * A CustomerFlight is A FlightObject wrapped in specific
 * date, booking, and customer information.
 * It contains a Flight object as a field (which can be Flight,
 * OneStopFlight, or TransferFlight). 
 */
public class CustomerFlight implements Serializable  {

    /* 
     * we don't need to store the cities themselves because they are 
     * already stored in the Cities table. Just store the id. 
     * Marking a field transient allows it to be part of an 
     * object without being stored.  To retrieve the city back 
     * into the object, use the getDepartCity or getArriveCity
     * methods.
     */
    transient public City departCity;

    /* city_id of the departure city*/
    public int departCity_id;

    /* CityObject of the arrival city*/
    transient public City arriveCity;

    /* city_id of the arrival city*/
    public int arriveCity_id;

    /* number of seats needed by the group*/
    public int numberSeats;

    /* fixed rate (per seat) */
    public BigDecimal fixedRate;

    /* total cost for the group*/
    public BigDecimal totalCost;

    private Flight aFlight;

    private Time departureTime;

    private Time arrivalTime;

    private Date travelDate;

    public short level;

    private int myGroupId;

    private transient PreparedStatement propose_a_flight;
    private transient PreparedStatement insert_new_information;
	private transient PreparedStatement fa_cursor;
	private transient PreparedStatement update_cursor;

    /**
     * Constructs a customer flight object. Initializes the date of 
     * travel, the city of origin, the destination city, the number 
     * of seats needed, and the level of travel. To add the actual 
     * Flight object to CustomerFlight, use the proposeFlight method. 
     * The bookFlight method adds the rate information.
     *
     * @param travel     Date of travel
     * @param city1      City of origin
     * @param city2      City of destination

     * @param num        Number of seats needed
     * @param     The level of the tour. See the static variables in Tour.java.
     */
    public CustomerFlight (Date travel, City city1, City city2, int num, short alevel, int gid) {
    
        departCity = city1;
        departCity_id = city1.city_id;
        arriveCity=  city2;
        arriveCity_id = city2.city_id;
        travelDate = travel;    
        numberSeats = num;
        fixedRate = new BigDecimal(0); 
        totalCost = new BigDecimal(0);
        aFlight = new Flight();
        departureTime = new Time(0, 0, 0);
        arrivalTime = new Time(0, 0, 0);
        level = alevel;
        myGroupId = gid;
    }

    /**
     * Checks the rate of the currently proposed flight. Does not fix the rate.
     * If rate is already fixed, returns that value.
     *
     * @param     conn          Connection to the database.
     * 
     * @exception SQLException  Thrown if there is a SQL error.
     */

    public BigDecimal checkRate(Connection conn) throws SQLException {
        if (fixedRate.equals(new BigDecimal(0.0)))
            return aFlight.getRate(conn, level).setScale(2, BigDecimal.ROUND_CEILING);
        return fixedRate;
    }

    /**
     * Checks the total of the currently proposed flight (rate * number of
     * seats). Does not fix the rate. If rate is already fixed, returns that value.
     *
     * @param     conn          Connection to the database.
     *
     * @exception SQLException  Thrown if there is a SQL error. 
     */
    public BigDecimal checkTotalCost(Connection conn) throws SQLException {
        if (totalCost.equals(new BigDecimal(0.0)))
            return (checkRate(conn)).multiply(new BigDecimal((double) numberSeats)).setScale(2, BigDecimal.ROUND_CEILING);
        return totalCost;
    }

    /**
     * Fixes the price of the flight. Called by the booking method.
     *
     * @param  Conn    Connection to the database
     *
     * @exception SQLException   if there is a database error
     * 
     */
    private void setPrice(Connection conn) throws SQLException {
        Util.println("Fixing the rate.");
        fixedRate = checkRate(conn);
        totalCost = fixedRate.multiply(new BigDecimal((double) numberSeats)).setScale(2, BigDecimal.ROUND_CEILING);
    }

    /**
     * Proposes a flight for the group.
     * Returns true if successfully proposed a flight.
     *
     * @param conn              Connection to db.
     * @param availableFunds    the amount of money the group has left to spend
     *
     * @return                  true if successfully proposed flight.
     *
     * @exception SQLException             Thrown if there is a SQL error.
     * @exception BudgetException          Thrown if the flight is out of the customer's budget.
     * @exception AvailabilityException    Thrown if the flight is not available.
     */
    public BigDecimal proposeFlight(Connection conn, BigDecimal availableFunds) throws SQLException, AvailabilityException, BudgetException {
        boolean available = false;
		boolean withinbudget=false;
        String flight_id = "";
        String airport1 = "";
        String airport2 = "";

		prepareStatements(conn);
        propose_a_flight.setInt(1, myGroupId);
		propose_a_flight.setString(2, getDepartCity(conn).getAirport());
		propose_a_flight.setString(3, getArriveCity(conn).getAirport());
		
		BigDecimal tempCost = new BigDecimal(0);
        ResultSet rs = propose_a_flight.executeQuery();
        while (rs.next()) {
            aFlight = (Flight) rs.getObject(1);
            Util.println("\nProposing flight " + aFlight.toString() 
						 + ". Checking if total cost within budget and availability");
			tempCost = this.checkTotalCost(conn);
            if (tempCost.compareTo(availableFunds) == -1) {
				withinbudget=true;
                if (this.checkAvailability(conn)) {
					available=true;
                }
			}
			if (withinbudget && available)
				break;
			
        }
             
        rs.close();
		propose_a_flight.close();
		if (withinbudget == false)
			throw new BudgetException(this);
		if (available == false)
			throw new AvailabilityException(this);
            departureTime = aFlight.departTime;
            arrivalTime = aFlight.arriveTime;
            Util.println("Flight proposed.");
            Util.println("Departing " + departureTime + 
                                     ", arriving " + arrivalTime + 
                                     ", rate $" + checkRate(conn));
            return tempCost;
    }



    /**
     * Proposes a flight for the group.
     * Returns true if successfully proposed a flight. Allows you to override level.
     *
     *
     * @param conn              Connection to db.
     * @param availableFunds    the amount of money the group has left to spend
     *
     * @return                  true if successfully proposed flight.
     *
     * @exception SQLException  Thrown if there is a SQL error.
     * @exception BudgetException          Thrown if the flight is out of the customer's budget.
     * @exception AvailabilityException    Thrown if the flight is not available.
     */
    public BigDecimal proposeFlight(Connection conn, BigDecimal availableFunds, short templevel) throws SQLException, AvailabilityException, BudgetException {
		this.level = templevel;
		Util.println("Downgrading flight level to: " + Tour.getStringForLevel(templevel));
		return proposeFlight(conn, availableFunds);
	}


	/**
     * Books the flight. --Updates the FlightBookings and the FlightAvailability tables.
     * @param conn     Connection to db.
     * @param gid      group_id.
     *
     * @return    the total cost of the flights for the group
     *
     * @exception SQLException  Thrown if there is a SQL error.
     */
    public BigDecimal bookFlight(Connection conn, int gid) throws SQLException {

        Util.println("\nBooking flight " + aFlight.toString());
        int n = 0;
        setPrice(conn);
  
        if (aFlight.storeBookingInfoForAirline(conn, gid, travelDate, numberSeats, fixedRate, level) == 1)
            Util.println();

        String updateAvail;
        String mycolumn = Flight.levelToSeats(level) + "_taken";
		String mycursor = Flight.levelToCursor(level);

        ResultSet rs;
  
        prepareStatements(conn);
        
   	    fa_cursor.setCursorName("FAROWS");

        fa_cursor.setString(1, getFlight());
        fa_cursor.setInt(2, aFlight.getSegmentNumber());
        fa_cursor.setString(3, aFlight.getAltFlightId());
        fa_cursor.setInt(4, aFlight.getAltSegmentNumber());
        fa_cursor.setDate(5, travelDate);
        rs = fa_cursor.executeQuery();

		// we used to test whether "mycolumn" had a value of null.
		// now, the column is defined with a default of 0. It iwll never be null.
 		update_cursor = conn.prepareStatement("UPDATE FlightAvailability " +
											  "SET "+ mycolumn +
											  " = " + mycolumn + " + ? " +
											  "WHERE CURRENT OF FAROWS");
		
        String fid = "";
		int sn =0;
        /*if flight_id/segment_number already has information, update it.*/
        while (rs.next()) {
            update_cursor.setInt(1, numberSeats);
            n += update_cursor.executeUpdate();
			fid = rs.getString(2);
			sn = rs.getInt(3);

        }


		if (n < 2) {

            if (!fid.equals(getFlight()) && aFlight.getSegmentNumber()!= sn) {
				insert_new_information.setString(1, getFlight());
				insert_new_information.setInt(2, aFlight.getSegmentNumber());
				insert_new_information.setDate(3, travelDate);
				insert_new_information.setInt(4, numberSeats);
				n += insert_new_information.executeUpdate(); 
			}
			if (aFlight.getAltSegmentNumber() > 0 &&
				(!fid.equals(aFlight.getAltFlightId()) && aFlight.getAltSegmentNumber()!= sn)) {
				insert_new_information.setString(1, aFlight.getAltFlightId());
				insert_new_information.setInt(2, aFlight.getAltSegmentNumber()); 
				n += insert_new_information.executeUpdate();
			}
        }
                           
                           
        Util.println("FlightAvailability Table Updated--" + n + " rows changed.");
        Util.println(this.toString(conn) + " booked.");
        // don't commit within method to allow method to be
		//used as a server-side method or replication work
		// unit.
		//conn.commit();
		rs.close();
		fa_cursor.close();
		update_cursor.close();
		insert_new_information.close();
        return totalCost;
    }

    /**
     * Gets the Flight ID of the Flight.
     *
     * @return  String   The flight_id
     *
     */
    public String getFlight() {
        return aFlight.getFlightId();
    }

    /**
     * Gets the Flight object (a field).
     *
     * @return  The flight (a Flight object)
     *
     * @see Flight
     * @see OneStopFlight
     * @see TransferFlight
     *
     */
     
    public Flight getFlightObject() {
        return aFlight;
    }

    /**
     * Checks the availability of the flight.
     * Returns true if flight is available.
     *
     * @param conn              Connection to db.
     *
     * @return                  true if on requested date.
     *
     * @exception SQLException  Thrown if there is a SQL error.
     */
    private boolean checkAvailability (Connection conn) throws SQLException {

        Util.println("Checking availability on . . ." + travelDate);

        int availableSeats = 0;
  
        
 
        return aFlight.checkSegmentAvailability(conn, 
                getTravelDate(), level, numberSeats);

    }

    /**
     * Gets the date of travel for the flight (date of departure).
     *
     * @return Date   Date flight departs.
     *
     */
    public Date getTravelDate() {
        return travelDate;
    }

    /** 
     * Date customers actually arrive in city (for long flights). 
     * If the flight arrives after 4 a.m., no hotel stay needed 
     * until date of actual arrival, so therefore this method
	 * returns the actual date that they arrive.  But if the flight arrives 
     * before 4 a.m, return the date before the actual arrival date
	 * because they will need a hotel stay for the remainder of the
	 * "night". 
     * Used by HotelStay.
     *
     * @see HotelStay
	 *
	 * @param     conn         A connection to the database.
     * @exception SQLException thrown on database access failure
     */
    public Date getArrivalDate(Connection conn) throws SQLException {

        double flyingTime = aFlight.getTotalDuration();
        double timeZoneDifference = getDepartCity(conn).getTimeDifference(travelDate, getArriveCity(conn));
        double tempTime = flyingTime - timeZoneDifference;
        double basicDifference = JCalendar.getTimeDifference(departureTime, arrivalTime);
        double discrepency = basicDifference - tempTime;
 
        /*
         * if the arrival time is before the departure time, 
         * check to see if a day really passed OR
         * if the difference is due to timezone differences.
         */
        if (arrivalTime.before(departureTime)) {
            if (discrepency < 2) {
                if (arrivalTime.after(new java.sql.Time(4,0,0))) {
                    return JCalendar.incrementDate(getTravelDate());
				}
                /*
                 * the flight arrives before four a.m., 
                 * so don't count it as the next day
                 */
                else {
					return getTravelDate();
				}
            } 
            if (discrepency > 22 & discrepency < 46)
                return JCalendar.incrementDate(getTravelDate());
            Util.println("It's a two-day flight!");
            return JCalendar.incrementDate(JCalendar.incrementDate(getTravelDate()));
        }
        /** 
         ** if the arrival is after the departure, 
         ** check to see that it's not a REALLY long flight!
          **/
        else {
            /*the total flight is at least a day, but under two days*/

            if (tempTime > 24 && tempTime < 48) {
                if (arrivalTime.after(new java.sql.Time(4,0,0)))
                    return JCalendar.incrementDate(getTravelDate());
                else return getTravelDate();
            }

            /* the total flights is over two days!*/
            if (tempTime > 48) {
                if (arrivalTime.after(new java.sql.Time(4,0,0)))
                    return JCalendar.incrementDate(getTravelDate());
                else
                    return JCalendar.incrementDate(JCalendar.incrementDate(getTravelDate()));
            }
            /* otherwise, it all occurs in the same day.*/
            return getTravelDate();  
        }
    }

    /** 
     * Returns a string value for the CustomerFlight. 
     *
     * @param conn     Conenction to the database.
     *
     * @return String  A string value for the CustomerFlight.   
     *
     * @exception SQLException  In case of a database error.
     *
     */
    public String toString(Connection conn) throws SQLException {

        return "Flight number " + aFlight.flight_id + " " 
            + getDepartCity(conn).getName() + " to " 
            + getArriveCity(conn).getName() + " flight for " 
            + numberSeats + " on " + travelDate.toString();
    }

    /** 
     * Returns a string value for the CustomerFlight. 
     *
     *
     * @return String  A string value for the CustomerFlight.   
     *
     */
    public String toString()  {

        return aFlight.toString() + 
            " flight for " + numberSeats + " on " + travelDate.toString();
    }

    /** 
     * Prints detailed information about the CustomerFlight. 
     *
     * @param conn  A connection to the database.
     *
     * @return  String   object information about the CustomerFlight 
     *
     * @exception SQLException   in case of a database error.  
     *
     */
    public String getInfo(Connection conn) throws SQLException {
        StringBuffer flightInfo = new StringBuffer();
        flightInfo.append("\n" + this.toString(conn));
        flightInfo.append("\nNumber of miles: " + aFlight.mileage);
        flightInfo.append("\nTotal duration of flight: " 
            + Util.round(aFlight.getTotalDuration(), 0.001) + " hours.\n\n");
        flightInfo.append("\nGuaranteed rate: $" + fixedRate);
        return flightInfo.toString();
    }

    /**  
     * Getter for the city where the flight arrives.
     *
     *
     * @param conn  A connection to the database.
     *
     * @return City City where the flight arrives.   
     *
     * @exception SQLException   in case of a database error
     *
     */
    public City getArriveCity(Connection conn) throws SQLException {
        if (arriveCity != null)
            return arriveCity;

        return City.findCity(conn, arriveCity_id);

    }

    /**  
     * Getter for the city where the flight departs from.
     *
     *
     * @param conn  A connection to the database.
     *
     * @return City City where the flight departs from.   
     *
     * @exception SQLException   in case of a database error
     *
     */
    public City getDepartCity(Connection conn) throws SQLException {
        if (departCity != null)
            return departCity;

        return City.findCity(conn, departCity_id);

    }

    /**
     * Prepares all the PreparedStatements for this class.
     *
     * @param conn    a connection to the database.
     *
     * @return boolean if statements are prepared without error
     *
     * @exception  SQLException   If there is an SQL error
     */
	public boolean prepareStatements(Connection conn) throws SQLException {
		
		//if (propose_a_flight == null)
			propose_a_flight = conn.prepareStatement("EXECUTE STATEMENT proposeAFlight");
		
			//if (insert_new_information == null) {
			String mycolumn = Flight.levelToSeats(level) + "_taken";
			insert_new_information = conn.prepareStatement(
														   "INSERT INTO FlightAvailability " +
														   "(flight_id, segment_number, flight_date, " +
														   mycolumn + ") VALUES (?, ?, ?, ?)");
			
			
			//}
		//if (fa_cursor == null) {
			String mycursor = Flight.levelToCursor(level);
			fa_cursor = conn.prepareStatement("EXECUTE STATEMENT flightAvailabilityCursor" + mycursor);
			//}		
		return true;
		
		
	}

   /**
     * Returns true if the two objects are equal.
     *
     */
	public boolean equals(Object o)  {
		CustomerFlight ocf=null;
		try {
			ocf = (CustomerFlight) o;
		}
		catch (ClassCastException cce) {

			return false;
		}
        if (this.departCity_id == ocf.departCity_id &&
			this.arriveCity_id == ocf.arriveCity_id &&
			this.travelDate == ocf.travelDate &&
			this.myGroupId == ocf.myGroupId &&
			this.numberSeats == ocf.numberSeats &&
			this.fixedRate == ocf.fixedRate && 
			this.aFlight == ocf.aFlight)
			return true;
		else
			return false;
	}


	
	
}


