package JBMSTours.serializabletypes;

import java.sql.Connection;
import java.sql.Statement;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Time;
import java.sql.Date;
import java.sql.SQLException;
import java.io.Serializable;
import java.math.BigDecimal;
import JBMSTours.Util;
import JBMSTours.JCalendar;

/**
 * Look at the <a href="../JBMSTours/serializabletypes/Flight.java">Source</a>.<p>
 * Flight describes an airline flight segment for the tours
 * database.  This is a superclass for one-leg and multi-leg
 * flights.  These objects are stored temporarily 
 * in the database in the <i>FlightObjects</i> table.
 *
 * @see OneStopFlight
 * @see TransferFlight
 */ 

public class Flight implements Serializable {

    /**
     * A unique id for this flight
     */
    public String flight_id;

    /**
     * The segment number of the flight.
     */

    public int segmentNumber;

    /**
     * The airline providing the flight
     */
    public String airline;

    /**
     * The airport where the flight starts
     */
    public String orig_airport;

    /**
     * The airport where the flight ends
     */
    public String dest_airport;

    /**
     * The departure time
     */
    public Time departTime;

    /**
     * The arrival time
     */
    public Time arriveTime;

    /**
     * The duration of the flight
     */
    public double flyingTime;
    
    public int mileage = 0;

    protected transient PreparedStatement insert_flight_bookings;
    protected transient PreparedStatement get_airline_pricing_info;

    public static final String ECONOMYFLIGHTLEVEL = "economy";
    public static final String BUSINESSFLIGHTLEVEL = "business";
    public static final String FIRSTCLASSFLIGHTLEVEL = "first class"; 


    static final long serialVersionUID = -7943693713518499018L;

    
    
    /**
     * Basic empty constructor.
     */
    public Flight () {
        flight_id = "000000";
        segmentNumber = 0;
        orig_airport = "unknown";
        dest_airport = "unknown";  
        departTime = new Time(0, 0, 0);
        arriveTime = new Time(0, 0, 0);
        flyingTime = 0.0;
        airline = "un";
    }


    
    /**
     * Constructs a Flight object. This is a non-stop flight.
     *
     * @param flightID         flight number
     * @param segment_number   flight segment number
     * @param origAirport      airport of departure
     * @param depart_time      time the flight departs
     * @param dest_airport     airport of arrival
     * @param arrive_time      time of arrival
     * @param flytime          flying time of flight
     * @param miles            number of miles in flight
     */
    public Flight (String flightId, int segment_number, String origAirport,
                   Time depart_time, String destAirport, Time arrive_time,
                   double flytime, int miles) {
        flight_id = flightId;
        segmentNumber = segment_number;
        orig_airport = origAirport;
        dest_airport = destAirport;  
        departTime = depart_time;
        arriveTime = arrive_time;
        flyingTime = flytime;
        mileage = miles;

        airline = flightId.substring(0,2);
    };


    /**
     * Returns flight information as a string.
     */
    public String toString() {
        return getType()  + ", flight " + flight_id + ", segment " + segmentNumber
            + ", " + orig_airport + " to " + dest_airport;
    }
 
    /**
     * Prints information about the flight.
     *
     */
    public void printInfo() {
        Util.println();
        Util.println("Flight number " + flight_id + " " + orig_airport+ " " + dest_airport );
        Util.println();
    } 
  
    /**
     * Checks the rate of the flight, using information from 
     * the airlines table. Algorithm for generating the rate 
     * is imaginary only. It probably does not resemble 
     * the actual way airlines price flights.
     *
     * @param conn database connection to use
     * @param     The level of the tour. See the static variables in Tour.java.
     *
     * @exception SQLException returned on errors in database access
     */
    public BigDecimal getRate(Connection conn, short level) 
            throws SQLException {

        double basicRate = 0.0;
        double distanceDiscount = 0.0 ;
        double tempDistance = 2000.00;
        double businessFactor= 0.0 ;
        double firstClassFactor= 0.0 ;
        double competitionFactor= 0.0 ;
        double factor = 1.0;
        double rate = 0.0;

        Statement stmt = conn.createStatement();
        prepareStatements(conn);
	get_airline_pricing_info.setString(1, airline);
        ResultSet rs = get_airline_pricing_info.executeQuery();
  
        while (rs.next()) {
            basicRate = rs.getDouble(1);
            distanceDiscount = rs.getDouble(2);
            businessFactor = rs.getDouble(3);
            firstClassFactor = rs.getDouble(4);
        }

        rs.close();
  
        while (mileage - tempDistance > 2000.00 ) {
            basicRate = basicRate * (1-distanceDiscount);
            tempDistance+=2000.00;
        }

   
        if (checkLevel(level).equals("first class"))
            factor = 1.0 + firstClassFactor;

        if (checkLevel(level).equals("business"))
            factor = 1.0 + businessFactor;

        rate = mileage*basicRate*factor;
        return new BigDecimal(rate).setScale(2, BigDecimal.ROUND_CEILING);
    }

    /**
     * Returns the flight's segment number.
     */
    public int getSegmentNumber() {
        return segmentNumber;
    }

    /**
     * Returns the flight_id.
     */
    public String getFlightId() {
        return flight_id;
    }

    public String getAltFlightId() {
        return "none";
    }

    public int getAltSegmentNumber() {
        return 0;
    }

    public String getType() {
        return "Non-stop flight";
    }

    public double getTotalDuration() {
        return flyingTime;
    }

  public String getUniqueId() {
  
    return getFlightId() + orig_airport + dest_airport;
  }

    /**
     * "Books" the flight by extracting the pertinent information 
     * to a separate table (FlightBookings). This is the 
     * information that eventually goes to the airline.
     *
     * @param conn        Database connection 
     * @param gid         The group_id
     * @param travelDate  Date that travel begins.
     * @param numberSeats number of seats reserved
     * @param fixedRate   the rate for those seats
     * @param     The level of the tour. See the static variables in Tour.java.
     *
     * @return    > 0 if FlightBookings table successfully updated.
     *    
     * @exception SQLException in case anything goes wrong during a SQL operation
     */

    public int storeBookingInfoForAirline(Connection conn, 
            int gid, Date travelDate, int numberSeats, 
            BigDecimal fixedRate, short level) throws SQLException {

    
        prepareStatements(conn);
        insert_flight_bookings.setString(1, flight_id);
        insert_flight_bookings.setInt(2, getSegmentNumber());
        insert_flight_bookings.setInt(3, gid);
        insert_flight_bookings.setDate(4, travelDate);
        insert_flight_bookings.setInt(5, numberSeats);
        insert_flight_bookings.setObject(6, fixedRate);
        insert_flight_bookings.setString(7, checkLevel(level));
   
        int n = insert_flight_bookings.executeUpdate();
   
        if (n > 0)
            Util.println("Updated FlightBookings table."); 
        return n;  

    }


    /**
     * Checks whether the flight is available on the specified 
     * date for the specified number of seat.
     * Called by CustomerFlight.checkAvailability.
     *
     * @param conn           Connection 
     * @param travel         The date flying.
     * @param seatsInBlock   The number of seats the airline 
     *                       reserves for JBMSTours on the flight
     * @param kindOfSeats    the name of the column in 
     *                       FlightAvailability to select from
     * @param numberSeatsNeeded    number of seats needed
     * @return true, if the flight is available
     * @exception SQLException in case anything goes 
     *                       wrong during a SQL operation
     */
    public boolean checkSegmentAvailability(Connection conn, 
            Date travel, short kindOfSeats, 
            int numberOfSeatsNeeded) throws SQLException {

            PreparedStatement check_segment_availability = prepareAvailabilityQuery(conn, kindOfSeats);

        check_segment_availability.setString(1, airline);
        check_segment_availability.setString(2, flight_id);
        check_segment_availability.setInt(3, getSegmentNumber());    
        check_segment_availability.setDate(4, travel);

        ResultSet rs = check_segment_availability.executeQuery();

        rs.next();
        int availableSeats = rs.getInt(1);
        check_segment_availability.close();

        if (!(availableSeats >= numberOfSeatsNeeded))
            return false;
        else
            return true;
    }

    /**
     * Prepares the query for checking the availability of the segment
     *
     * @param conn           Connection 
     * @param kindOfSeats    the name of the column in 
     * @return the prepared statement 
     * @exception SQLException in case anything goes 
     *                       wrong during a SQL operation
     */
  protected PreparedStatement prepareAvailabilityQuery(Connection conn, short kindOfSeats) throws SQLException {


    /* even though the statement is prepared for each instance, the statement cache may find exact duplicates, which will allow us to avoid compilation.*/

        String seats = levelToSeats(kindOfSeats);
        String seatstaken = seats + "_taken";

		// the conditional expression allows us to test whether 0 rows are returned by
		// the inner query--convert a null into a 0.
        String query = "Select Block_Seats.SeatsStore - " +
                              "(Seats_taken.seats IS NULL ? 0 : Seats_taken.seats)    FROM " +
                              "(SELECT " + seats + 
                              " FROM Airlines WHERE Airline = ?) " +
                              " Block_Seats(SeatsStore) LEFT OUTER JOIN (SELECT  " +
                                seatstaken +
                              " FROM FlightAvailability " +
                              " WHERE flight_id = ? " +
                              " AND segment_number = ? " +
                              " AND flight_date = ?) Seats_taken(seats) " +
                              "ON SeatsStore >= seats";
 
        PreparedStatement ps = 
        conn.prepareStatement(query);   
       return ps;

  }

    /**
     * Converts a group's level to the kind of seats needed
     *
     * @param     The level of the tour. See the static variables in Tour.java.
     *
     * @return the appropriate column name for the kind of airline seats
     */

  public static String levelToSeats(short l) {
  
		String sts = levelToCursor(l);
        sts = sts + "_seats"; 
	return sts;

  }

    /**
     * Converts a group's level to the kind of seats needed
     *
     * @param     The level of the tour. See the static variables in Tour.java.
     *
     * @return the appropriate column name for the kind of airline seats
     */

  public static String levelToCursor(short l) {
  

        String sts = checkLevel(l);
        if (sts.equals("first class")) 
		sts = "firstclass";
        //sts = sts + "_seats"; 
	return sts;

  }


	
    /**
     * Prepares the PreparedStatements for this class.
     *
     * @return boolean if statements are prepared without error
     *
     * @exception   SQLException   If there is a SQL error
     */

  public boolean prepareStatements(Connection conn) throws SQLException {

           if (insert_flight_bookings == null) 

           insert_flight_bookings = conn.prepareStatement(
            "INSERT INTO FlightBookings (flight_id, "+
            "segment_number, group_id, travel_date, "+
            "number_seats, fixed_rate, travel_level) "+
            "VALUES (?, ?, ?, ?, ?, ?, ?)");

	   if (get_airline_pricing_info == null)
	     get_airline_pricing_info = conn.prepareStatement(
                "SELECT basic_rate, distance_discount, "+
                    "business_level_factor, firstclass_level_factor "+
                 "FROM airlines "+
                 "WHERE airline = ?");


	   if (insert_flight_bookings != null && get_airline_pricing_info != null)
	     return true;
	   else
	     return false;

  }

  public static String checkLevel(short l) {
  
  if (l == Tour.FIRSTCLASSTOURLEVEL)
    return FIRSTCLASSFLIGHTLEVEL;

  if (l == Tour.STANDARDTOURLEVEL)
    return BUSINESSFLIGHTLEVEL;

  if (l == Tour.ECONOMYTOURLEVEL)
    return ECONOMYFLIGHTLEVEL;

  return "unknown level";


  }


}














