package JBMSTours.serializabletypes;

import java.sql.Date;
import java.sql.PreparedStatement;
import java.sql.Statement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.DriverManager;
import java.sql.Connection;
import java.math.BigDecimal;
import JBMSTours.Hotel;
import JBMSTours.Util;
import JBMSTours.JCalendar;
import JBMSTours.Group;
import JBMSTours.AvailabilityException;
import JBMSTours.BudgetException;

/** 
 * Look at the <a href="../JBMSTours/serializabletypes/HotelStay.java">source</a>.<p>
 * Definition of class HotelStay.
 * <p>
 * A HotelStay defines a group's stay at a particular hotel. It includes
 * methods to calculate the price of the stay, to book the stay, and
 * the like. In a Tour, a group has one HotelStay per city (except their
 * home city.)
 * <p>
 * a HotelStay is stored only as part of a Tour object.
 */ 
public class HotelStay implements java.io.Serializable  {

    //////////////////
    // fields
    //////////////////

    public Date arrival;
    public Date departure;

    public String cityName;
    public int cityId;
    public int numberRooms;
    public int numberInParty;
    public short level;
    public int gid;

    /** 
     * Marking a field as transient means that the field is not stored. 
     * We do not need to store theHotel field, since hotels are already 
     * stored in the Hotels table. Storing the hotelId allows us to 
     * retrieve a hotel object if needed. See the 
     * <code>getTheHotel</code> method.
     */
    private transient Hotel theHotel;

    public int hotelId;
    public String hotelName;
    public BigDecimal fixedRate;
    public BigDecimal totalCost;

    /** 
     * We do not need to store the prepared Statements as part of the class,
     * so we mark them transient.
     */
    private transient PreparedStatement find_available_hotel;
    private transient PreparedStatement find_another_available_hotel;
    private transient PreparedStatement get_number_of_rooms_taken;
    private transient PreparedStatement update_number_of_rooms_taken; 
    private transient PreparedStatement insert_into_hotelAvailability;
    private transient PreparedStatement insert_into_hotelbookings;
    private transient PreparedStatement find_me;
  
    static final long serialVersionUID = -7335746095070637734L;
  

    //////////////////
    // constructors
    //////////////////

    /**
     * Create a new HotelStay, initializing it with default/uset values.
     */
    public HotelStay () {

        arrival         = new java.sql.Date(100, 0, 1);
        departure       = new java.sql.Date(100, 0, 1);
        cityName        = "unknown";
        cityId          = 1000;
        numberRooms     = 0;
        numberInParty   = 0;
        level           = 0;
        gid             = 0;
        hotelId         = 0;
        hotelName       = "unknown";
        fixedRate       = new BigDecimal(0);
        totalCost       = new BigDecimal(0);

    }

    /**
     * Constructs a HotelStay group object.
     *
     * <p>
     * A basic constructor taking the information from the passed in objects
     * and initializing the state of the HotelStay object.
     *
     * @param acity           The city to stay in.
     * @param agroup          Description of group (size, number of rooms, ...)
     * @param arrivalFlight   The flight the group is arriving on.
     * @param departureFlight The flight the group is departing on.
     *
     * @exception java.sql.SQLException a error occurred.
     *
     */
    public HotelStay (Connection conn, City acity, Group agroup, 
            CustomerFlight arrivalFlight, 
            CustomerFlight departureFlight) throws java.sql.SQLException {

        arrival       = arrivalFlight.getArrivalDate(conn);
        departure     = departureFlight.getTravelDate();
        cityName      = acity.getName();
        cityId        = acity.city_id;
        numberRooms   = agroup.getNumberRooms();
        numberInParty = agroup.getNumberPeople();
        level         = agroup.getLevel();
        gid           = agroup.getGroupId();
        hotelId       = 0;
        hotelName     = "unknown";
        fixedRate     = new BigDecimal(0);
        totalCost     = new BigDecimal(0);
    }

    /**
     * Constructs a HotelStay group object.
     *
     * <p>
     * A constructor taking the information from the passed in objects
     * and initializing the state of the HotelStay object.
     * Takes dates instead of flights as parameters.
     *
     * @param conn            Connection to the database.
     * @param cid             The id of the city to stay in.
     * @param gid             Id of the group (size, number of rooms, ...)
     * @param arrival          Date the group arrives.
     * @param departure       Date the group departs.
     *
     * @exception java.sql.SQLException a error occurred.
     *
     */
    public HotelStay (Connection conn, int cid, int gid, 
            java.sql.Date arrival, java.sql.Date departure) throws java.sql.SQLException {

        this.arrival       = arrival;
        this.departure     = departure;
        cityName      = City.findCity(conn, cid).getName();
        cityId        = cid;
        Group agroup = Group.getGroup(conn, gid);
        numberRooms   = agroup.getNumberRooms();
        numberInParty = agroup.getNumberPeople();
        level         = agroup.getLevel();
        this.gid      = gid;
        hotelId       = 0;
        hotelName     = "unknown";
        fixedRate     = new BigDecimal(0);
        totalCost     = new BigDecimal(0);
    }

    ///////////////////////
    // private methods
    ///////////////////////

    /**
     * Set the price of the hotel room guaranteed to the customer.
     * <p>
     * This function is called in order to fix the rate, guaranteed to the
     * customer.  As part of this process a message is delivered to 
     * "System.out" indicating the rate charged, and information about the
     * reservation.
     * <p>
     * This method is called by the hotel booking method.
     *
     * @param conn the connection to the database.
     *
     * @exception SQLException thrown on database access failure.
     **/
    private void setPrice(Connection conn) 
        throws java.sql.SQLException {

        fixedRate = checkRate(conn);
        totalCost = fixedRate.multiply(
            new BigDecimal((double) (numberRooms * getNumberOfDays())));

        Util.println("the fixed rate is $" + fixedRate.toString() +
                           " for " +
                           numberRooms +
                           " rooms, for " +
                           getNumberOfDays() +
                           " days.");
    }

    ///////////////////////
    // database accessors
    ///////////////////////

    /** 
     * Suggests a hotel stay.
     * <p>
     * Selects all hotels which can fulfill the reservation request and then
     * returns a hotel which is within the price range of the Group.
     * <p>
     * Results of this query are then looked at one at a time to see if they
     * meet the price requirements of the group.
     * 
     * @param      conn           a Connection object
     * @param      availableFunds the group's available funds for the stay.
     *
     * @exception  SQLException            if a SQLException occurs
     * @exception  BudgetException         if the hotel stay is over budget
     * @exception  AvailabilityException   if the hotel stay is not available.
     *
     */
    public BigDecimal proposeHotelStay(Connection conn, BigDecimal availableFunds) 
         throws SQLException, AvailabilityException, BudgetException {
      
        Util.println();
        Util.println(
            "Proposing hotel stay in " + cityName + " from " + arrival + 
            " to " + departure+ " . . . ");

        prepareStatements(conn);

        // set the variable arguments to the prepared statement.
        find_available_hotel.setInt(1, cityId);
        find_available_hotel.setShort(2, level);                 
        find_available_hotel.setInt(3, numberRooms); 
        find_available_hotel.setDate(4, arrival);
        find_available_hotel.setDate(5, departure);

        ResultSet rs = find_available_hotel.executeQuery();
		return establishTheStay(conn, rs, availableFunds);
	}
    /** 
     * Suggests a hotel stay, given a hotel_id to avoid.
     * <p>
     * Selects all hotels (except the one specified) that can fulfill
	 * the reservation request and then
     * returns a hotel which is within the price range of the Group.
     * <p>
     * Results of this query are then looked at one at a time to see if they
     * meet the price requirements of the group.
     * 
     * @param      conn           a Connection object
     * @param      availableFunds the group's available funds for the stay.
	 * @param      bad_hotel_id   the hotel id to avoid
     *
     * @exception  SQLException            if a SQLException occurs
     * @exception  BudgetException         if the hotel stay is over budget
     * @exception  AvailabilityException   if the hotel stay is not available.
     *
     */
    public BigDecimal proposeHotelStay(Connection conn, BigDecimal availableFunds, int bad_hotel_id) 
         throws SQLException, AvailabilityException, BudgetException {
        Util.println();
        Util.println(
            "Proposing hotel stay in " + cityName + " from " + arrival + 
            " to " + departure+ " . . . ");


		if (find_another_available_hotel == null) 
			find_another_available_hotel = conn.prepareStatement("EXECUTE STATEMENT findAnotherAvailableHotel");
		
        // set the variable arguments to the prepared statement.
        find_another_available_hotel.setInt(1, cityId);
        find_another_available_hotel.setShort(2, level);                 
		find_another_available_hotel.setInt(3, bad_hotel_id);
        find_another_available_hotel.setInt(4, numberRooms); 
        find_another_available_hotel.setDate(5, arrival);
        find_another_available_hotel.setDate(6, departure);

        ResultSet rs = find_another_available_hotel.executeQuery();
		return establishTheStay(conn, rs, availableFunds);
	}

    private BigDecimal establishTheStay(Connection conn,
									   ResultSet rs,
									   BigDecimal availableFunds)
		 throws AvailabilityException, BudgetException, SQLException {
			 boolean withinbudget = false;
			 boolean available = false;
			 theHotel = new Hotel();
			 BigDecimal tempCost = new BigDecimal(0);
			 
			 
			 while (rs.next()) {
				 available = true;
				 theHotel = new Hotel(rs); 
				 
				 Util.println(". . . at the " + theHotel.getName());
				 tempCost = this.checkTotalCost(conn);
				 if (tempCost.compareTo(availableFunds) == -1) {
					 
					 // This hotel fills all requirements of request, including cost.
					 
					 withinbudget = true;
					 hotelId = theHotel.hotelId;
					 
					 break;
				 }
			 }
			 if (available == false)
				 throw new AvailabilityException(this);
			 if (withinbudget == false)
				 throw new BudgetException(this);
			 hotelName = theHotel.getName();
			 return tempCost;             
    }
	
	    /** 
     * Suggests a hotel stay given a hotel_id to avoid, allowing you to temporarily change level.
     * 
     * @param      conn           a Connection object
     * @param      availableFunds the group's available funds for the stay.
	 * @param      templevel  the new level
     *
     * @exception  SQLException            if a SQLException occurs
     * @exception  BudgetException         if the hotel stay is over budget
     * @exception  AvailabilityException   if the hotel stay is not available.
     *
     */
    public BigDecimal proposeHotelStay(Connection conn, BigDecimal availableFunds, short templevel, int bad_hotel_id)
		 
         throws SQLException, AvailabilityException, BudgetException {
			 this.level=templevel;
			 Util.println("Changing hotel level to: " + Tour.getStringForLevel(templevel));
			 return proposeHotelStay(conn, availableFunds, bad_hotel_id);
	}

    /** 
     * Books the hotel stay.
     * <p>
     * This function is responsible for actually booking the hotel stay.  It
     * updates the HotelAvailability and the HotelBookings tables and fixes
     * the rate.
     * <P>
     * It calls the storeBookingInfoForHotel method to update the HotelBookings table.
     * <p>
     * It alters the HotelAvailability table as follows:
     * For each day in the hotel stay, the method finds out whether 
     * there is any data for that particular hotel for that particular day. If so
     * it updates that row. If not, it inserts a row.
     * Using PreparedStatements is most efficient because the method must loop
     * through multiple days. The statements get prepared only once, even though
     * they are executed multiple times.
     * <p>
     * query to check if any rooms for this hotel have been booked on a given date:
     * <p>
     *
     *     SELECT rooms_taken FROM HotelAvailability 
     *        WHERE hotel_id = ? AND booking_date = ? 
     *
     * <P>
     * If there is already a row for this hotel and date, update the "rooms_taken" column
     * (increase it by the number of rooms in this hotelstay) 
     * <p>
     *     UPDATE HotelAvailability SET rooms_taken = (rooms_taken + ?) 
     *         WHERE hotel_id = ? AND booking_date = ?
     *
     * <P>
     * If there are no rows for this hotel and date, add a row showing the number
     * of rooms taken in this hotelstay
     * <p>
     *     INSERT INTO HotelAvailability (hotel_id, booking_date, rooms_taken) 
     *         VALUES (?, ?, ?, ?)
     * 
     * @param      conn           The current database connection.
     *
     * @exception  SQLException   if a SQLException occurs
     *
     */
    public BigDecimal bookHotel(Connection conn) 
		 throws SQLException {
			 
			 Util.println("");
			 Util.println("Booking hotel stay at " + theHotel.getName());
			 int n = 0;
			 prepareStatements(conn);
			 
			 // guarantee the current price to the customer. 
			 setPrice(conn);
			 
			 // store the booking in the database.
			 if (storeBookingInfoForHotel(conn) == 1)
            Util.println(this.toString() + " booked ");


        ResultSet rs;
        // For each day, check whether there are any bookings. If not, 
        // do an insert.  If there are bookings already, do an update.

        Date checkDate = arrival;

        while (checkDate.before(departure)) {

            get_number_of_rooms_taken.setInt(1, getTheHotel(conn).hotelId);
            get_number_of_rooms_taken.setDate(2, checkDate);
            rs = get_number_of_rooms_taken.executeQuery();

            if (rs.next()) {
                // bookings exist - update hotel booking with reservation
                
                // set the variable arguments to the prepared statement.
                update_number_of_rooms_taken.setInt(1, numberRooms);
                update_number_of_rooms_taken.setInt(
                    2, getTheHotel(conn).hotelId);
                update_number_of_rooms_taken.setDate(3, checkDate);

                // Execute the prepared statement.
                n += update_number_of_rooms_taken.executeUpdate();   
            }
            else {
                // bookings don't exist - insert current reservation
       

                // set the variable arguments to the prepared statement.
                insert_into_hotelAvailability.setInt(
                    1, getTheHotel(conn).hotelId);
                insert_into_hotelAvailability.setDate(2, checkDate);
                insert_into_hotelAvailability.setInt(3, numberRooms);
                 n += insert_into_hotelAvailability.executeUpdate();
            }

            checkDate = JCalendar.incrementDate(checkDate);
        }

        Util.println(
            "HotelAvailability table updated -- " + n + " rows changed.");

        // don't commit inside the method to 
        //allow it to be used as a server-side method or replication
        // work unit.
        //conn.commit();

        return totalCost;
    }

    /** 
     * Return the hotel where the group is staying.
     * <p>
     * Return the hotel where the group is staying.  This information may
     * either be stored in the object or may require a database lookup.
     * The database lookup executes the following SQL query:
     * <p>
     *     SELECT Hotel from Hotels WHERE hotel_id = ?
     * 
     * @param      conn           The current database connection.
     *
     * @exception  SQLException   if a SQLException occurs
     *
     */
    public Hotel getTheHotel(Connection conn) throws SQLException {

        if (theHotel != null) {
            return theHotel;
        }
        else { 
	    prepareStatements(conn);

            // set the variable arguments to the prepared statement.
            find_me.setInt(1, hotelId);

            // execute the select getting the hotel row.
            ResultSet rs = find_me.executeQuery(); 
            while (rs.next()) {

                theHotel = new Hotel(rs); 
            }

            
            find_me.close();
        }

        return theHotel;
    }

    /** 
     * Store the booking information into the database.
     * <p>
     * Insert a new row into the HotelBookings table which represents this
     * HotelStay.  This is done with the following SQL statement:
     * <p>
     *     INSERT INTO HotelBookings VALUES (?, ?, ?, ?, ?)
     * 
     * @param      conn           The current database connection.
     *
     * @return The number of rows inserted into the HotelBookings table, 1 row
     *         indicates success, anything else is failure.
     *
     * @exception  SQLException   if a SQLException occurs
     */
    public int storeBookingInfoForHotel(Connection conn) 
        throws SQLException {

        // set the variable arguments to the prepared statement.
        insert_into_hotelbookings.setInt(1, hotelId);
        insert_into_hotelbookings.setInt(2, gid);
        insert_into_hotelbookings.setDate(3, arrival);
        insert_into_hotelbookings.setDate(4, departure);
        insert_into_hotelbookings.setInt(5, numberRooms);
        insert_into_hotelbookings.setObject(6, totalCost);

        int rows_inserted = insert_into_hotelbookings.executeUpdate();
        if (rows_inserted > 0)
            Util.println("Updated HotelBookings table.");

        return rows_inserted;
    }

    /** 
     * Create the History database and a HotelBookings table for it.
     * <p>
     * The HotelBookings table contains the history of all bookings made.  There
     * is one row per booking made.  This routine connects to the History 
     * database, * creating the database in the process, it does this by 
     * connecting to the following url:
     * <p>
     *    "jdbc:cloudscape:History;create=true;autocommit=false"
     * <p>
     * Once the database is created, the HotelBookings table is created with
     * the following SQL statement:
     * <p>
     *     CREATE TABLE HotelBookings 
     *         (hotel_id INT, group_id INT, arrival DATE, departure DATE, 
     *          number_rooms INT, CONSTRAINT HOTELBOOKINGS_PK PRIMARY KEY
     *          (hotel_id, group_id, arrival))
     *
     * @exception  SQLException   if a SQLException occurs
     */
    public static void makeHistoryDatabase() throws SQLException {
		
        Connection conn1 = 
            DriverManager.getConnection(
										"jdbc:cloudscape:History;create=true;autocommit=false");
		
        Statement s1 = conn1.createStatement();
		
		try {
			s1.execute(
					   "CREATE TABLE HotelBookings " + 
					   "(hotel_id INT, group_id INT, arrival DATE, departure DATE, " +
					   "number_rooms INT, total_cost DECIMAL(10,2), " +
					   "CONSTRAINT HOTELBOOKINGS_PK PRIMARY " +
					   "KEY (hotel_id, group_id, arrival))");
			s1.execute("CREATE BTREE INDEX HotelDates2 ON HotelBookings (arrival)");
			
			
			
		}
		
		/* catch any exceptions from this statement, 
           because we will need to close the  connection*/
        catch (SQLException e) {
			//if the database already exists and that table already exists,
			// we'll get a SQLException with state X0x08. if so,
			// just ignore the error.
			if (!e.getSQLState().equals("X0X08")) {
				Util.println("exception thrown:");
				conn1.close();
				throw e;
			}
        }
		
		conn1.commit();
		conn1.close();
    }
	
    /** 
     * Move records of reservations into the History Database.
     * <p>
     * This method is an example of how you can connect to two different 
     * databases at once.  You can execute this method as an application-side
     * or as a server-side method.
     * <p>
     * In this example, the old bookings are selected from the HotelBookings
     * table and inserted into the History database. The method must handle
     * the results.<p>
     * Execute the method
     * makeHistoryDatabase() if the database doesn't exist yet.
     * Since the actions span two connections, they exist within two
     * separate transactions and are thus not protected by transaction
     * control.<p>
     * See archiveRecordsReadWriteVTI for an example of method in which
     * Cloudscape handles
     * the results (use the Cloudscape VirtualTableInterface feature).
	 * Note how much simpler that method is.
     *
     * @param      theDate        The date up to which to archive records.
     *
     * @exception  SQLException   if a SQLException occurs
     */

    public static void archiveRecords(java.sql.Date theDate) 
        throws SQLException {

	/* get a new connection to toursDB*/
        Connection source_connection =
            DriverManager.getConnection(
                "jdbc:cloudscape:toursDB;autocommit=false");

        Connection destination_connection = 
            DriverManager.getConnection(
                "jdbc:cloudscape:History;autocommit=false");

        PreparedStatement get_source_hotel_bookings = 
            source_connection.prepareStatement(
                "SELECT hotel_id, group_id, arrival, departure, number_rooms, total_cost " +
                    "from HotelBookings WHERE arrival < ? "+
				"ORDER BY arrival");
        get_source_hotel_bookings.setDate(1, theDate);

        ResultSet result_of_get_source_hotel_bookings = 
            get_source_hotel_bookings.executeQuery();

        PreparedStatement insert_into_destination_table = 
            destination_connection.prepareStatement(
                "INSERT INTO HotelBookings " + 
                    "(hotel_id, group_id, arrival, departure, number_rooms, total_cost)" + 
                "VALUES (?, ?, ?, ?, ?, ?)");

        // Loop through the records in source database representing
        // HotelStay's before "theDate", and for each record insert a
        // corresponding record in the destination database.

        Util.println("archiving some records from hotelbooking table.");

        while (result_of_get_source_hotel_bookings.next()) {

            // set the variable arguments to the prepared statement.
            insert_into_destination_table.setInt(
                1, result_of_get_source_hotel_bookings.getInt(1));
            insert_into_destination_table.setInt(
                2, result_of_get_source_hotel_bookings.getInt(2));
            insert_into_destination_table.setDate(
                3, result_of_get_source_hotel_bookings.getDate(3));
            insert_into_destination_table.setDate(
                4, result_of_get_source_hotel_bookings.getDate(4));
            insert_into_destination_table.setInt(
                5, result_of_get_source_hotel_bookings.getInt(5));
			insert_into_destination_table.setBigDecimal(
				6, result_of_get_source_hotel_bookings.getBigDecimal(6, 6));
			
            // execute the insert
            insert_into_destination_table.executeUpdate();
        }

        result_of_get_source_hotel_bookings.close();

        // Dump to System.out list of dates archived to the HotelBookings table.
        ResultSet result_of_dest_hotelbookings_select = 
             destination_connection.createStatement().executeQuery(
                 "Select * from HotelBookings ORDER BY arrival");

        while (result_of_dest_hotelbookings_select.next()) {
           Util.println(result_of_dest_hotelbookings_select.getDate(3));
        }
        result_of_dest_hotelbookings_select.close();

        // Delete the HotelBookings which were archived to the history database
        PreparedStatement delete_source_hotel_bookings = 
            source_connection.prepareStatement(
                "Delete from HotelBookings WHERE arrival < ?");

        delete_source_hotel_bookings.setDate(1, theDate);
        delete_source_hotel_bookings.executeUpdate();

        // commit transactions and close down database connections.
        destination_connection.commit();
        source_connection.commit();
        destination_connection.close();
        source_connection.close();
    }



    /** 
     * Move records of reservations into the History Database.
     * <p>
     * This method is an example of using a class that implements Cloudscape's Read-Write
	 * virtual table interface to create an
     * ExternalVirtualTable. We select rows from HotelBookings and insert them into
	 * the external table.
     *
	 * @param      conn           A connection to the database.
     * @param      theDate        The date up to which to archive records.
     *
     * @exception  SQLException   if an SQLException occurs
     */
	
    public static void archiveRecordsReadWriteVTI(Connection conn, java.sql.Date theDate) 
		 throws SQLException {

		String externalinsert = "INSERT INTO NEW JBMSTours.vti.jdbc1_2.ExternalCloudscapeTable( " +
			"'jdbc:cloudscape:History', " +
			"'HotelBookings') SELECT *  FROM HotelBookings WHERE arrival < ?";
		String internaldelete = "DELETE FROM HotelBookings WHERE arrival < ?";
		PreparedStatement ps = conn.prepareStatement(externalinsert);
		ps.setDate(1, theDate);
		int numRecordsInserted = ps.executeUpdate();
		ps=conn.prepareStatement(internaldelete);
		ps.setDate(1, theDate);
		int numRecordsDeleted = ps.executeUpdate();
		System.out.println(numRecordsInserted + " records inserted into History database.");
		System.out.println(numRecordsDeleted + " records deleted from toursDB.");
		ps.close();
    }
	



	
    //////////////////////////////////
    // state modifiers or accessors
    //////////////////////////////////
    
    /**
     * Check the rate of the current hotel stay. It does not fix the rate.
     * <p>
     * This method gets the current rate from the hotel object.  If the stay
     * is during the the "high season", this routine prints a message to 
     * "System.out" indicating the high season rate.
     * <p>
     *
     * @return The current rate from the hotel object.
     *
     * @param conn  The current SQL connection.
     *
     * @exception  java.sql.SQLException Standard SQL exception thrown on error.
     **/
    public BigDecimal checkRate(Connection conn) 
        throws java.sql.SQLException {

        BigDecimal ratePerDay;

        if (getTheHotel(conn).isHigh(arrival, departure)) {
            ratePerDay = theHotel.getHighRate();

            Util.println(
                "(You must pay the high-season rate for this hotel.)");
        }
        else {
            ratePerDay = theHotel.getNormalRate();
        }

        return ratePerDay;
    }

    /**
     * Returns the total cost of the group's stay at the hotel.
     *
     * @return The total cost of the group's stay at the hotel.
     *
     * @exception  java.sql.SQLException Standard SQL exception thrown on error.
     **/
    public BigDecimal checkTotalCost(Connection conn) 
        throws SQLException {
        return (checkRate(conn).multiply(new BigDecimal((double) numberRooms * getNumberOfDays())).setScale(2, BigDecimal.ROUND_DOWN));

    }

    /**
     * Return number of days of the current Hotel Stay.
     *
     * @return The number of days of the current hotel stay.
     *
     **/
    public int getNumberOfDays() {

        int difference = 0;
        if (departure.after(arrival) || departure.equals(arrival)) {
            difference = JCalendar.getDateDifference(departure,arrival);
        }

        return difference;
    }

    /**
     * Gets info about a HotelStay..
     *
     * @return  String   object information about the HotelStay 
     *
     * @exception  java.sql.SQLException Standard SQL exception thrown on error.
     *
     **/
    public String getInfo(Connection conn) throws SQLException {

        StringBuffer hotelInfo = new StringBuffer();

        hotelInfo.append(getTheHotel(conn).getInfo());
        hotelInfo.append("\nYou are arriving on " + arrival + " and leaving on " + departure);

        hotelInfo.append("\nNumber of rooms booked: " + numberRooms);
        hotelInfo.append("\nYour guaranteed rate per room is: $" + fixedRate);
        hotelInfo.append("\nAt a total cost of: $" + totalCost);
        return hotelInfo.toString();
    }

    /**
     * Convert the object to a string. Supplies the arrival and departure
     * dates, hotel name, and city name.
     *
     * @return string for the group object.
     */
    public String toString() {
        return("Hotel booking from " + arrival + " to " + departure +
               " at the " + hotelName + " in " + cityName);
    }



    /**
     * Prepares all the PreparedStatements for this class.
     *
     * @return boolean if statements are prepared without error
     *
     * @exception   SQLException   If there is a SQL error
     */
  public boolean prepareStatements(Connection conn) throws SQLException {

    if (find_available_hotel == null) 
        find_available_hotel = conn.prepareStatement("EXECUTE STATEMENT findAvailableHotel");
    
    

	if (get_number_of_rooms_taken == null)
        get_number_of_rooms_taken = 
            conn.prepareStatement(
                "SELECT rooms_taken FROM HotelAvailability " + 
                    "WHERE hotel_id = ? AND booking_date = ? ");

	if (update_number_of_rooms_taken == null)
        update_number_of_rooms_taken = 
            conn.prepareStatement(
                "UPDATE HotelAvailability " + 
                "SET rooms_taken = (rooms_taken + ?) " +
                    "WHERE hotel_id = ? AND booking_date = ?");

	if (insert_into_hotelAvailability == null)
        insert_into_hotelAvailability = 
            conn.prepareStatement(
                "INSERT INTO HotelAvailability " + 
                    "(hotel_id, booking_date, rooms_taken) " +
                "VALUES (?, ?, ?)");
        if (insert_into_hotelbookings == null)
	insert_into_hotelbookings = conn.prepareStatement(
                    "INSERT INTO HotelBookings VALUES (?, ?, ?, ?, ?, ?)");

            find_me = 
                conn.prepareStatement(
                    "SELECT * from hotels WHERE hotel_id = ?");
        


	if (find_available_hotel != null && get_number_of_rooms_taken != null && update_number_of_rooms_taken != null && insert_into_hotelAvailability != null && find_me != null)
	  return true;
	else
	  return false;

  }

   /**
     * @return    boolean        if the two objects are equals.
     *
	 * @exception ClassCastException if the passed-in object is not a HotelStay
     */
	public boolean equals(Object o) throws ClassCastException {
        HotelStay ohs = (HotelStay) o;

        if (this.arrival== ohs.arrival &&
			this.departure == ohs.departure &&
			this.cityId == ohs.cityId &&
			this.gid == ohs.gid &&
			this.hotelId == ohs.hotelId &&
			this.fixedRate == ohs.fixedRate)
			return true;
		else
			return false;
	}


	

}
