package JBMSTours.serializabletypes;

import java.sql.*;
import java.io.Serializable;
//import java.lang.Comparable;
import java.lang.ClassCastException;
import JBMSTours.Util;
import JBMSTours.JCalendar;


/**
 * Look at the <a href="../JBMSTours/serializabletypes/Location.java">Source</a>.<p>
 * The Location class includes stores longitude and latitude
 * information. In this application, part of the City object.
 *
 *
 * @see City
 */
public class Location implements Serializable {

    ///////////
    // fields
    ///////////

    private int latitudeDegrees;
    private int latitudeMinutes;

    /**
     * latitude direction: N or S
     */
    private String latitudeDir;
    private int longitudeDegrees;
    private int longitudeMinutes;

    /**
     * longitude direction: E or W
     */
    private String longitudeDir;

    static final long serialVersionUID = 8430573461302288890L;
	

    /////////////////
    // constructors
    /////////////////

    /**
     * Construct a Location at 0N 0E.
     */
    public Location () {
        latitudeDegrees = 0;
        latitudeMinutes = 0;
        latitudeDir = "N";
        longitudeDegrees = 0;
        longitudeMinutes = 0;
        longitudeDir = "E";        
    }

    /**
     * Constructs a Location object.
     *
     * @param     latDeg      degrees latitude
     * @param     latMin      minutes latitude
     * @param     latDir      "N" or "S" for  latitude
     * @param     longDeg     degrees longitude
     * @param     longMin     minutes longitude
     * @param     longDir     "E" or "W" for  latitude
     */ 
    public Location (
            int latDeg, int latMin, String latDir,
            int longDeg, int longMin, String longDir) {

        latitudeDegrees = latDeg;
        latitudeMinutes = latMin;
        latitudeDir = latDir;
        longitudeDegrees = longDeg;
        longitudeMinutes = longMin;
        longitudeDir = longDir;        
    }

    ////////////////////
    // toString value
    ////////////////////

    /**
     * Convert the object to a string.
     *
     * @return string representation of the object.
     */
    public String toString() {
        return latitudeDegrees + " degrees, " 
             + latitudeMinutes + " minutes, " + latitudeDir + ", " 
             + longitudeDegrees + " degrees, " 
             + longitudeMinutes + " minutes " + longitudeDir;
    }


	public int getLatitudeDegrees() {

		return latitudeDegrees;
	}
	public int getLatitudeMinutes() {

		return latitudeMinutes;
	}
	public int getLongitudeDegrees() {

		return longitudeDegrees;
	}
	public int getLongitudeMinutes() {

		return longitudeMinutes;
	}

	public String getLatitudeDir() {

		return latitudeDir;
	}

	public String getLongitudeDir() {

		return longitudeDir;
	}


    ///////////////////////
    // conversion methods
    ///////////////////////

    /**
     * Convert the latitude degrees and minutes to a decimal value.
     *
     * @return the latitude as a double
     */
    private double getDecimalLatitude() {

        double lat = (double) latitudeDegrees
                   + minutesToDecimal(latitudeMinutes);

        if (latitudeDir.equals("S"))
            lat = 0 - lat;

        return lat; 
    }

    /**
     * Convert the longitude degrees and minutes to a decimal value.
     *
     * @return the longitude as a double
     */
    private double getDecimalLongitude() {

        double lon = (double) longitudeDegrees 
                   + minutesToDecimal(longitudeMinutes);

        if (longitudeDir.equals("W"))
            lon = 0 - lon;

        return lon;
    }

    /**
     * Get the mathematical sin value of the latitude
     *
     * @return the latitude as a sin value.
     */
    private double getSinLatitude() {
        return Math.sin(degreesToRadians(getDecimalLatitude()));
    }

    /**
     * Get the mathematical cos value of the latitude
     *
     * @return the latitude as a cos value.
     */
    private double getCosLatitude() {
        return Math.cos(degreesToRadians(getDecimalLatitude()));
    }

    /**
     * To compare two locations, get the cosine of the difference
     * of this location from the other location.
     *
     * @param b the other location
     * @return the cosine of the difference of two locations.
     */
    private double getCosLongDif(Location b) {
        double dif, cosDif;
        dif = this.getDecimalLongitude() - b.getDecimalLongitude();
        cosDif = Math.cos(degreesToRadians(dif));
        return cosDif;
    }

    ///////////////////////
    // comparison methods
    ///////////////////////

    /**
     * Determine the distance (in miles) between this location and another one.
     * Computes the distance using the great circle formula, which is not entirely
     * precise but reasonably accurate. 69.16 is the number of miles in a degree.
     * 
     * @param b the other location
     */
    public int getDistance(Location b) {

        double cosDistance, acosDistance;
        int distance;

        cosDistance = (this.getSinLatitude() * b.getSinLatitude()) 
                    + (this.getCosLatitude() * b.getCosLatitude() 
                      * this.getCosLongDif(b));

        acosDistance = radiansToDegrees(Math.acos(cosDistance));

        distance = (int) (acosDistance * 69.16);

        return distance;
    }

    /**
     * Determine if the location is tropical or not.
     *
     * @return true if the location is tropical.
     */
    public boolean isTropical() {
        return (latitudeDegrees <= 24);
    }

    /**
     * Get the decimal latitude for comparing locations by
     * north/south relative location.
     *
     * @return the latitude as a comparable value
     */
    public double getNorthSouthOrderableLocation() {
        return getDecimalLatitude();
    }

    /**
     * Get the decimal longitude for comparing locations by
     * east/west relative location.
     *
     * @return the longitude as a comparable value
     */
    public double getEastWestOrderableLocation() {
        return getDecimalLongitude();
    }

    //////////////////////////
    // static helper methods
    //////////////////////////

    /**
     * Convert a minutes measurement to a fraction.
     *
     * @return the minutes as a fractional value.
     */
    private static double minutesToDecimal(int mins) {
        return (double) mins/60;
    }

    /**
     * Convert a degree measurement to radians.
     *
     * @return the degrees as radians.
     */
    private static double degreesToRadians(double degrees) {
        return (degrees * Math.PI)/180;
    }

    /**
     * Convert a radians measurement to degrees.
     *
     * @return the radians as degrees.
     */
    private static double radiansToDegrees(double radians) {
        return (radians * 180) / Math.PI;
    }

	


	/**
     * This method allows a Location object to be <i>orderable</i> within SQL-J
	 * statements. <p>
	 * Returns -1 if this object is "less than" the parameter, 1 if this object
	 * is "greater than" the parameter, and "0" if they are equal.<p>
	 * We have arbitrarily decided to order Locations by their latitudes (from North to South),
	 * and secondarily by their longitudes (from East to West).
     *
     * @exception ClassCastException  Thrown if the objects cannot be compared.
     *
	 * @see City
     *
     */

	public int compareTo(Object o) throws ClassCastException {
		//handle nulls;
		// nulls sort lower than everything
		if (o == null)
			return 1;
		// may throw an exception
        Location ol = (Location) o;
		double latvalue1 = this.getNorthSouthOrderableLocation();
		double latvalue2 = ol.getNorthSouthOrderableLocation();
		double longvalue1 = this.getEastWestOrderableLocation();
		double longvalue2 = ol.getEastWestOrderableLocation();

		if (latvalue1 == latvalue2 && longvalue1 == longvalue2)
			return 0;
		// order the objects first by their latitudes
		if (latvalue1 > latvalue2)
			return 1;
		if (latvalue1 < latvalue2)
			return -1;
		//if their latitudes are equal, then compare their longitudes
		if (longvalue1 > longvalue2)
			return 1;
		//the only other possibility is that longvalue1 < longvalue2
		
		else
			return -1;
	}
   /**
     * For correct = comparisons.
     *
     */
	public boolean equals(Object o)  {
		Location ol;
		try {
			ol = (Location) o;
		}
		catch (ClassCastException cce) {
			return false;
		}
		if (this.compareTo(ol)==0)
			return true;
		else
			return false;
	}

	/* for correct equality comparisons
	 *
     *
	 * @return int   a hash code
	 *
	 */
	 
	public int hashCode()   {
		double latvalue1 = Math.abs(this.getNorthSouthOrderableLocation()*100);
		double longvalue1 = Math.abs(this.getEastWestOrderableLocation()*100);
		int lat = (int) latvalue1;
		int lon = (int) longvalue1;
		String s1 = new Integer(lat).toString();
		String s2 = new Integer(lon).toString();
		String s3 = s1+s2;
		return new Integer(s3).intValue();

	}

	
	

}
