package JBMSTours.serializabletypes;

import java.sql.*;
import java.math.BigDecimal;
import JBMSTours.Util;
import JBMSTours.JCalendar;


/**
  * Look at the <a href="../JBMSTours/serializabletypes/OneStopFlight.java">Source</a>.<p>
  * Definition of a flight with one stop.
  *
  */ 
public class OneStopFlight    extends        Flight
                            implements    java.io.Serializable {

    /////////////////////////////////////////////////////////////////////////
    //
    //    STATE
    //
    /////////////////////////////////////////////////////////////////////////

 
    public int      segment2Number;
    public String   stop_airport;
    public Time     stopArriveTime;
    public Time     stopDepartTime;
    public transient PreparedStatement insert_stop_flight_bookings;			      
    static final long serialVersionUID = 8831183975706341087L;


    /////////////////////////////////////////////////////////////////////////
    //
    //    CONSTRUCTORS
    //
    /////////////////////////////////////////////////////////////////////////

 
    /**
     * Default niladic constructor for a flight with one stop.
     *
     */
      
    public OneStopFlight() {

        super();

        segment2Number= 0;
        stop_airport = "unknown";
        stopArriveTime = new Time(0, 0, 0);
        stopDepartTime = new Time(0, 0, 0);
    }


  /**
    * Constructs a one-stop flight from its two legs.
    *
    * @param segone     Flight object for first flight.
    * @param segtwo     Flight object for second flight.
    *
    */

    public OneStopFlight (Flight segone, Flight segtwo) {

        super( segone.flight_id,
               segone.segmentNumber,
               segone.orig_airport,
               segone.departTime,
               segtwo.dest_airport,
               segtwo.arriveTime,
               segone.flyingTime + segtwo.flyingTime,
               segone.mileage + segtwo.mileage );

        segment2Number = segtwo.segmentNumber;
        stop_airport = segone.dest_airport;
        stopArriveTime = segone.arriveTime;
        stopDepartTime = segtwo.departTime;
    }


    /////////////////////////////////////////////////////////////////////////
    //
    //    SHADOW OBJECT METHODS
    //
    /////////////////////////////////////////////////////////////////////////

    /**
     * Prints out this object.
     *
     */

    public String toString() {
        return    getType() +
                ", flight " + flight_id +  ", " +
                orig_airport + " to " + dest_airport +
                " with a stop at " + stop_airport;
    } 


    /////////////////////////////////////////////////////////////////////////
    //
    //    ACCESSORS
    //
    /////////////////////////////////////////////////////////////////////////

    /**
      *    Get the segment number of the second leg of this Flight.
      *
      *    @return    segment number of the second leg of this Flight.
      */
    public int getAltSegmentNumber() { 
        return segment2Number; 
    }


    /**
      *    Get the duration, in hours, of the layover.
      *
      *    @return    the duration, in hours, of the layover.
      */
    public double getLayoverTime() { 
        return JCalendar.getTimeDifference(stopArriveTime, stopDepartTime);
    }
   
    ///////////////////////////////
    //                           //
    //    SHADOW FLIGHT ACCESSORS  //
    //                           //
    ///////////////////////////////


    /**
      *    Get the flight id of the second leg.
      *
      *    @return    flight id of the second leg.
      */
    public String getAltFlightId() { 
        return flight_id; 
    }


    /**
      *    Get the Total Duration of the flight, including layover Time
      *
      *    @return    flight's duration.
      */
    public double getTotalDuration() { 
        return flyingTime + getLayoverTime(); 
    }

    /**
      *    Get the flight type.
      *
      *    @return    flight type.
      */
    public String getType() {  
        return "One-stop flight"; 
    } 


  public String getUniqueId() {
  
    return getFlightId() + orig_airport + stop_airport + dest_airport;
  }


    /////////////////////////////////////////////////////////////////////////
    //
    //    SHADOW OTHER FLIGHT METHODS
    //
    /////////////////////////////////////////////////////////////////////////


    /**
     * "Books" the flight by extracting the pertinent information to 
     * a separate table (FlightBookings). This is the information 
     * that eventually goes to the airline.
     *
     * @param conn        Connection 
     * @param gid         The group_id
     * @param travelDate  Date that travel begins.
     * @param numberSeats number of seats reserved
     * @param fixedRate   the rate for those seats
     * @param     The level of the tour. See the static variables in Tour.java.
     *
     * @return    > 0 if FlightBookings table successfully updated.
     *    
     * @exception SQLException in case anything goes wrong during a SQL operation
     */

    public    int    storeBookingInfoForAirline
    (
        Connection    conn,
        int         gid,
        Date        travelDate,
        int         numberSeats,
        BigDecimal    fixedRate,
        short         level
    )
        throws SQLException
    {
   
      prepareStatements(conn);
    
        insert_stop_flight_bookings.setString(1, flight_id);
        insert_stop_flight_bookings.setInt(2, getSegmentNumber());
        insert_stop_flight_bookings.setInt(3, gid);
        insert_stop_flight_bookings.setDate(4, travelDate);
        insert_stop_flight_bookings.setString(5, flight_id);
        insert_stop_flight_bookings.setInt(6, segment2Number);
        insert_stop_flight_bookings.setInt(7, numberSeats);
        insert_stop_flight_bookings.setObject(8, fixedRate);
        insert_stop_flight_bookings.setString(9, checkLevel(level));

        int n = insert_stop_flight_bookings.executeUpdate();

        return n;
    }


    /**
     * Checks whether the flight is available on the specified date for the specified
     * number of seats.
     *
     * Called by CustomerFlight.checkAvailability().
     *
     * @param conn           Connection 
     * @param travel         The date flying.
     * @param kindOfSeats    the name of the column in FlightAvailability to select from
     * @param numberSeatsNeeded    number of seats needed
     * @return true, if the flight is available
     * @exception SQLException in case anything goes wrong during a SQL operation
     */

    public boolean checkSegmentAvailability
    (
        Connection    conn,
        Date         travel,
        short         kindOfSeats,
        int            numberOfSeatsNeeded
    )
        throws SQLException
    {

        PreparedStatement check_segment_availability = prepareAvailabilityQuery(conn, kindOfSeats);

        check_segment_availability.setString(1, airline);
        check_segment_availability.setString(2, flight_id);
        check_segment_availability.setInt(3, getSegmentNumber());    
        check_segment_availability.setDate(4, travel);

        ResultSet rs = check_segment_availability.executeQuery();

        rs.next();
        int availableSeats = rs.getInt(1);
       
        if (!(availableSeats >=numberOfSeatsNeeded))
            return false;


        check_segment_availability.setString(1, airline);
        check_segment_availability.setString(2, flight_id);
        check_segment_availability.setInt(3, getAltSegmentNumber());    
        check_segment_availability.setDate(4, travel);

        rs = check_segment_availability.executeQuery();
        rs.next();

        availableSeats = rs.getInt(1);
        check_segment_availability.close();
        if (!(availableSeats >=numberOfSeatsNeeded))
            return false;

        else return true;
        
    }

    /**
     * Prepares the PreparedStatements for this class.
     *
     * @return boolean if statements are prepared without error
     *
     * @exception   SQLException   If there is a SQL error
     */

  public boolean prepareStatements(Connection conn) throws SQLException {

           super.prepareStatements(conn);

           if (insert_stop_flight_bookings == null) 

           insert_stop_flight_bookings = 
	     conn.prepareStatement("INSERT INTO FlightBookings (" +
				   "flight_id, segment_number, group_id, travel_date, " +
				   "segment_two_flight_id, segment_two_segment_number, " +
				   "number_seats, fixed_rate, travel_level) " +
				   "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)");


	   if (insert_stop_flight_bookings != null)
	     return true;
	   else
	     return false;
  }

}









