package JBMSTours.serializabletypes;

import java.io.Externalizable;
import java.io.ObjectOutput;
import java.io.ObjectInput;
import java.io.IOException;
import java.sql.*;
import java.util.Calendar;
import java.util.GregorianCalendar;
import JBMSTours.Util;
import JBMSTours.JCalendar;
import JBMSTours.ApplicationMode;


/**
 * Look at the <a href="../JBMSTours/serializabletypes/Person.java">Source</a>.<p>
 * A Person is stored in the person column of the <em>People</em> table.
 * Usually only subclasses (Adult and Child) are stored.
 * A Person is a member of a Group.
 * <p>These classes illustrate how you can
 * store instances of subclasses in a column defined to store
 * the superclass. (In reality, these classes are probably too
 * simple to store as objects--probably it is better to store the fields
 * as simple data types.)x
 * <p>Cloudscape automatically assigns the integer value in the person_id column
 * (DEFAULT AUTOINCREMENT) whenever a row is inserted. This class uses the
 * <a href="#updatePersonWithAutoincrementValue">updatePersonWithAutoincrementValue</a>
 * method to find out what that value was and then to update the Person object to
 * assign that value to <code>this.id</code>.
 *
 * BuildATour constructs the Persons.
 *
 * @see Adult
 * @see Child
 * @see JBMSTours.Group
 * @see JBMSTours.BuildATour
 *
 */

public class Person implements Externalizable {

    public int id=0;
    protected String firstName;
    protected String lastName;
    protected char middleInitial;
    protected Date birthdate;
	private static transient PreparedStatement find_person;
    private static final int currentClassVersion = 1;

    public static final long serialVersionUID = 6968972189667483941L;

    public Person () {
        firstName = "unknown";
        lastName = "unknown";
		nullOutPreparedStatements();

    }

    /**
     * Constructor.
     * Takes a first name and a last name.
     *
     * @param first    the first name.
     * @param last     the last name.
     */
    public Person (String first, String last) {
        this(first, last, ' ');

    }

    /**
     * Constructor.
     * Takes a first name, a last name, and a middle initial.
     * @param first     thefirst name.
     * @param last      the last name.
     * @param a         the middle initial.
     */
    public Person (String first, String last, char a) {
        firstName = first;
        lastName = last;
        middleInitial = a;
		nullOutPreparedStatements();
		
    }

	private static void nullOutPreparedStatements() {
		find_person = null;
	}

    /**
     * Sets the person's age.
     *
     * @param bd   birth date.
     */
    public void setBirthdate(Date bd) {
        birthdate = bd;
    }

    /**
     * Gets the full name.
     */

    public String getName() {
        String theName;
        theName = this.firstName + " " + this.lastName;
        return theName;
    }

    /**
     * Gets the first name.
     */

    public String getFirstName() {
        return firstName;
    }

    /**
     * Gets the last name.
     */

    public String getLastName() {
        return lastName;
    }


	public static boolean prepareStaticStatements(Connection conn) throws SQLException {
		find_person = conn.prepareStatement("SELECT person " +
                            "FROM People " +
                            "WHERE person_id = ?"); 
		
		return true;
		
	}


    /**
     * Does a system out of the name.
     */
    public void printName() {

        System.out.println("");
        System.out.println("Name is " + getName());
        System.out.println("");
    }

    /**
     * Sets the person's ID.
     *
     * @param i    the ID.
     */
    public void setId(int i) {

        id = i;
    }

    /**
     * Gets the person's ID.
     *
     * @return int    The ID.
     */
    public int getId() {
        return id;
    }

    /**    
     * Gets the type (person, child, or adult).
     */
    //public String getType() {
	//return "JBMSTours.serializabletypes.Person";
    //} 

    /**
     * Gets the age.
     */
    public int getAge() {
				Date now = JCalendar.getToday();
				return JCalendar.getDateDifference(now, birthdate)/365;
    }

    /**
     * Returns getName()
     */
    public String toString() {
        return getName();
    }


    /**
     * Queries the People table to return the appropriate person 
     * object based on the person's id.
     *
     *
     * @param     conn       Connection 
     * @param     personId   id of person
     *
     * @return    Person object
     *
     * @exception SQLException  Thrown if there is a SQL error.
     */
    public static Person findPerson(Connection conn,
                                int personId)
            throws SQLException {

        Person myperson = null;

		prepareStaticStatements(conn);
		find_person.setInt(1,personId);
		
        ResultSet rs = find_person.executeQuery();

        /* Should get at most one row, since city_id is the primary key */
        while (rs.next()) {
            myperson = (Person) rs.getObject(1);
        }
             
        //close the result set 
        rs.close();


        return myperson;
    }


    /**
     * Get info about the last AUTOINCREMENT value inserted into the People table,
	 * and assign that number to the Person object inserted in the same row,
	 * and also return the new (updated) Person object. It is expected that you would
	 * call this method after inserting a person into the database.
     *
     * @param conn the connection to the database.
     *
     * @return the Person object updated with its id (that Person object is also updated
	 * in the database during this method).
     *
     * @exception SQLException thrown if database access fails.
     */

	public void updatePersonWithAutoincrementValue(Connection conn) throws SQLException {

		Statement s3 = conn.createStatement();
		ResultSet rs3 = s3.executeQuery(
			"VALUES ConnectionInfo.lastAutoincrementValue('APP', 'PEOPLE', 'PERSON_ID')");
		int id = 0;
		while (rs3.next()) {
			
			id = rs3.getInt(1);
		}
		rs3.close();
		s3.close();

		this.setId(id);
		PreparedStatement ps3=conn.prepareStatement("UPDATE People SET Person = ? WHERE person_id = ?");
		ps3.setObject(1, this);
		ps3.setInt(2, getId());
		ps3.executeUpdate();
		ps3.close();

	}

    /**
     * Is the child's parent already in the table. (Used in a check
	 * constraint.)
     *
     * @param     conn       Connection 
     *
     * @return    boolean     if parent is in table.
     *
     * @exception SQLException  Thrown if there is a SQL error.
     */
    public boolean parentInTable(Connection conn)
		 throws SQLException {
			 //do nothing unless the object is a Child.
			 return true;
    }


    /**
     * Good Cloudscape practice requires that objects override
	 * equals and hashCode so that = and <> operators work correctly
	 * in SQL comparisons.
     *
     * @param     Object      object to compare it to
     *
     * @return    boolean     if the objects are equal
     *
     */

	public boolean equals(Object o)  {

		Person op = null;
		// doesn't throw classCastException; instead
		// return false if Object is not Person
		try {
			op = (Person) o;
		}
		catch (ClassCastException cce) {
			return false;
		}
		

		if (this.id != op.id)
			return false;
		if (!this.firstName.equals(op.firstName))
			return false;
		if (!this.lastName.equals(op.lastName))
			return false;
		if (this.middleInitial != op.middleInitial)
			return false;
		if (birthdate != null) {
			if (!this.birthdate.equals(op.birthdate))
				return false;
		}
		return true;
	}

    /**
     * Good Cloudscape practice requires that objects override
	 * equals and hashCode so that = and <> operators work correctly
	 * in SQL comparisons.
     *
     * @param     Object      object to compare it to
     *
     * @return    boolean     if the objects are equal
     *
     */

	public int hashCode() {

		if (id != 0)
			return id;
		else
			return (int) birthdate.getTime();
	}
		

    /* Externalizable interface */

    public void writeExternal(ObjectOutput out) throws IOException {
        out.writeInt(currentClassVersion);
        out.writeInt(id);
        out.writeUTF(firstName);
        out.writeUTF(lastName);
        out.writeChar(middleInitial);
        out.writeObject(birthdate);
    }
 
    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
        int    version = in.readInt();
        if (version != currentClassVersion) {
            throw new IOException(
                "Expected class version " +
                currentClassVersion +
                " when reading Person, but got class version "
                + version);
        }

        id = in.readInt();
        firstName = in.readUTF();
        lastName = in.readUTF();
        middleInitial = in.readChar();
        birthdate = (Date) in.readObject();
    }




}
