package JBMSTours.serializabletypes;

import java.sql.*;
import java.util.Calendar;
import java.io.Serializable;
import java.util.Calendar;
import java.util.GregorianCalendar;
import JBMSTours.Util;
import JBMSTours.JCalendar;
import java.lang.Math;

//import java.lang.Comparable;

/**
 * Look at the <a href="../JBMSTours/serializabletypes/TimeZone.java">Source.</a><p>
 * This class represents a time zone.  Each time zone is some number
 * of hours from Greenwich Mean Time.  Time zones also differ in when
 * Daylight Savings Time starts and ends (and whether Daylight Savings
 * Time is used at all in the time zone).  In the northern hemisphere,
 * Daylight Savings Time always starts on a lower-numbered month than
 * it ends on, while in the southern hemisphere, Daylight Savings Time
 * always starts on a higher-numbered month than it ends on, because it
 * starts in winter and ends in spring of the next year. 
 * TimeZones are part of the City class (and are thus stored as part of
 * City objects in the Cities table).
 *
 */
public class TimeZone implements Serializable {

    //////////////////
    // fields
    //////////////////

    /**
     * Offset in hours from Greenwich Mean Time.
     */
    private double GMToffset;

    /**
     * Offset in hours from Greenwich Mean Time when Daylight Savings Time
     * is in effect.
     */
    private double DSTGMToffset;

    /**
     * A 3-character string representing the day   Daylight Savings Time begins.
     */
    private String beginDST;

    /**
     * A 3-character string representing the day that DaylightSavings Time ends.
     */
    private String endDST;

    /**
     * serialVersionUID allows us to make compatible changes to the class.
     */
    static final long serialVersionUID = -6875777552867642683L;

    //////////////////
    // constructors
    //////////////////


    /**
     * Creates a TimeZone with the given offset from Greenwich Mean Time,
     * and a given Daylight Savings Time behavior.
     *
     * @param GMToff    Offset in hours from Greenwich Mean Time
     * @param DSTGMToff Offset in hours from Greenwich Mean Time when
     *                  Daylight Savings Time is in effect.
     * @param beginTheDST  A String code that represents when Daylight Savings
     *                  Time begins.  The possible values are:
     *  <PRE>
     *      "FSA" - First Sunday in April
     *      "FSO" - First Sunday in October
     *      "SSO" - Second Sunday in October
     *      "SSM" - Second Sunday in March
     *      "SSF" - Second Sunday in February
     *      "FSS" - First Sunday in September
     *      "FSM15" - First Sunday on or after March 15
     *      "LSO" - Last Sunday in October
     *      "LSM" - Last Sunday in March
     *      "FSM" - First Sunday in May
     *      "TSO" - Third Sunday in October
     *      "TSS" - Third Sunday in September
     *      "FFM" - First Friday in May
     *      "LWS" - Last Wednesday in September
     *      "LSS" - Last Sunday in September
     *      "SSA" - Second Sunday in April
     *      "TSM" - Third Sunday in March
     *  </PRE>
     * @param endTheDST    A String code that represents when Daylight Savings
     *                  Time ends.  See above for possible values.
     */ 
    public TimeZone (
        double GMToff,
        double DSTGMToff,
        String beginTheDST,
        String endTheDST) {

        GMToffset = GMToff;
        DSTGMToffset = DSTGMToff;

        beginDST =beginTheDST;
        endDST = endTheDST;

    }

    /**
     * Creates a TimeZone with a given offset from Greenwich Mean Time,
     * and no Daylight Savings Time.
     *
     * @param GMToff    Offset in hours from Greenwich Mean Time
     */
    public TimeZone (
        double GMToff) {

        GMToffset = GMToff;
        DSTGMToffset = GMToff;
        beginDST = "000";
        endDST = "000";

    }

    ////////////////////
    // toString value
    ////////////////////

    /**
     * Format this TimeZone as a String.
     */
    public String toString() {
        return "GMT " + GMToffset;
    }

    //////////////////////////////////
    // accessors
    //////////////////////////////////

	public double getGMTOffset() {

		return GMToffset;

	}

	public double getDSTGMTOffset() {
		
		return DSTGMToffset;
		
	}

	public String getBeginDST() {
		
		return beginDST;
		
	}

	public String getEndDST() {
		
		return endDST;
		
	}

    ///////////////////////////////
    // various Date calculations
    ///////////////////////////////

    /**
     * Return true if the given date is in Daylight Savings Time for this
     * time zone.
     *
     * @param tzDate   A Date to check whether it is Daylight Savings Time
     *                  for this time zone.
     */
    private boolean isInDST(java.sql.Date tzDate) {

		
        Calendar calendar = new GregorianCalendar();
		calendar.setTime(tzDate);

		int dayOfYear = calendar.get(Calendar.DAY_OF_YEAR);

        if (beginDST.equals("000"))
              return false;

        int begin = parseString(beginDST, tzDate);
        int end = parseString(endDST, tzDate);

        if (end < begin) {
             
            if (dayOfYear > end && dayOfYear < begin) 
                return false;
            else return true;             
        }
        if (dayOfYear < begin  || dayOfYear > end)
            return false;
        else return true;

    }

    /**
     * Return the offset in hours between this time zone and Greenwich Mean
     * Time on the given Date.  The offset can depend on whether the Date
     * is in Daylight Savings Time.
     *
     * @param theDate   The Date for which to return the number of hours
     *                  between this time zone and Greenwich Mean Time.
     */
    public double getGMTOffset(java.sql.Date theDate) {
        if (isInDST(theDate))
            return  DSTGMToffset;
        else
            return GMToffset;
    }

    /**
     * Return the difference in time zones, in hours, between this time
     * zone and a given time zone on the given date. Returns a positive
     * number if the passed-in TimeZone is LATER, and a negative if the
     * passed-in TimeZone is earlier.
     *
     * @param tzb   The time zone for which we want to know how many
     *              hours difference there are with this time zone.
     * @param tzDate   The date on which we are interested in the hours
     *                  difference.  The difference can depend on whether
     *                  one time zone is in Daylight Savings Time and the
     *                  other time zone is not.
     */
    public double getTimeDifference(TimeZone tzb, java.sql.Date anyDate){

        double timea = this.getGMTOffset(anyDate);
        double timeb = tzb.getGMTOffset(anyDate);
        return timea - timeb;
    }

    //////////////////////////
    // static helper methods
    //////////////////////////

    /**
     * Numbers for days of the month, starting at zero for Sunday.
     */
    private static final int SUNDAY = Calendar.SUNDAY;
    private static final int MONDAY = Calendar.MONDAY;
    private static final int TUESDAY = Calendar.TUESDAY;
    private static final int WEDNESDAY = Calendar.WEDNESDAY;
    private static final int THURSDAY = Calendar.THURSDAY;
    private static final int FRIDAY = Calendar.FRIDAY;
    private static final int SATURDAY = Calendar.SATURDAY;

    /**
     * Convert the given String code to a Date.
     *
     * @param tzDate   A String code telling what Date to return.
     *                  The possible values are:
     *  <PRE>
     *      "FSA" - First Sunday in April
     *      "FSO" - First Sunday in October
     *      "SSO" - Second Sunday in October
     *      "SSM" - Second Sunday in March
     *      "SSF" - Second Sunday in February
     *      "FSS" - First Sunday in September
     *      "FSM15" - First Sunday on or after March 15
     *      "LSO" - Last Sunday in October
     *      "LSM" - Last Sunday in March
     *      "FSM" - First Sunday in May
     *      "TSO" - Third Sunday in October
     *      "TSS" - Third Sunday in September
     *      "FFM" - First Friday in May
     *      "LWS" - Last Wednesday in September
     *      "LSS" - Last Sunday in September
     *      "SSA" - Second Sunday in April
     *      "TSM" - Third Sunday in March
     *  </PRE>
     */
    private static int parseString(String offset, Date adate) {

    
        int year = JCalendar.getYear(adate);
        String Upper = offset.toUpperCase();
        // First Sunday in April
        if (Upper.equals("FSA"))
            return firstDayOnOrAfter(SUNDAY, year, Calendar.APRIL, 1);

        // First Sunday in October
        if (Upper.equals("FSO"))
            return firstDayOnOrAfter(SUNDAY, year, Calendar.OCTOBER, 1);

        // Second Sunday in October
        if (Upper.equals("SSO"))
            return firstDayOnOrAfter(SUNDAY, year, Calendar.OCTOBER, 8);

        // Second Sunday in March
        if (Upper.equals("SSM"))
            return firstDayOnOrAfter(SUNDAY, year, Calendar.MARCH, 8);

        // Second Sunday in February
        if (Upper.equals("SSF"))
            return firstDayOnOrAfter(SUNDAY, year, Calendar.FEBRUARY, 8);

        // First Sunday in September
        if (Upper.equals("FSS"))
            return firstDayOnOrAfter(SUNDAY, year, Calendar.SEPTEMBER, 1);

        // First Sunday on or after March 15
        if (Upper.equals("FSM15"))
            return firstDayOnOrAfter(SUNDAY, year, Calendar.MARCH, 15);

        // Last Sunday in October
        if (Upper.equals("LSO"))
            return lastDayOnOrBefore(SUNDAY, year, Calendar.OCTOBER);

        // Last Sunday in March
        if (Upper.equals("LSM"))
            return lastDayOnOrBefore(SUNDAY, year, Calendar.MARCH);

        // First Sunday in May
        if (Upper.equals("FSM"))
            return firstDayOnOrAfter(SUNDAY, year, Calendar.MAY, 1);

        // Third Sunday in October
        if (Upper.equals("TSO"))
            return firstDayOnOrAfter(SUNDAY, year, Calendar.OCTOBER, 15);

        // Third Sunday in September
        if (Upper.equals("TSS"))
            return firstDayOnOrAfter(SUNDAY, year, Calendar.SEPTEMBER, 15);
           
        // First Friday in May
        if (Upper.equals("FFM"))
            return firstDayOnOrAfter(FRIDAY, year, Calendar.MAY, 1);

        // Last Wednesday in September
        if (Upper.equals("LWS"))
            return lastDayOnOrBefore(WEDNESDAY, year, Calendar.SEPTEMBER);

        // Last Sunday in September
        if (Upper.equals("LSS"))
            return lastDayOnOrBefore(SUNDAY, year, Calendar.SEPTEMBER);

        // Second Sunday in April
        if (Upper.equals("SSA"))
            return firstDayOnOrAfter(SUNDAY, year, Calendar.APRIL, 8);

        // Third Sunday in March
        if (Upper.equals("TSM"))
            return firstDayOnOrAfter(SUNDAY, year, Calendar.MARCH, 15);

        return 0;
    }

    /**
     * Get the first day (DAY_OF_YEAR, an int) on the given
	 * day of the week that is on or after the given month, day, and year.
     *
     * @param dayOfWeek     The day of the week we're looking for, with
     *                      Sunday as 0.
     * @param year          The given year (see description above).
     * @param month         The given month (see description above).
     *                      January is 0.
     * @param day           The given day of the month.
     */
    private static int firstDayOnOrAfter(
                                    int dayOfWeek,
                                    int year,
                                    int month,
                                    int day) {

        Calendar calendar = new GregorianCalendar();
		calendar.set(year, month, day);
		int tempDayOfWeek = calendar.get(Calendar.DAY_OF_WEEK);
		if (tempDayOfWeek != dayOfWeek) {
			int diff = dayOfWeek - tempDayOfWeek;
			if (diff < 0)
				diff = diff+7;
			calendar.add(Calendar.DAY_OF_YEAR, diff);
		}
		return calendar.get(Calendar.DAY_OF_YEAR);
        
    }

    /**
     * Get the last day (DAY_OF_YEAR, an int, on the given day of the
	 * week that is on or before the given month, day, and year.
     *
     * @param dayOfWeek     The day of the week we're looking for, with
     *                      Sunday as 0.
     * @param year          The given year (see description above).
     * @param month         The given month (see description above).
     *                      January is 0.
     */
    private static int lastDayOnOrBefore(
                                    int dayOfWeek,
                                    int year,
                                    int month) {

        Calendar calendar = new GregorianCalendar();
		calendar.set(year, month, 1);
        int maxDay = calendar.getMaximum(GregorianCalendar.DAY_OF_MONTH);
		//get around bug in JDK that says that last day in September is the 31st;
		if (month == Calendar.SEPTEMBER && maxDay == 31)
			maxDay = 30;
        calendar.set(Calendar.DAY_OF_MONTH, maxDay);
        int lastDayInMonth = calendar.get(Calendar.DAY_OF_WEEK);
		if (lastDayInMonth!= dayOfWeek) {
			int diff = dayOfWeek - lastDayInMonth;
			if (diff > 0)
				diff = diff - 7;
			calendar.add(Calendar.DAY_OF_YEAR, diff);
		}
		return calendar.get(Calendar.DAY_OF_YEAR);

    }

	
	private static String parseDayOfWeek(int i) {
		
		switch(i) {
		case SUNDAY: return "Sunday";
		case MONDAY: return "Monday";
		case TUESDAY: return "Tuesday";
		case WEDNESDAY: return "Wednesday";
		case THURSDAY: return "Thursday";
		case FRIDAY: return "Friday";
		case SATURDAY: return "Saturday";
		default: return "unknown";
		}
		
	}
	private static String parseMonth(int i) {
		
		switch(i) {
		case Calendar.JANUARY: return "January";
		case Calendar.FEBRUARY: return "February";
		case Calendar.MARCH: return "March";
		case Calendar.APRIL: return "April";
		case Calendar.MAY: return "May";
		case Calendar.JUNE: return "June";
		case Calendar.JULY: return "July";
		case Calendar.AUGUST: return "August";
		case Calendar.SEPTEMBER: return "September";
		case Calendar.OCTOBER: return "October";
		case Calendar.NOVEMBER: return "November";
		case Calendar.DECEMBER: return "December";
		default: return "unknown";
		}
		
	}
	private static String parseDayOfYear(int i) {
		int year = 2000;
		return parseDayOfYear(i, year);
	}

	private static String parseDayOfYear(int i, int year) {
		Calendar c = new GregorianCalendar();
		c.set(year, Calendar.JANUARY, 1);
		c.add(Calendar.DAY_OF_YEAR, i-1);

		StringBuffer sb = new StringBuffer();

		sb.append(parseDayOfWeek(c.get(Calendar.DAY_OF_WEEK)));
		sb.append(", " + parseMonth(c.get(Calendar.MONTH)));
		sb.append(" " + c.get(Calendar.DATE) + ", ");
		sb.append(c.get(Calendar.YEAR));
		return sb.toString();
	}



    /**
     * Allows us to test DST settings for a given year.
	 *
     * @param year          The given year (see description above).
     */
	public static void showAllDSTZones(int year) {

		System.out.println("FSA, First Sunday in April");
        System.out.println(parseDayOfYear(firstDayOnOrAfter(SUNDAY, year, Calendar.APRIL, 1)));

		System.out.println("FSO, first Sunday in October");
		System.out.println(parseDayOfYear(firstDayOnOrAfter(SUNDAY, year, Calendar.OCTOBER, 1)));

		System.out.println("SSO, second Sunday in October");
		System.out.println(parseDayOfYear(firstDayOnOrAfter(SUNDAY, year, Calendar.OCTOBER, 8)));

		System.out.println("SSM, second Sunday in March");
		System.out.println(parseDayOfYear(firstDayOnOrAfter(SUNDAY, year, Calendar.MARCH, 8)));

		System.out.println("SSF, second Sunday in February");
		System.out.println(parseDayOfYear(firstDayOnOrAfter(SUNDAY, year, Calendar.FEBRUARY, 8)));

		System.out.println("FSS, First Sunday in September");
		System.out.println(parseDayOfYear(firstDayOnOrAfter(SUNDAY, year, Calendar.SEPTEMBER, 1)));
		
		System.out.println("FSM15, First Sunday on or after March 15");
		System.out.println(parseDayOfYear(firstDayOnOrAfter(SUNDAY, year, Calendar.MARCH, 15)));
		
		System.out.println("LSO, Last Sunday in October");
		System.out.println(parseDayOfYear(lastDayOnOrBefore(SUNDAY, year, Calendar.OCTOBER)));

		System.out.println("LSM, Last Sunday in March");
		System.out.println(parseDayOfYear(lastDayOnOrBefore(SUNDAY, year, Calendar.MARCH)));
		
		System.out.println("FSM, first Sunday in May");
		System.out.println(parseDayOfYear(firstDayOnOrAfter(SUNDAY, year, Calendar.MAY, 1)));
		
		System.out.println("TSO, third Sunday in October");
		System.out.println(parseDayOfYear(firstDayOnOrAfter(SUNDAY, year, Calendar.OCTOBER, 15)));
		
		System.out.println("TSS, third Sunday in September");
		System.out.println(parseDayOfYear(firstDayOnOrAfter(SUNDAY, year, Calendar.SEPTEMBER, 15)));
		
		System.out.println("FFM, First friday in May");
		System.out.println(parseDayOfYear(firstDayOnOrAfter(FRIDAY, year, Calendar.MAY, 1)));
		
		System.out.println("LWS, last Wednesday in September");
		System.out.println(parseDayOfYear(lastDayOnOrBefore(WEDNESDAY, year, Calendar.SEPTEMBER)));
		
		System.out.println("LSS, last Sunday in September");
		System.out.println(parseDayOfYear(lastDayOnOrBefore(SUNDAY, year, Calendar.SEPTEMBER)));
		
		System.out.println("SSA, second Sunday in April");
		System.out.println(parseDayOfYear(firstDayOnOrAfter(SUNDAY, year, Calendar.APRIL, 8)));
		
		System.out.println("TSM, third Sunday in March");
		System.out.println(parseDayOfYear(firstDayOnOrAfter(SUNDAY, year, Calendar.MARCH, 15)));
		
	}


    /**
     * This method allows a TimeZone object to be <i>orderable</i> within SQL-J
	 * statements.<p>
	 * Returns -1 if this object is "less than" the parameter, 1 if this object
	 * is "greater than" the parameter, and "0" if they are equal.<p>
	 * We have arbitrarily decided to order TimeZones by their
	 * Greenwich Mean Time offset (from high to low),
	 * and secondarily by their Daylight Savings Times Offsets.
     *
     * @return    int        -1, 0, or 1.
	 *
     * @exception ClassCastException  Thrown if the objects cannot be compared.
     *
	 * @see City
     *
     */


	public int compareTo(Object o) throws ClassCastException {
		//handle nulls;
		// nulls sort lower than everything
		if (o == null)
			return 1;
		//may throw an exception
        TimeZone otz = (TimeZone) o;
		double go1 = this.getGMTOffset();
		double go2 = otz.getGMTOffset();
		double bd1 = this.getDSTGMTOffset();
		double bd2 = otz.getDSTGMTOffset();
		
		if (go1 == go2 && bd1 == bd2)
			return 0;
		//first compare their GMT offsets
		if (go1 > go2)
			return 1;
		if (go1 < go2)
			return -1;
		// if their GMT offsets are equal,
		// then look at their Daylight Savings Times offsets
		if (bd1 > bd2)
			return 1;
		//the only other possibility is that bd2 < bd1
		
		else
			return -1;
	}

   /**
	*  For correct equality comparisons. Returns true if the two objects are "equal".
	*
    */
	public boolean equals(Object o)  {
        TimeZone oz;
		try {
			oz = (TimeZone) o;
		}
		catch (ClassCastException cce) {
			return false;
		}
		
		if (this.compareTo(oz)==0)
			return true;
		else
			return false;
	}

	/* for correct equality comparisons
	 *
     *
	 * @return int   a hash code
	 *
	 */
	 
	public int hashCode() {
		double go1 = Math.abs(this.getGMTOffset()*10);
		double bd1 = Math.abs(this.getDSTGMTOffset()*10);
		int go = (int) go1;
		int bd = (int) bd1;
		String s1 = new Integer(go).toString();
		String s2 = new Integer(bd).toString();
		String s3 = s1+s2;
		return new Integer(s3).intValue();


	}

}
