package JBMSTours.serializabletypes;

import java.sql.*;

import java.io.Externalizable;
import java.io.ObjectOutput;
import java.io.ObjectInput;
import java.io.IOException;
import java.math.BigDecimal;

import JBMSTours.Util;
import JBMSTours.JCalendar;
import JBMSTours.Group;
import JBMSTours.Hotel;
import JBMSTours.FlightBuilder;

import JBMSTours.CityChoiceException;
import JBMSTours.BudgetException;
import JBMSTours.AvailabilityException;


/**
 * Look at the <a href="../JBMSTours/serializabletypes/Tour.java">Source</a>.<p>
 * Basic class for building a customer tour. 
 * This is a two-city tour (first city of three is user's home city).
 * Main client application BuildATour constructs the object.
 * A tour consists of a Group, three cities, three dates
 * of travel, two Customer Flights and a two HotelStays.
 * A Tour gets stored in the CustomizedTours table.
 */
public class Tour implements Externalizable {

    //////////////////
    // fields
    //////////////////

    /**
     * first city (home city), transient (not stored). Store id only.
     */
    transient protected City city1;
    protected int    cityId1;
    protected String city1Name;

    /**
     * second city (middle city), transient (not stored). Store id only.
     */
    transient protected City city2;
    protected int cityId2;
    protected String city2Name;

    /**
     * third city (destination city), transient (not stored). Store id only.
     */
    transient protected City city3;
    protected int cityId3;
    protected String city3Name;
 
    /**
     * first date of travel (home city to first city).
     */
      public Date begin;

    /**
     * second date of travel (first city to second city).
     */
    public Date middle;

    /**
     * final date of travel (second city to home city).
     */
    public Date end;

    protected CustomerFlight myflight;
    protected CustomerFlight myflight2;
    protected CustomerFlight myflight3;

    protected HotelStay myStay;
    protected HotelStay secondStay;    

     /**
     * Group taking the tour, transient (not stored). Store id only.
     */
    transient protected Group myGroup;
    protected int        myGroupId;

    /**
     * Since some of these methods can be called as server-side methods, 
     * we set a flag to determine whether we can call the commit method.
     * By default, commits are allowed.
     *
     **/

    boolean commitsAllowed=true;


    public static final short ECONOMYTOURLEVEL = 1;
    public static final short STANDARDTOURLEVEL = 2;
    public static final short FIRSTCLASSTOURLEVEL = 3; 

    protected static final int currentClassVersion = 1;

    static final long serialVersionUID = -7862405295363648721L;

    /**
     * No arg constructer required. All externalizable objects need 
     * a no arg constructor. This one is only used when reading
     * the object in.
     *
     */
    public Tour() {
    }

    /**
     * Constructs a two-city tour object. 
     *
     * @param agroup    The group taking the tour.
     * @param one       The group's home city.
     * @param two       The first city on the tour.
     * @param three     The second city on the tour.
     * @param date1     The date the group travels to the first city.
     * @param date2     The date the group travels to the second city.
     * @param date3     The date the group returns home.
     */
    public Tour(
            Group    agroup,
            City    one,
            City    two,
            City    three,
            Date    date1,
            Date    date2,
            Date    date3) {
        city1 = one;
        cityId1 = city1.city_id;
        city1Name = city1.getName();
        city2 = two;
        cityId2 = city2.city_id;
        city2Name = city2.getName();
        city3 = three;
        cityId3 = city3.city_id;
        city3Name = city3.getName();

        begin = date1;
        middle = date2;
        end = date3;

        myGroup = agroup;
        myGroupId = myGroup.getGroupId();
        areCommitsAllowed();
     }


    /**
     * Find out whether the property "databaseSide" has been set. If so,
     * set the variable commitsAllowed to false.
     *     
     */

    private void areCommitsAllowed() {

    String p = System.getProperty("JBMSTours.databaseSide");
        if (p == null)
	  p = "false";
        if (p.equals("true")) 
	  commitsAllowed=false;

	}


    /**
     * Fills out the objects that are part of the tour.
     * Dynamically builds and stores flight objects based on cities
     * of travel. Suggests flights based on dates and cities of travel;
     * books flights; suggests hotels; books hotels.
     * Tracks the cost.
     *
     * @param conn      Connection to the db.
     *
     * @exception SQLException          Thrown if there is a SQL error.
     * @exception CityChoiceException   Thrown if if can't build tour
     *                                  to requested cities.
	 * @exception AvailabilityException if the tour can't find any available hotels or flights
     * @exception BudgetException       if the tour can't be made within budget
     */
    public void buildTheTour(Connection conn)
                    throws SQLException, CityChoiceException, BudgetException, AvailabilityException {

        FlightBuilder buildFlightObjects;
        buildFlightObjects = new FlightBuilder(conn, getCity1(conn), getCity2(conn), getCity3(conn), myGroupId);
	
        buildFlightObjects.constructFlightObjectTable();
       

        myflight =  new CustomerFlight(begin,
                           getCity1(conn),
                           getCity2(conn),
                           getGroup(conn).getSeats(conn),
                           getGroup(conn).getLevel(),
                           myGroupId);
        BigDecimal tempFunds = getGroup(conn).getAvailableFunds();
		try {
			tempFunds = tempFunds.subtract(myflight.proposeFlight(conn, tempFunds));
		}
		catch (Exception e) {
			if (e instanceof BudgetException || e instanceof AvailabilityException) {
				short templevel = temporarilyDowngradeTourLevel(getGroup(conn).getLevel());
				tempFunds = tempFunds.subtract(myflight.proposeFlight(conn, tempFunds, templevel));
				
			}
			else throw (SQLException) e;
		}
        commit(conn);
        Util.println();
        Util.println();


         myflight2 =  new CustomerFlight(middle,
                        getCity2(conn),
                        getCity3(conn),
                        getGroup(conn).getSeats(conn),
                        getGroup(conn).getLevel(),
                        myGroupId);
        try {
			tempFunds = tempFunds.subtract(myflight2.proposeFlight(conn, tempFunds));
		}
		catch (Exception e) {
			if (e instanceof BudgetException || e instanceof AvailabilityException) {
				short templevel = temporarilyDowngradeTourLevel(getGroup(conn).getLevel());
				tempFunds = tempFunds.subtract(myflight2.proposeFlight(conn, tempFunds, templevel));
			}
			else throw (SQLException) e;
		}
        commit(conn);
        Util.println();
        Util.println();

        myflight3 =  new CustomerFlight(end,
                        getCity3(conn),
                        getCity1(conn),
                        getGroup(conn).getSeats(conn),
                        getGroup(conn).getLevel(),
                        myGroupId);
        try {
			tempFunds = tempFunds.subtract(myflight3.proposeFlight(conn, tempFunds));
		}
		catch (Exception e) {
			if (e instanceof BudgetException || e instanceof AvailabilityException) {
				short templevel = temporarilyDowngradeTourLevel(getGroup(conn).getLevel());
				tempFunds = tempFunds.subtract(myflight3.proposeFlight(conn, tempFunds, templevel));
			}
			else
				throw (SQLException) e;
		}
        commit(conn);
    
        getGroup(conn).addToRunning(myflight.bookFlight(conn,
                                getGroup(conn).getGroupId())
                                    );
        commit(conn);
        getGroup(conn).addToRunning(myflight2.bookFlight(conn,
                                 getGroup(conn).getGroupId())
                                    );
        commit(conn);
        getGroup(conn).addToRunning(myflight3.bookFlight(conn,
                                 getGroup(conn).getGroupId())
                                                                        );
        commit(conn);

        myStay = new HotelStay(conn,
                       getCity2(conn),
                       getGroup(conn),
                       myflight,
                       myflight2
                            );
			tempFunds = getGroup(conn).getAvailableFunds();
		

		try {
			tempFunds = tempFunds.subtract(myStay.proposeHotelStay(conn, tempFunds));
		}
		catch (Exception e) {
			if (e instanceof BudgetException || e instanceof AvailabilityException) {
				short templevel = temporarilyDowngradeTourLevel(getGroup(conn).getLevel());
				tempFunds = tempFunds.subtract(myStay.proposeHotelStay(conn, tempFunds, templevel));
				
			}
			else
				throw (SQLException) e;
		}
		
        commit(conn);

        secondStay = new HotelStay(conn, 
                       getCity3(conn),
                       getGroup(conn),
                       myflight2,
                       myflight3
                                );
		try {
			tempFunds=tempFunds.subtract(secondStay.proposeHotelStay(conn, tempFunds));
		}
		catch (Exception e) {
			if (e instanceof BudgetException || e instanceof AvailabilityException) {
				short templevel = temporarilyDowngradeTourLevel(getGroup(conn).getLevel());
			    tempFunds = tempFunds.subtract(secondStay.proposeHotelStay(conn, tempFunds, templevel));
			}
			else throw (SQLException) e;
		}
        commit(conn);

        getGroup(conn).addToRunning(myStay.bookHotel(conn));
        commit(conn);

        getGroup(conn).addToRunning(secondStay.bookHotel(conn));
        commit(conn);

    }

    /**
     * Allows you to alter the tour to change one hotel stay for another one. (does not
	 * automatically update the CustomizedTours table).
     *
     * @param conn      Connection to the db.
     *
     * @exception SQLException          Thrown if there is a SQL error.
	 * @exception AvailabilityException if the tour can't find any available hotels or flights
     * @exception BudgetException       if the tour can't be made within budget
	 */
	public void findANewHotelStay(Connection conn, int old_hotel_id, BigDecimal old_cost)  throws BudgetException, AvailabilityException, SQLException {

		//find out which stay it was.
		int hid1 = myStay.hotelId;
        int hid2 = secondStay.hotelId;

		Statement s = conn.createStatement();
		//since the old hotel stay is no longer valid, delete the its cost from the Group's running total
		s.executeUpdate("UPDATE GROUPS set running_total = running_total - " + old_cost);
		if (old_hotel_id == hid1)
			updateMyStay(conn, old_hotel_id);
		if (old_hotel_id == hid2)
			updateSecondStay(conn, old_hotel_id);
		return;
	}

	private void updateMyStay(Connection conn, int bad_hotel_id) throws BudgetException, AvailabilityException, SQLException {
        myStay = new HotelStay(conn,
                       getCity2(conn),
                       getGroup(conn),
                       myflight,
                       myflight2
                            );
			BigDecimal tempFunds = getGroup(conn).getAvailableFunds();
			BigDecimal tempFunds2 = tempFunds;

		try {
			tempFunds = tempFunds.subtract(myStay.proposeHotelStay(conn, tempFunds, bad_hotel_id));
		}
		catch (Exception e) {

			if (e instanceof BudgetException || e instanceof AvailabilityException) {
				if (e instanceof BudgetException && getGroup(conn).getLevel() > 1) {
					short templevel = temporarilyDowngradeTourLevel(getGroup(conn).getLevel());
					tempFunds = tempFunds2.subtract(myStay.proposeHotelStay(conn, tempFunds, templevel, bad_hotel_id));
				}
				
				if (e instanceof AvailabilityException && getGroup(conn).getLevel() ==1) {
					short templevel = temporarilyUpgradeTourLevel(getGroup(conn).getLevel());
					tempFunds = tempFunds2.subtract(myStay.proposeHotelStay(conn, tempFunds, templevel, bad_hotel_id));
				}
			}
			
			
			
			else
				throw (SQLException) e;
		}
        getGroup(conn).addToRunning(myStay.bookHotel(conn));
		return;

	}

	private void updateSecondStay(Connection conn, int bad_hotel_id) throws BudgetException, AvailabilityException, SQLException {
         secondStay = new HotelStay(conn,
                       getCity3(conn),
                       getGroup(conn),
                       myflight2,
                       myflight3
                            );
			BigDecimal tempFunds = getGroup(conn).getAvailableFunds();
		    BigDecimal tempFunds2 = tempFunds;

		try {



			
			tempFunds = tempFunds.subtract(secondStay.proposeHotelStay(conn, tempFunds, bad_hotel_id));
		}
		catch (Exception e) {
			if (e instanceof BudgetException || e instanceof AvailabilityException) {
				if (e instanceof BudgetException && getGroup(conn).getLevel() > 1) {
					short templevel = temporarilyDowngradeTourLevel(getGroup(conn).getLevel());
					tempFunds = tempFunds2.subtract(secondStay.proposeHotelStay(conn, tempFunds, templevel, bad_hotel_id));
				}
				
				if (e instanceof AvailabilityException && getGroup(conn).getLevel() ==1) {
					short templevel = temporarilyUpgradeTourLevel(getGroup(conn).getLevel());
					tempFunds = tempFunds2.subtract(secondStay.proposeHotelStay(conn, tempFunds, templevel, bad_hotel_id));
				}
			}
			
			
			
			else
				throw (SQLException) e;


			
			
		}
        getGroup(conn).addToRunning(secondStay.bookHotel(conn));
		return;

	}

	
    /**
     * Prints a welcome message.
     *
     * @param conn      Connection to the db.
     *
     * @exception SQLException  Thrown if there is a SQL error.
     *
     */
    public void printWelcomeGroup(Connection conn) throws SQLException {
 
        StringBuffer welcomeMessage = new StringBuffer();
        welcomeMessage.append(getGroup(conn).getWelcomeMessage(conn));
		welcomeMessage.append(Util.lineSeparator());
        welcomeMessage.append("This application will build a customized tour" +
                           " for you from your home city to two cities.");
		welcomeMessage.append(Util.lineSeparator());
        welcomeMessage.append("This tour includes flights and hotel stays.");
		welcomeMessage.append(Util.lineSeparator());
        welcomeMessage.append("You wish to travel from your home city of ");
        welcomeMessage.append(city1.getName());
        welcomeMessage.append(" to ");
        welcomeMessage.append(city2.getName());
        welcomeMessage.append(" on ");
        welcomeMessage.append(begin);
        welcomeMessage.append(".");
		welcomeMessage.append(Util.lineSeparator());
        welcomeMessage.append("You wish to travel from ");
        welcomeMessage.append(city2.getName());
        welcomeMessage.append(" to ");
        welcomeMessage.append(city3.getName());
        welcomeMessage.append(" on ");
        welcomeMessage.append(middle);
        welcomeMessage.append(".");
		welcomeMessage.append(Util.lineSeparator());
        welcomeMessage.append("You wish to travel from ");
        welcomeMessage.append(city3.getName());
        welcomeMessage.append(" back to your home city of ");
        welcomeMessage.append(city1.getName());
        welcomeMessage.append(" on ");
        welcomeMessage.append(end);
        welcomeMessage.append(".");
		welcomeMessage.append(Util.lineSeparator());
        Util.println(welcomeMessage.toString());

    } 

    /**
     * Prints information about the tour.
     *
     * @param conn      Connection to the db.
     *
     * @exception SQLException  Thrown if there is a SQL error.
     *
     */
    public void printTourInfo(Connection conn) throws SQLException {
        Util.println(getTourInfo(conn));
    } 

    /**
     *
     * @param conn      Connection to the db.
     *
     * @exception SQLException  Thrown if there is a SQL error.
     *
     */
    public String getTourInfo(Connection conn) throws SQLException {
        StringBuffer sb = new StringBuffer();
		sb.append(getGroup(conn).getInfo(conn));
		sb.append(Util.lineSeparator());
        sb.append(getCity2(conn).getInfo(begin, getCity1(conn)));
		sb.append(Util.lineSeparator());
        sb.append(myStay.getInfo(conn));
		sb.append(Util.lineSeparator());
        sb.append(myflight.getInfo(conn));
		sb.append(Util.lineSeparator());
        sb.append(getCity3(conn).getInfo(begin, getCity1(conn)));
		sb.append(Util.lineSeparator());
        sb.append(secondStay.getInfo(conn));
		sb.append(Util.lineSeparator());
        sb.append(myflight2.getInfo(conn));
		sb.append(Util.lineSeparator());
        sb.append(myflight3.getInfo(conn));
		sb.append(Util.lineSeparator());
		return sb.toString();
    } 

        

    /**
     * Stores the tour as an object in the CustomizedTours table.
     * Transient fields are not stored.<p>
     * Uses query <em>INSERT INTO CUSTOMIZEDTOURS VALUES (?, ?)</em>
     *
     * @param conn      Connection to the db.
     *
     * @exception SQLException  Thrown if there is a SQL error.
     *
     */
    public void storeInDatabase (Connection conn) throws SQLException {
        PreparedStatement storeTour;

        storeTour = conn.prepareStatement(
                            "INSERT INTO CUSTOMIZEDTOURS VALUES (?, ?)");
        storeTour.setInt(1, getGroup(conn).getGroupId());
        storeTour.setObject(2, this);

        int n = storeTour.executeUpdate();

        Util.println(
            "This customized tour has been stored in the database.");
    }

    /**
     * Get a string value for the tour.
     *
     */
    public String toString() {
        return "customized tour of " +
            (city2Name) +
            " and " +
            (city3Name) + 
                        " from " +
                        city1Name +
                        ", dates " +
                        begin + 
                        " to " +
                        end + ".";
    }

    /**
     * Returns the first CustomerFlight object (flight from home city
     * to first city in tour).
     *
     * @see CustomerFlight
     *
     * @return      CustomerFlight      first flight in the tour.
     */
    public CustomerFlight getCustomerFlight1() {
        return myflight;
    }
 
    /**
     * Returns the first Flight object (flight from home city
     * to first city in tour). The Flight object (flight information
     * without specific date and number of seat information) is a field in the
     * CustomerFlight object.
     *
     * @see Flight
     * @see OneStopFlight
     * @see TransferFlight
     *
     */
    public Flight getFlight1() {
        return getCustomerFlight1().getFlightObject();
    }

    /**
     * Returns the second CustomerFlight object (flight from home city
     * to first city in tour).
     *
     * @see CustomerFlight
     *
     */
    public CustomerFlight getCustomerFlight2() {
        return myflight2;
    }

    /**
     * Returns the second Flight object (flight from home city
     * to first city in tour). The Flight object (flight information
     * without specific date and number of seat information) is a field in the
     * CustomerFlight object.
     *
     * @see Flight
     * @see OneStopFlight
     * @see TransferFlight
     *
     */
    public Flight getFlight2() {
        return getCustomerFlight2().getFlightObject();
    }

    /**
     * Returns the third CustomerFlight object (flight from home city
     * to first city in tour).
     *
     * @see CustomerFlight
     *
     */
    public CustomerFlight getCustomerFlight3() {
        return myflight3;
    }

    /**
     * Returns the third Flight object (flight from home city
     * to first city in tour). The Flight object (flight information
     * without specific date and number of seat information) is a field in the
     * CustomerFlight object.
     *
     * @see Flight
     * @see OneStopFlight
     * @see TransferFlight
     *
     */
    public Flight getFlight3() {
        return getCustomerFlight3().getFlightObject();
    }

    /**
     * Returns the first HotelStay object (HotelStay in
     * first city in tour).
     *
     * @see HotelStay
     *
     */
    public HotelStay getStay1() {
        return myStay; 
    }

    /**
     * Returns the first HotelStay object (HotelStay in
     * first city in tour).
     *
     * @see HotelStay
     *
     *
     */
    public HotelStay getStay2() {
        return secondStay; 
    }

    /**
     * Returns the first Hotel in the tour.
     * Hotel is a field in the HotelStay object.
     *
     * @see Hotel
     *
     * @param  conn            connection to the database.
     *
     *
     * @exception SQLException  Thrown if there is a SQL error.
     */
    public Hotel getHotel1(Connection conn) throws SQLException {
        return getStay1().getTheHotel(conn); 
    }

    /**
     * Returns the second Hotel in the tour.
     * Hotel is a field in the HotelStay object.
     *
     * @see Hotel
     *
     * @param  conn            connection to the database.
     *
     * @exception SQLException  Thrown if there is a SQL error.
     */
    public Hotel getHotel2(Connection conn) throws SQLException {
        return getStay2().getTheHotel(conn);
    }

    /**
     * Returns the total cost of the tour 
     * (cost of three flights for all group members and two hotel stays).
     *
     *
     */
    public BigDecimal getTotalCost() {
        BigDecimal runningCost;
        runningCost = getStay1().fixedRate;
        runningCost = runningCost.add(getStay2().fixedRate);
        runningCost = runningCost.add(getCustomerFlight1().totalCost);
        runningCost = runningCost.add(getCustomerFlight2().totalCost);
        runningCost = runningCost.add(getCustomerFlight3().totalCost); 
        return runningCost.setScale(2, BigDecimal.ROUND_CEILING);
    }

    /**
     * Returns the "level" of the tour (first class, standard, or economy).
     *
     * @param       conn        Connection to the database
     *
     *
     * @exception SQLException  Thrown if there is a SQL error.
     */
    public String getLevel(Connection conn) throws SQLException {
        return getStringForLevel(getGroup(conn).getLevel());
    }

    /**
     * Returns average cost of the tour per person, per day.
     *
     * @param       conn        Connection to the database
     *
     * @exception SQLException  Thrown if there is a SQL error.
     */
    public BigDecimal getAverageCostPerDayPerPerson(Connection conn) throws SQLException {

        int numberOfPeople;
        int numberOfDays;

        numberOfPeople = getGroup(conn).getNumberPeople();
        numberOfDays = (int) ((getCustomerFlight3().getArrivalDate(conn)).getTime() - begin.getTime())/86400000;

        return getTotalCost().divide(new BigDecimal((double) (numberOfPeople * numberOfDays)), 2);
    }

    /**
     * Returns total miles traveled (the mileage of all three flights)
     *
     */
    public int getTotalMilesTraveled() {
      return getFlight1().mileage + getFlight2().mileage + getFlight3().mileage;
    }

    /* Externalizable interface */
    public void writeExternal(ObjectOutput out) throws IOException {
        out.writeInt(currentClassVersion);
        out.writeInt(cityId1);
        out.writeInt(cityId2);
        out.writeInt(cityId3);
        out.writeUTF(city1Name);
        out.writeUTF(city2Name);
        out.writeUTF(city3Name);
        out.writeObject(begin);
        out.writeObject(middle);
        out.writeObject(end);
        out.writeObject(myflight);
        out.writeObject(myflight2);
        out.writeObject(myflight3);
        out.writeObject(myStay);
        out.writeObject(secondStay);
        out.writeInt(myGroupId);
    }

    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
        int    version = in.readInt();
        if (version != currentClassVersion) {
            throw new IOException(
                        "Expected class version " +
                        currentClassVersion +
                        " when reading Tour, but got class version "
                        + version);
        }

        cityId1 = in.readInt();
        cityId2 = in.readInt();
        cityId3 = in.readInt();

        city1Name = in.readUTF();
        city2Name = in.readUTF();
        city3Name = in.readUTF();
        begin = (Date) in.readObject();
        middle = (Date) in.readObject();
        end = (Date) in.readObject();
        myflight = (CustomerFlight) in.readObject();
        myflight2 = (CustomerFlight) in.readObject();
        myflight3 = (CustomerFlight) in.readObject();
        myStay = (HotelStay) in.readObject();
        secondStay = (HotelStay) in.readObject();

        myGroupId = in.readInt();
    }

    /*
    ** Accessor methods for the three cities.  If we have the right city
    ** object, just return it.  If we don't have it, find it and return it.
    */

    /**
     * Gets the first city (home city).
     *
     * @param       conn        Connection to the database
     *
     * @return      City        home city
     *
     * @exception SQLException  Thrown if there is a SQL error.
     */
    public City getCity1(Connection conn) throws SQLException{
        if (city1 != null)
            return city1;
        else
            return (city1 = City.findCity(conn, cityId1));
    }

    /**
     * Gets the first city visited (second city).
     *
     * @param       conn        Connection to the database
     *
     * @return      City        second city
     *
     * @exception SQLException  Thrown if there is a SQL error.
     */
    public City getCity2(Connection conn) throws SQLException {
        if (city2 != null)
            return city2;
        else
            return (city2 = City.findCity(conn, cityId2));
    }

   /**
     * Gets the second city visited (third city).
     *
     * @param       conn        Connection to the database
     *
     * @return      City        third city
     *
     * @exception SQLException  Thrown if there is a SQL error.
     */
    public City getCity3(Connection conn) throws SQLException {
        if (city3 != null)
            return city3;
        else
            return (city3 = City.findCity(conn, cityId3));
    }

    /**
     *
     * Accessor method for the group.  If we have the group object,
     * just return it.  If we don't have it, find it and return it.
     *
     * @param     conn       Connection 
     *
     * @return    Group      The Tour's Group field.
     *
     * @exception SQLException  Thrown if there is a SQL error.
     *
     */
    public Group getGroup(Connection conn) throws SQLException {
        if (myGroup != null)
            return myGroup;
        else
            return (myGroup = Group.getGroup(conn, myGroupId));
    }

   /**
     * Gets the String corresponding for the short level value.
     *
     * @param       level       the short representing the level of the tour
     *
     * @return      String      the corresponding String (e.g., "economy", "first class", etc.)
     */

  public static String getStringForLevel(short level) {

   if (level==1)
      return "economy";

   if (level==2)
     return "standard";

   if (level==3)
     return "first class";

   return "invalid level";
  }


   /**
     * Gets the short corresponding to the String representing the tour level.
     *
     * @param       level       the String representing the level of the tour (e.g., "first class", etc.)
     *
     */

  public static short getShortForLevel(String level) {

   if (level.equals("economy"))
      return 1;

   if (level.equals("standard"))
     return 2;

   if (level.equals("first class"))
     return 3;

   return 0;
  }

   /**
     * Gets the name of the static field in this class corresponding to short representing
     * the level of the tour.
     *
     * @param       level       the short representing the level of the tour
     *
     * @return      String      fullname of the Java field
     */

  public static String getJavaFieldForLevel (short level) {
   if (level==1)
      return "Tour.ECONOMYTOURLEVEL";

   if (level==2)
     return "Tour.STANDARDTOURLEVEL";

   if (level==3)
     return "Tour.FIRSTCLASSTOURLEVEL";
   return "invalid level";
  }
   /**
     * Gets the SQL-J version of the 
     * name of the static field in this class corresponding to short representing
     * the level of the tour.
     *
     * @param       level       the short representing the level of the tour
     *
     * @return      String      fullname of the Java field
     */

  public static String getSQLJFieldForLevel (short level) {
   if (level==1)
      return "Tour->ECONOMYTOURLEVEL";

   if (level==2)
     return "Tour->STANDARDTOURLEVEL";

   if (level==3)
     return "Tour->FIRSTCLASSTOURLEVEL";
   return "invalid level";
  }

   /**
     * Go down a level in the tour for booking reasons. 
     *
     * @param       level       the short representing the desired level of the tour
     *
     * @return      short       the desired level-1 if the level is greater than 1; otherwise, the desired level
     */


	public static short temporarilyDowngradeTourLevel(short level) {
		if (level > 1)
			return (short) (level- 1);
		else
			return level;
	}

	/**
     * Go up a level in the tour for booking reasons. 
     *
     * @param       level       the short representing the desired level of the tour
     *
     * @return      short       the desired level+1 if the level is less than 3; otherwise, the desired level
     */

	public static short temporarilyUpgradeTourLevel(short level) {
		if (level < 3)
			return (short) (level+ 1);
		else
			return level;
	}

	
  /* commits the connection, unless commits are not allowed */
  private void commit(Connection conn) throws SQLException {
        if (commitsAllowed)
            conn.commit();
  }

	/**
     * Good Cloudscape practice requires us to override equals and hashCode
	 * from Object.
     *
     * @return      boolean      if the objects are equal
     */

	public boolean equals(Object o)  {
        Tour ot = null;
		try {
			ot = (Tour) o;
		}
		catch (ClassCastException cce) {
			return false;
		}
		
		if (this.myGroupId != ot.myGroupId)
			 return false;
		if (this.cityId1 != ot.cityId1)
			return false;
		if (this.cityId2 != ot.cityId2)
			return false;
		if (this.cityId3 != ot.cityId3)
			return false;
		if (!this.begin.equals(ot.begin))
			return false;
		if (!this.middle.equals(ot.middle))
			return false;
		if (!this.end.equals(ot.end))
			return false;
		return true;

	}
	
	/**
     * Good Cloudscape practice erquires us to override equals and hashCode
	 * from Object.
     *
     */

	public int hashCode() {
		return myGroupId;
	}

}




