package JBMSTours.vti.jdbc1_2;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Timestamp;

import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;
import java.util.Enumeration;

import java.io.InputStream;
import java.io.IOException;

/**
 * Look at the <a href="../JBMSTours/vti/jdbc1_2/jarvti.java">Source</a>.<p>
 * This read-only VTI class opens a jar file and returns the metadata information 
 * in the form of
 * a table. You can select from an instance of this class provided that you give it
 * a proper parameter and a correlation name.
 * <p>
 * This class also provides some costing information to the optimizer.
 * If the actual jar file name is known at compilation, this class can provide very
 * accurate costing information. If not, the class returns some default information
 * that show that in general instantiating this class is very cheap (it's very cheap
 * to iterate through the metadata information in a jar file).
 */

public class jarvti extends COM.cloudscape.vti.VTITemplate implements COM.cloudscape.vti.VTICosting {

	private ResultSetMetaData metaData;
	private String jarName;
	private ZipFile zip;
	private Enumeration rows;
	private boolean open;
	private ZipEntry entry;
	public double myEstimatedRowCount = 5000;
	public double myEstimatedCost = 439;
	public Integer rowcount;


	// can be private, because implementation of VTICosting means
	// that Cloudscape will never call this method. Instead, will
	// always instantiate the class during optimization, and can
	// instead call the non-static version of this method.
	private static ResultSetMetaData getResultSetMetaData(String jarName) {
		return new jarmetadata();
	}

    /**
     * Instantiates the vti given a jar name.
     * 
     * @exception SQLException  Thrown if there is a SQL error.
     *
     */
	
	public jarvti(String jarName) throws SQLException {
		rowcount = null;
		metaData = getResultSetMetaData(jarName);
		if (metaData == null)
			return;
		this.jarName = jarName;
		//if the jarName is null, it is compile time AND the parameter
		// is a non-literal such as a ? or a column reference.
		// if so, don't open the jar. (Return only default optimization info).
		if (jarName == null)
			return;
		try {
			zip = new ZipFile(jarName);
		} catch (IOException ioe) {
			throw new SQLException(ioe.toString());
		}
		rows = zip.entries();
		open = true;

	}

   /**
     * Go to the next "row".
     * 
     * @exception SQLException  Thrown if there is a SQL error.
     *
     */

	public boolean next() throws SQLException {
		if (!rows.hasMoreElements()) {
			entry = null;
			return false;
		}

		entry = (ZipEntry) rows.nextElement();
		return true;
	}
   /**
     * Get the value of the specified data type from a column.
     * 
     * @exception SQLException  Thrown if there is a SQL error.
     *
     */

	public String getString(int column) throws SQLException {
		//System.out.println("getString " + column);
		checkOnRow();
		switch (column) {
		case 1: return entry.getName();
		default: return super.getString(column); // throw an exception
		}
	}
   /**
     * Get the value of the specified data type from a column.
     * 
     * @exception SQLException  Thrown if there is a SQL error.
     *
     */
	public boolean getBoolean(int column) throws SQLException {
		//System.out.println("getBoolean " + column);
		checkOnRow();
		switch (column) {
		case 2: return entry.isDirectory();
		default: return super.getBoolean(column); // throw an exception
		}
	}
   /**
     * Get the value of the specified data type from a column.
     * 
     * @exception SQLException  Thrown if there is a SQL error.
     *
     */
	public Timestamp getTimestamp(int column) throws SQLException {
		//System.out.println("getTimestamp " + column);
		checkOnRow();
		switch (column) {
		case 3: return new Timestamp(entry.getTime());
		default: return super.getTimestamp(column); // throw an exception
		}
	}
   /**
     * Get the value of the specified data type from a column.
     * 
     * @exception SQLException  Thrown if there is a SQL error.
     *
     */
	public long getLong(int column) throws SQLException {
		//System.out.println("getLong " + column);
		checkOnRow();
		switch (column) {
		case 4: return entry.getSize();
		case 5: return entry.getCompressedSize();
		default: return super.getLong(column); // throw an exception
		}
	}

   /**
     * Get the value of the specified data type from a column.
     * 
     * @exception SQLException  Thrown if there is a SQL error.
     *
     */
	public InputStream getBinaryStream(int column) throws SQLException {
		checkOnRow();
		switch (column) {
		case 6: 
			try {
				return zip.getInputStream(entry);
			} catch (IOException ioe) {
				throw new SQLException(ioe.toString());
			}
		default: return super.getBinaryStream(column); // throw an exception
		}
	}

   /**
     * Get the value of the specified data type from a column.
     * 
     * @exception SQLException  Thrown if there is a SQL error.
     *
     */
	public Object getObject(int col) throws SQLException {
		//System.out.println("getObject " + col);
		return Boolean.TRUE;
	}

   /**
     * Close the vti.
     * 
     *
     */
	public void close() {
		open = false;
		if (zip != null) {
			try {
				zip.close();
			} catch (IOException ioe) {
			}
			zip = null;
		}
	}

   /**
     * Return the metadata needed by Cloudscape for the ResultSet.
     * 
     *
     */
	public ResultSetMetaData getMetaData() {
		return metaData;
	}

   /**
     * See if the row is valid.
     * 
     * @exception SQLException  Thrown if there is a SQL error.
     *
     */
	private void checkOnRow() throws SQLException {

		if (!open)
			throw new SQLException("jar vti closed");

		if (entry == null)
			throw new SQLException("jar vti not on row");

	}


   /**
     * Return the estimated row count. Used by the Cloudscape optimizer.
     * 
     * @exception SQLException  Thrown if there is a SQL error.
     *
     */

	public double getEstimatedRowCount(COM.cloudscape.vti.VTIEnvironment vtiEnvironment) throws SQLException {
		if (rowcount!= null)
			return rowcount.doubleValue();
		// if it's compile time and the actual jar file is not known,
		// return the default costing information.
		if (!open)	 
			return myEstimatedRowCount;
		int count = 0;
		while (rows.hasMoreElements()) {
			count++;
			rows.nextElement();
		}
		
		rows = zip.entries();
		open = true;
		rowcount = new Integer(count);
		return rowcount.doubleValue();
		
		
	}
	
	/**
     * Return the estimated cost.. Used by the Cloudscape optimizer.
     * 
     * @exception SQLException  Thrown if there is a SQL error.
     *
     */
	
	public double getEstimatedCostPerInstantiation(COM.cloudscape.vti.VTIEnvironment vtiEnvironment) throws SQLException {
		
		// if it's compile time and the actual jar file is not known,
		// return the default costing information.
		if (!open)	 
			return myEstimatedCost;
        return .0879 * getEstimatedRowCount(vtiEnvironment);
}
  /**
     * Return whether this vti can be instantiated more than once (yes).
     * 
     * @exception SQLException  Thrown if there is a SQL error.
     *
     */

	public boolean supportsMultipleInstantiations(COM.cloudscape.vti.VTIEnvironment vtiEnvironment) throws SQLException {
		
        return true;
		
	}
	
	
	
}
