package JBMSTours.vti.jdbc2_0;
import COM.cloudscape.vti20.UpdatableVTITemplate;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.DriverManager;

/**
 * A read-write VTI class that gives access to a table in an Informix database.<p>
 * This class uses EITResult as a wrapper around a JDBC20 updatable ResultSet as the
 * ResultSet it returns in ExecuteQuery.<p>
 * Here is an example SQL-J statement using this VTI class as an ExternalVirtualTable:
 *<p>
 * <code>DELETE FROM NEW JBMSTours.vti.jdbc2_0.ExternalInformixTable(
 * 'jdbc:informix-sqli://janets:1025/toys:INFORMIXSERVER=ol_janets;user=informix;password=informix', * 'toys') WHERE "toy" like '%barbie%';</code><p>
 *
 * DO NOT compile or run this class in a pre-JDK 1.2 environment. It requires
 * JDK 1.2 or above.
 *
 *
 */
public class ExternalInformixTable extends UpdatableVTITemplate  {
	private Connection	conn;
	int		columnCount = 0;
	private String		url;
	private String		tableName;
	private String      query;
	boolean closed = false;
	private PreparedStatement ps;
	private EITResult rs;

	/**
	 *   Constructs the VTI class. Loads the Informix JDBC Driver and constructs
	 *   and prepares the query.
	 *
	 * @param url    The url of the external Informix database
	 * @param tableName  The name of the table
	 * 
	 *
	 * @exception SQLException on unexpected JDBC error
	 */
	public ExternalInformixTable(String url, String tableName) throws SQLException {
		this.url = url;
		this.tableName = tableName;
		this.query = "SELECT * FROM " + tableName;
		try {
			Class.forName("com.informix.jdbc.IfxDriver");

		}
		catch (Exception e) {
			throw (SQLException) e;
		}

		prepareTheStatement();

	}

	/**
	 * Must return java.sql.ResultSet.CONCUR_UPDATABLE to be a read-write VTI class.
	 * 
	 * @see java.sql.Statement
	 *
 	 * @exception SQLException on unexpected JDBC error
	 */
	public int getResultSetConcurrency() throws SQLException {
		return ResultSet.CONCUR_UPDATABLE;
	}

	/**
	 * Returns a ResultSet (an EITResult) to Cloudscape. Cloudscape calls this\
	 * method when it executes the SQL-J statement that uses this VTI class.
	 *
	 * @see java.sql.Statement
	 *
 	 * @exception SQLException on unexpected JDBC error
	 */
	public ResultSet executeQuery() throws SQLException {
		if (closed) {
	        throw new SQLException("close - already closed");
		}
		getAResultSet();
		return rs;
	}

    /**
     *  Provide the metadata for the query against the given table. Cloudscape
	 *  calls this method when it compiles the SQL-J statement that uses this
	 *  VTI class.
     *
     *   @return the result set metadata for the query
	 * 
     *   @exception SQLException thrown by JDBC calls
     */
    public ResultSetMetaData getMetaData() throws SQLException {
		if (closed) {
	        throw new SQLException("close - already closed");
		}

		prepareTheStatement();
		ResultSetMetaData rsmd = ps.getMetaData();
		columnCount = rsmd.getColumnCount();
		return rsmd;
    }

	/**
	 * Explicitly closes this PreparedStatement class. (Note: Cloudscape
	 * calls this method only after compiling a SELECT statement that uses
	 * this class.)<p>
	 *
	 * @see java.sql.Statement
	 *
 	 * @exception SQLException on unexpected JDBC error
	 */
	public void close() throws SQLException {
		if (closed) {
	        throw new SQLException("close - already closed");
		}

		conn.close();
		closed = true;
	}

	/**
	 * Return the number of columns in the user-specified table.
	 *
	 * @exception SQLException	Thrown if there is an error getting the
	 *							metadata.
	 */
	int getColumnCount() throws SQLException {
		if (closed) {
	        throw new SQLException("close - already closed");
		}

		if (columnCount == 0) {
			getMetaData();
		}
		
		return columnCount;
	}

	/**
	 * Get a connection for the user-supplied URL. Get it only once
	 * per use of this VTI.
	 */
	private Connection getAConnection() throws SQLException {
		if (conn == null || conn.isClosed()) {
			conn = DriverManager.getConnection(url);
			conn.setAutoCommit(false);
		}


		return conn;
	}

	
	/**
	 * Instantiate a user-defined ResultSet (EITResult) that is a wrapper for the
	 * DBMS's result set.
	 *
	 * We need a wrapper because we need to be able to explictly
	 * commit the connection at the close
	 * of each result set. .
	 */

	private void getAResultSet() throws SQLException {
		prepareTheStatement();
		rs = new EITResult(conn, ps);
		
	}

	/**
	 * Prepare the statement that SELECTs against the table.
	 * Prepares the statement
	 * with ResultSet.TYPE_FORWARD_ONLY, ResultSet.CONCUR_UPDATABLE.
	 * 
	 */
	
	private void prepareTheStatement() throws SQLException {

		getAConnection();
		if (ps == null || !(rs == null)) {
			ps=conn.prepareStatement(query,
									 ResultSet.TYPE_FORWARD_ONLY,
									 ResultSet.CONCUR_UPDATABLE);
			
		}
	}
	
}
